-------------------------------------------------------------------------------
--
-- Copyright 2020
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-------------------------------------------------------------------------------

-- Author : J Hargreaves
-- Purpose:  
--   Wrapper for full adc input test design
-- Description:
--   Unb2b version for lab testing
--   Contains complete AIT input stage with 12 ADC streams


LIBRARY IEEE, common_lib, unb2b_board_lib, diag_lib, dp_lib, tech_jesd204b_lib, lofar2_unb2b_adc_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE unb2b_board_lib.unb2b_board_pkg.ALL;
USE diag_lib.diag_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;

ENTITY lofar2_unb2b_adc_full IS
  GENERIC (
    g_design_name      : STRING  := "lofar2_unb2b_adc_full";
    g_design_note      : STRING  := "Lofar2 adc with all streams";
    g_sim              : BOOLEAN := FALSE; --Overridden by TB
    g_sim_unb_nr       : NATURAL := 0;
    g_sim_node_nr      : NATURAL := 0;
    g_stamp_date       : NATURAL := 0;  -- Date (YYYYMMDD) -- set by QSF
    g_stamp_time       : NATURAL := 0;  -- Time (HHMMSS)   -- set by QSF
    g_revision_id      : STRING := ""   -- revision ID     -- set by QSF
  );
  PORT (
    -- GENERAL
    CLK          : IN    STD_LOGIC; -- System Clock
    PPS          : IN    STD_LOGIC; -- System Sync
    WDI          : OUT   STD_LOGIC; -- Watchdog Clear
    INTA         : INOUT STD_LOGIC; -- FPGA interconnect line
    INTB         : INOUT STD_LOGIC; -- FPGA interconnect line

    -- Others
    VERSION      : IN    STD_LOGIC_VECTOR(c_unb2b_board_aux.version_w-1 DOWNTO 0);
    ID           : IN    STD_LOGIC_VECTOR(c_unb2b_board_aux.id_w-1 DOWNTO 0);
    TESTIO       : INOUT STD_LOGIC_VECTOR(c_unb2b_board_aux.testio_w-1 DOWNTO 0);
    
    -- I2C Interface to Sensors
    SENS_SC      : INOUT STD_LOGIC;
    SENS_SD      : INOUT STD_LOGIC;
  
    PMBUS_SC     : INOUT STD_LOGIC;
    PMBUS_SD     : INOUT STD_LOGIC;
    PMBUS_ALERT  : IN    STD_LOGIC := '0';

    -- 1GbE Control Interface
    ETH_CLK      : IN    STD_LOGIC;
    ETH_SGIN     : IN    STD_LOGIC_VECTOR(c_unb2b_board_nof_eth-1 DOWNTO 0);
    ETH_SGOUT    : OUT   STD_LOGIC_VECTOR(c_unb2b_board_nof_eth-1 DOWNTO 0);

    -- LEDs
    QSFP_LED     : OUT   STD_LOGIC_VECTOR(c_unb2b_board_tr_qsfp_nof_leds-1 DOWNTO 0);

     -- back transceivers (note only 6 are used in unb2b)
    BCK_RX       : IN    STD_LOGIC_VECTOR(c_unb2b_board_nof_tr_jesd204b + c_unb2b_board_start_tr_jesd204b-1  downto c_unb2b_board_nof_tr_jesd204b);
    BCK_REF_CLK  : IN    STD_LOGIC; -- Use as JESD204B_REFCLK
 
    -- jesd204b syncronization signals (2 syncs)
    JESD204B_SYSREF : IN    STD_LOGIC;
    JESD204B_SYNC_N : OUT   STD_LOGIC_VECTOR(c_unb2b_board_nof_sync_jesd204b-1 DOWNTO 0)
  );
END lofar2_unb2b_adc_full;
 
ARCHITECTURE str OF lofar2_unb2b_adc_full IS

  SIGNAL JESD204B_SERIAL_DATA       : STD_LOGIC_VECTOR((c_unb2b_board_tr_jesd204b.bus_w*c_unb2b_board_tr_jesd204b.nof_bus)-1  downto 0);
  SIGNAL jesd204b_sync_n_arr        : STD_LOGIC_VECTOR((c_unb2b_board_tr_jesd204b.bus_w*c_unb2b_board_tr_jesd204b.nof_bus)-1  downto 0);
  SIGNAL JESD204B_REFCLK            : STD_LOGIC;


BEGIN

  -- Mapping between JESD signal names and UNB2B pin/schematic names
  JESD204B_REFCLK <=  BCK_REF_CLK;
  JESD204B_SERIAL_DATA(0) <= BCK_RX(42);
  JESD204B_SERIAL_DATA(1) <= BCK_RX(43);
  JESD204B_SERIAL_DATA(2) <= BCK_RX(44);
  JESD204B_SERIAL_DATA(3) <= BCK_RX(45);
  JESD204B_SERIAL_DATA(4) <= BCK_RX(46);
  JESD204B_SERIAL_DATA(5) <= BCK_RX(47);
  JESD204B_SERIAL_DATA(6) <= '0';
  JESD204B_SERIAL_DATA(7) <= '0';
  JESD204B_SERIAL_DATA(8) <= '0';
  JESD204B_SERIAL_DATA(9) <= '0';
  JESD204B_SERIAL_DATA(10) <= '0';
  JESD204B_SERIAL_DATA(11) <= '0';
  JESD204B_SYNC_N(c_unb2b_board_nof_sync_jesd204b-1 DOWNTO 0) <= jesd204b_sync_n_arr(c_unb2b_board_nof_sync_jesd204b-1 DOWNTO 0);


  u_revision : ENTITY lofar2_unb2b_adc_lib.lofar2_unb2b_adc
  GENERIC MAP (
    g_design_name => g_design_name,
    g_design_note => g_design_note,
    g_sim         => g_sim,
    g_sim_unb_nr  => g_sim_unb_nr,
    g_sim_node_nr => g_sim_node_nr,
    g_stamp_date  => g_stamp_date,
    g_stamp_time  => g_stamp_time,
    g_revision_id => g_revision_id
  )
  PORT MAP (
    -- GENERAL
    CLK          => CLK,
    PPS          => PPS,
    WDI          => WDI,
    INTA         => INTA,
    INTB         => INTB,

    -- Others
    VERSION      => VERSION,
    ID           => ID,
    TESTIO       => TESTIO,

    -- I2C Interface to Sensors
    SENS_SC      => SENS_SC,
    SENS_SD      => SENS_SD,

    PMBUS_SC     => PMBUS_SC,
    PMBUS_SD     => PMBUS_SD,
    PMBUS_ALERT  => PMBUS_ALERT,

    -- 1GbE Control Interface
    ETH_clk      => ETH_clk,
    ETH_SGIN     => ETH_SGIN,
    ETH_SGOUT    => ETH_SGOUT,

    -- LEDs
    QSFP_LED     => QSFP_LED,

    -- back transceivers
    JESD204B_SERIAL_DATA   => JESD204B_SERIAL_DATA,
    JESD204B_REFCLK        => JESD204B_REFCLK,
  
    -- jesd204b syncronization signals
    JESD204B_SYSREF        => JESD204B_SYSREF,
    JESD204B_SYNC_N        => jesd204b_sync_n_arr
  );
END str;
