-------------------------------------------------------------------------------
--
-- Copyright 2020
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--
-- Author: R. van der Walle
-- Purpose: Self-checking testbench for simulating lofar2_unb2b_adc using WG data.
--
-- Description:
--   MM control actions:
--
--   1) Enable calc mode for WG via reg_diag_wg with:
--        freq = 20MHz
--        ampl = 0.5 * 2**13
--   
--   2) Read current BSN from reg_bsn_scheduler_wg and write reg_bsn_scheduler_wg 
--      to trigger start of WG at BSN.
--     
--   3) Read WG data via ram_aduh_mon into sp_sample and replay sp_sample for
--      analogue view in Wave window:
--   
--   4) Read ADUH monitor power sum for via reg_aduh_mon and verify with 
--      c_exp_wg_power_sp.
--      View sp_power_sum in Wave window
--
-- Usage:
--   > as 7    # default
--   > as 12   # for detailed debugging
--   > run -a  
--
-------------------------------------------------------------------------------
LIBRARY IEEE, common_lib, unb2b_board_lib, i2c_lib, mm_lib, dp_lib, diag_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE unb2b_board_lib.unb2b_board_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE common_lib.common_str_pkg.ALL;
USE mm_lib.mm_file_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE mm_lib.mm_file_unb_pkg.ALL;
USE diag_lib.diag_pkg.ALL;

ENTITY tb_lofar2_unb2b_adc_wg IS
END tb_lofar2_unb2b_adc_wg;

ARCHITECTURE tb OF tb_lofar2_unb2b_adc_wg IS

  CONSTANT c_sim             : BOOLEAN := TRUE;
  CONSTANT c_unb_nr          : NATURAL := 0; -- UniBoard 0
  CONSTANT c_node_nr         : NATURAL := 0; 
  CONSTANT c_id              : STD_LOGIC_VECTOR(7 DOWNTO 0) := "00000000";
  CONSTANT c_version         : STD_LOGIC_VECTOR(1 DOWNTO 0) := "00";
  CONSTANT c_fw_version      : t_unb2b_board_fw_version := (1, 0);

  CONSTANT c_eth_clk_period      : TIME := 8 ns;  -- 125 MHz XO on UniBoard
  CONSTANT c_ext_clk_period      : TIME := 5 ns;
  CONSTANT c_bck_ref_clk_period  : TIME := 5 ns;
  CONSTANT c_pps_period          : NATURAL := 1000;

  CONSTANT c_tb_clk_period       : TIME := 100 ps; -- use fast tb_clk to speed up M&C
  CONSTANT c_cable_delay         : TIME := 12 ns
;
  CONSTANT c_sample_freq         : NATURAL := c_unb2b_board_ext_clk_freq_200M/10**6;  -- 200 MSps
  CONSTANT c_sample_period       : TIME := (10**6 / c_sample_freq) * 1 ps; 

  CONSTANT c_nof_sync            : NATURAL := 5;
  CONSTANT c_nof_block_per_sync  : NATURAL := 16; 
 
  CONSTANT c_percentage          : REAL := 0.05;  -- percentage that actual value may differ from expected value
  CONSTANT c_lo_factor           : REAL := 1.0 - c_percentage;  -- lower boundary  
  CONSTANT c_hi_factor           : REAL := 1.0 + c_percentage;  -- higher boundary

  CONSTANT c_nof_points          : NATURAL := 1024; 
  CONSTANT c_nof_taps            : NATURAL := 16; 
  
  CONSTANT c_subband_period      : TIME := c_nof_points * c_sample_period;

  -- WG
  CONSTANT c_full_scale_ampl      : REAL := REAL(2**(18-1)-1);  -- = full scale of WG
  CONSTANT c_bsn_start_wg         : NATURAL := 2;  -- start WG at this BSN to instead of some BSN, to avoid mismatches in exact expected data values
  CONSTANT c_ampl_sp              : NATURAL := 2**(14-1)/2;  -- in number of lsb
  CONSTANT c_subband_sp           : REAL := 51.2;  -- Select subband at index 512/10 = 51.2 = 20 MHz
  CONSTANT c_wg_subband_freq_unit : REAL := c_diag_wg_freq_unit/512.0;  -- subband freq = Fs/512 = 200 MSps/512 = 390625 Hz sinus
  CONSTANT c_wg_ampl_lsb          : REAL := c_diag_wg_ampl_unit / c_full_scale_ampl;  -- amplitude in number of LSbit resolution steps
  CONSTANT c_exp_wg_power_sp      : REAL := REAL(c_ampl_sp**2)/2.0 * REAL(c_nof_points*c_nof_block_per_sync);
   
  -- ADUH
  CONSTANT c_mon_buffer_nof_samples  : NATURAL := 1024; --samples per stream 
  CONSTANT c_mon_buffer_nof_words    : NATURAL := c_mon_buffer_nof_samples; 
  
  -- MM  
  CONSTANT c_mm_file_reg_ppsh               : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "PIO_PPS";
  CONSTANT c_mm_file_reg_bsn_source         : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "REG_BSN_SOURCE";
  CONSTANT c_mm_file_reg_bsn_scheduler_wg   : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "REG_BSN_SCHEDULER";
  CONSTANT c_mm_file_reg_diag_wg            : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "REG_WG";
  CONSTANT c_mm_file_reg_aduh_mon           : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "REG_ADUH_MONITOR";
  CONSTANT c_mm_file_ram_aduh_mon           : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "RAM_ADUH_MONITOR";


  -- Tb
  SIGNAL tb_end              : STD_LOGIC := '0';
  SIGNAL sim_done            : STD_LOGIC := '0';
  SIGNAL tb_clk              : STD_LOGIC := '0';  
  SIGNAL rd_data             : STD_LOGIC_VECTOR(c_32-1 DOWNTO 0);

  -- WG
  SIGNAL dbg_c_exp_wg_power_sp : REAL := c_exp_wg_power_sp;
  SIGNAL sp_samples            : t_integer_arr(0 TO c_mon_buffer_nof_samples-1) := (OTHERS=>0); 
  SIGNAL sp_sample             : INTEGER := 0;
  SIGNAL sp_power_sum          : UNSIGNED(63 DOWNTO 0);
  SIGNAL current_bsn_wg        : STD_LOGIC_VECTOR(c_dp_stream_bsn_w-1 DOWNTO 0);

  -- DUT
  SIGNAL ext_clk             : STD_LOGIC := '0';
  SIGNAL pps                 : STD_LOGIC := '0';
  SIGNAL ext_pps             : STD_LOGIC := '0'; 
  SIGNAL pps_rst             : STD_LOGIC := '0';

  SIGNAL WDI                 : STD_LOGIC;
  SIGNAL INTA                : STD_LOGIC;
  SIGNAL INTB                : STD_LOGIC;

  SIGNAL eth_clk             : STD_LOGIC := '0';
  SIGNAL eth_txp             : STD_LOGIC_VECTOR(c_unb2b_board_nof_eth-1 downto 0);
  SIGNAL eth_rxp             : STD_LOGIC_VECTOR(c_unb2b_board_nof_eth-1 downto 0);

  SIGNAL sens_scl            : STD_LOGIC;
  SIGNAL sens_sda            : STD_LOGIC;
  SIGNAL pmbus_scl           : STD_LOGIC;
  SIGNAL pmbus_sda           : STD_LOGIC;

  -- back transceivers
  SIGNAL JESD204B_SERIAL_DATA : STD_LOGIC_VECTOR((c_unb2b_board_tr_jesd204b.bus_w * c_unb2b_board_tr_jesd204b.nof_bus)-1 downto 0);
  SIGNAL JESD204B_REFCLK      : STD_LOGIC := '1';

  -- jesd204b syncronization signals
  SIGNAL jesd204b_sysref     : STD_LOGIC;
  SIGNAL jesd204b_sync_n     : STD_LOGIC_VECTOR((c_unb2b_board_tr_jesd204b.nof_bus * c_unb2b_board_tr_jesd204b.bus_w)-1 DOWNTO 0);

BEGIN

  ----------------------------------------------------------------------------
  -- System setup
  ----------------------------------------------------------------------------
  ext_clk <= NOT ext_clk AFTER c_ext_clk_period/2;  -- External clock (200 MHz)
  eth_clk <= NOT eth_clk AFTER c_eth_clk_period/2;  -- Ethernet ref clock (125 MHz)
  JESD204B_REFCLK <= NOT JESD204B_REFCLK AFTER c_bck_ref_clk_period/2;  -- JESD sample clock (200MHz) 

  INTA <= 'H';  -- pull up
  INTB <= 'H';  -- pull up

  sens_scl <= 'H';  -- pull up
  sens_sda <= 'H';  -- pull up
  pmbus_scl <= 'H';  -- pull up
  pmbus_sda <= 'H';  -- pull up

  ------------------------------------------------------------------------------
  -- External PPS
  ------------------------------------------------------------------------------  
  proc_common_gen_pulse(1, c_pps_period, '1', pps_rst, ext_clk, pps);
  jesd204b_sysref <= pps;
  ext_pps <= pps;

  ------------------------------------------------------------------------------
  -- DUT
  ------------------------------------------------------------------------------
  u_lofar_unb2b_adc : ENTITY work.lofar2_unb2b_adc
  GENERIC MAP (
    g_design_name => "lofar2_unb2b_adc_one_node",
    g_design_note => "Lofar2 adc with one node",
    g_sim         => c_sim,
    g_sim_unb_nr  => c_unb_nr,
    g_sim_node_nr => c_node_nr
  )
  PORT MAP (
    -- GENERAL
    CLK          => ext_clk,
    PPS          => pps,
    WDI          => WDI,
    INTA         => INTA,
    INTB         => INTB,

    -- Others
    VERSION      => c_version,
    ID           => c_id,
    TESTIO       => open,

    -- I2C Interface to Sensors
    SENS_SC      => sens_scl,
    SENS_SD      => sens_sda,

    PMBUS_SC     => pmbus_scl,
    PMBUS_SD     => pmbus_sda,
    PMBUS_ALERT  => open,

    -- 1GbE Control Interface
    ETH_CLK      => eth_clk,
    ETH_SGIN     => eth_rxp,
    ETH_SGOUT    => eth_txp,

    -- LEDs
    QSFP_LED     => open,

    -- back transceivers
    JESD204B_SERIAL_DATA  => JESD204B_SERIAL_DATA,
    JESD204B_REFCLK       => JESD204B_REFCLK,
  
    -- jesd204b syncronization signals
    JESD204B_SYSREF => jesd204b_sysref,
    JESD204B_SYNC_N => jesd204b_sync_n
  );

  ------------------------------------------------------------------------------
  -- MM slave accesses via file IO
  ------------------------------------------------------------------------------
  tb_clk  <= NOT tb_clk AFTER c_tb_clk_period/2;    -- Testbench MM clock
  
  p_mm_stimuli : PROCESS
    VARIABLE v_bsn                   : NATURAL;
    VARIABLE v_sp_power_sum          : REAL;
  BEGIN
    -- Wait for DUT power up after reset
    WAIT FOR 1 us;
    
    proc_common_wait_until_hi_lo(ext_clk, ext_pps);
        
    ----------------------------------------------------------------------------
    -- Enable BS
    ----------------------------------------------------------------------------
    mmf_mm_bus_wr(c_mm_file_reg_bsn_source, 3,                    0, tb_clk);
    mmf_mm_bus_wr(c_mm_file_reg_bsn_source, 2,                    0, tb_clk);  -- Init BSN = 0
    mmf_mm_bus_wr(c_mm_file_reg_bsn_source, 1, c_nof_block_per_sync, tb_clk);  -- nof_block_per_sync
    mmf_mm_bus_wr(c_mm_file_reg_bsn_source, 0,         16#00000003#, tb_clk);  -- Enable BS at PPS
    
    ----------------------------------------------------------------------------
    -- Enable WG
    ----------------------------------------------------------------------------
    --   0 : mode[7:0]           --> off=0, calc=1, repeat=2, single=3)
    --       nof_samples[31:16]  --> <= c_ram_wg_size=1024
    --   1 : phase[15:0]
    --   2 : freq[30:0]
    --   3 : ampl[16:0]
    mmf_mm_bus_wr(c_mm_file_reg_diag_wg, 0, 1024*2**16 + 1, tb_clk);  -- nof_samples, mode calc
    mmf_mm_bus_wr(c_mm_file_reg_diag_wg, 1, INTEGER(  0.0 * c_diag_wg_phase_unit), tb_clk);  -- phase offset in degrees
    mmf_mm_bus_wr(c_mm_file_reg_diag_wg, 2, INTEGER(c_subband_sp * c_wg_subband_freq_unit), tb_clk);  -- freq
    mmf_mm_bus_wr(c_mm_file_reg_diag_wg, 3, INTEGER(REAL(c_ampl_sp) * c_wg_ampl_lsb), tb_clk);  -- ampl
    
    -- Read current BSN
    mmf_mm_bus_rd(c_mm_file_reg_bsn_scheduler_wg, 0, current_bsn_wg(31 DOWNTO  0), tb_clk);
    mmf_mm_bus_rd(c_mm_file_reg_bsn_scheduler_wg, 1, current_bsn_wg(63 DOWNTO 32), tb_clk);
    proc_common_wait_some_cycles(tb_clk, 1);
    
    -- Write scheduler BSN to trigger start of WG at next block
    v_bsn := TO_UINT(current_bsn_wg) + 2;
    ASSERT v_bsn <= c_bsn_start_wg REPORT "Too late to start WG: " & int_to_str(v_bsn) & " > " & int_to_str(c_bsn_start_wg) SEVERITY ERROR;
    v_bsn := c_bsn_start_wg;
    mmf_mm_bus_wr(c_mm_file_reg_bsn_scheduler_wg, 0, v_bsn, tb_clk);  -- first write low then high part
    mmf_mm_bus_wr(c_mm_file_reg_bsn_scheduler_wg, 1,     0, tb_clk);  -- assume v_bsn < 2**31-1
    
    -- Wait for ADUH monitor to have filled with WG data
    WAIT FOR c_subband_period*c_nof_taps;
    WAIT FOR c_subband_period*2;

    ----------------------------------------------------------------------------
    -- WG data : read ADUH monitor buffer
    ----------------------------------------------------------------------------
    -- Wait for start of sync interval 
    mmf_mm_wait_until_value(c_mm_file_reg_bsn_scheduler_wg, 0,                  -- read BSN low
                            "UNSIGNED", rd_data, ">=", c_nof_block_per_sync*2,  -- this is the wait until condition
                            c_subband_period, tb_clk);
    
    WAIT FOR c_subband_period;  -- ensure that one block of samples has filled the ADUH monitor buffer after the sync

    -- Read via MM    
    FOR I IN 0 TO c_mon_buffer_nof_words-1 LOOP
      mmf_mm_bus_rd(c_mm_file_ram_aduh_mon, I, rd_data, tb_clk);
      sp_samples(I) <= TO_SINT(rd_data(15 DOWNTO 0)); 
    END LOOP;
    
    -- Play to have waveform in time to allow viewing as analogue in the Wave Window
    FOR I IN 0 TO c_mon_buffer_nof_words-1 LOOP
      proc_common_wait_some_cycles(ext_clk, 1);
      sp_sample <= sp_samples(I);
    END LOOP;
 
    WAIT FOR c_subband_period*3;

    ---------------------------------------------------------------------------
    -- Read ADUH monitor power sum 
    ---------------------------------------------------------------------------
    -- Wait for start of sync interval
    mmf_mm_wait_until_value(c_mm_file_reg_bsn_scheduler_wg, 0,                   -- read BSN low
                            "UNSIGNED", rd_data, ">=", c_nof_block_per_sync*3,   -- this is the wait until condition
                            c_subband_period, tb_clk);
    
    -- Read ADUH monitor power sum 
    mmf_mm_bus_rd(c_mm_file_reg_aduh_mon, 2, rd_data, tb_clk);  -- read low part
    sp_power_sum(31 DOWNTO 0) <= UNSIGNED(rd_data);
    mmf_mm_bus_rd(c_mm_file_reg_aduh_mon, 3, rd_data, tb_clk);  -- read high part
    sp_power_sum(63 DOWNTO 32) <= UNSIGNED(rd_data);
    proc_common_wait_some_cycles(tb_clk, 1);

    ---------------------------------------------------------------------------
    -- Verification 
    ---------------------------------------------------------------------------
    -- Convert UNSIGNED sp_power_sum to REAL
    v_sp_power_sum := REAL(REAL(TO_INTEGER(sp_power_sum(61 DOWNTO 30)))*REAL(2**30) + REAL(TO_INTEGER(sp_power_sum(29 DOWNTO 0)))); 

    ASSERT v_sp_power_sum > c_lo_factor * c_exp_wg_power_sp REPORT "Wrong SP power for SP 0" SEVERITY ERROR;
    ASSERT v_sp_power_sum < c_hi_factor * c_exp_wg_power_sp REPORT "Wrong SP power for SP 0" SEVERITY ERROR;

    ---------------------------------------------------------------------------
    -- End Simulation 
    ---------------------------------------------------------------------------   
    sim_done <= '1';
    proc_common_wait_some_cycles(ext_clk, 100);
    proc_common_stop_simulation(TRUE, ext_clk, sim_done, tb_end);
    WAIT;
  END PROCESS;
  
END tb;
