-------------------------------------------------------------------------------
--
-- Copyright 2020
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--
-- Author: R. van der Walle (original), E. Kooistra (updates)
-- Purpose: Self-checking testbench for simulating lofar2_unb2b_sdp_station_fsub using WG data.
--
-- Description:
--   MM control actions:
--
--   1) Enable calc mode for WG via reg_diag_wg with:
--        freq = 19.921875MHz
--        ampl = 0.5 * 2**13
--   
--   2) Read current BSN from reg_bsn_scheduler_wg and write reg_bsn_scheduler_wg 
--      to trigger start of WG at BSN.
--     
--   3) Read subband statistics (SST) via MM and verify with exp_subband_sst at g_subband.
--      . use weighted subbands (default selected by MM)
--
--   4) View in wave window
--      . in_sosi.sop and in_data in u_si_arr(g_sp) to check that:
--        - WG starts with zero phase sine when c_subband_phase = 0.0 degrees
--        - WG amplitude = 8191 (is full scale - 1) when wg_ampl = 1.0
--      . pfb_sosi_arr(c_pfb_index).im/re and fsub_sosi_arr(c_pfb_index).im/re
--        in u_fsub in decimal radix and analog format to check that subband
--        phase is g_subband_weight_phase phase as set by the subband weight.
--        - Raw: pfb_sosi_arr = atan2(-65195 / 0) = -90 degrees
--        - Weighted: fsub_sosi_arr = atan2(-56457 / 32598) = -60 degrees
--          --> rotated expected g_subband_weight_phase = -60 - -90 = +30 degrees.
--   
-- Usage:
--   > as 7    # default
--   > as 12   # for detailed debugging
--   # Manually add missing signal
--   > add wave -position insertpoint  \
--     sim:/tb_lofar2_unb2b_sdp_station_fsub/sp_subband_ssts_arr2
--   > run -a
--   # Takes about 30 m when g_read_all_SST = FALSE
--   # Takes about 40 m when g_read_all_SST = TRUE
--
-------------------------------------------------------------------------------
LIBRARY IEEE, common_lib, unb2b_board_lib, i2c_lib, mm_lib, dp_lib, diag_lib, lofar2_sdp_lib, wpfb_lib, lofar2_unb2b_sdp_station_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE IEEE.math_real.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE common_lib.common_str_pkg.ALL;
USE mm_lib.mm_file_pkg.ALL;
USE mm_lib.mm_file_unb_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE diag_lib.diag_pkg.ALL;
USE wpfb_lib.wpfb_pkg.ALL;
USE lofar2_sdp_lib.sdp_pkg.ALL;
USE unb2b_board_lib.unb2b_board_pkg.ALL;

ENTITY tb_lofar2_unb2b_sdp_station_fsub IS
  GENERIC (
    g_sp                   : NATURAL := 3;    -- signal path index in range(S_pn = 12)
    g_wg_ampl              : REAL := 1.0;     -- WG normalized amplitude
    g_subband              : NATURAL := 102;  -- select subband at index 102 = 102/1024 * 200MHz = 19.921875 MHz
    g_subband_weight_gain  : REAL := 1.0;     -- subband weight normalized gain
    g_subband_weight_phase : REAL := 30.0;    -- subband weight phase rotation in degrees
    g_read_all_SST         : BOOLEAN := TRUE  -- when FALSE only read SST for g_subband, to save sim time
  );
END tb_lofar2_unb2b_sdp_station_fsub;

ARCHITECTURE tb OF tb_lofar2_unb2b_sdp_station_fsub IS

  CONSTANT c_sim             : BOOLEAN := TRUE;
  CONSTANT c_unb_nr          : NATURAL := 0; -- UniBoard 0
  CONSTANT c_node_nr         : NATURAL := 0;
  CONSTANT c_init_bsn        : NATURAL := 17;  -- some recognizable value >= 0

  CONSTANT c_id              : STD_LOGIC_VECTOR(7 DOWNTO 0) := "00000000";
  CONSTANT c_version         : STD_LOGIC_VECTOR(1 DOWNTO 0) := "00";
  CONSTANT c_fw_version      : t_unb2b_board_fw_version := (1, 0);

  CONSTANT c_eth_clk_period      : TIME := 8 ns;  -- 125 MHz XO on UniBoard
  CONSTANT c_ext_clk_period      : TIME := 5 ns;
  CONSTANT c_bck_ref_clk_period  : TIME := 5 ns;

  CONSTANT c_tb_clk_period       : TIME := 100 ps; -- use fast tb_clk to speed up M&C

  CONSTANT c_nof_block_per_sync  : NATURAL := 16;
  CONSTANT c_nof_clk_per_sync    : NATURAL := c_nof_block_per_sync * c_sdp_N_fft;
  CONSTANT c_pps_period          : NATURAL := c_nof_clk_per_sync;
  CONSTANT c_wpfb_sim            : t_wpfb := func_wpfb_set_nof_block_per_sync(c_sdp_wpfb_subbands, c_nof_block_per_sync);
  CONSTANT c_stat_data_sz        : NATURAL := c_wpfb_sim.stat_data_sz;  -- = 2

  CONSTANT c_percentage          : REAL := 0.05;  -- percentage that actual value may differ from expected value
  CONSTANT c_lo_factor           : REAL := 1.0 - c_percentage;  -- lower boundary  
  CONSTANT c_hi_factor           : REAL := 1.0 + c_percentage;  -- higher boundary

  -- WG
  CONSTANT c_bsn_start_wg         : NATURAL := c_init_bsn + 2;  -- start WG at this BSN to instead of some BSN, to avoid mismatches in exact expected data values
  -- .ampl
  CONSTANT c_wg_ampl              : NATURAL := NATURAL(g_wg_ampl * REAL(c_sdp_FS_adc));  -- in number of lsb
  CONSTANT c_exp_sp_power         : REAL := REAL(c_wg_ampl**2) / 2.0;
  CONSTANT c_exp_sp_ast           : REAL := c_exp_sp_power * REAL(c_nof_clk_per_sync);
  -- . phase
  CONSTANT c_subband_phase        : REAL := 0.0;  -- wanted subband phase in degrees = WG phase at sop
  CONSTANT c_subband_freq         : REAL := REAL(g_subband) / REAL(c_sdp_N_fft);                   -- normalized by fs = f_adc = 200 MHz is dp_clk
  CONSTANT c_wg_latency           : INTEGER := c_diag_wg_latency - 0;  -- -0 to account for BSN scheduler start trigger latency
  CONSTANT c_wg_phase_offset      : REAL := 360.0 * REAL(c_wg_latency) * c_subband_freq;  -- c_diag_wg_latency is in dp_clk cycles
  CONSTANT c_wg_phase             : REAL := c_subband_phase + c_wg_phase_offset;  -- WG phase in degrees

  -- FSUB
  -- . WPFB
  CONSTANT c_pol_index                      : NATURAL := g_sp MOD c_sdp_Q_fft;
  CONSTANT c_pfb_index                      : NATURAL := g_sp / c_sdp_Q_fft;  -- only read used WPFB unit out of range(c_sdp_P_pfb = 6)
  CONSTANT c_exp_subband_ampl_raw           : REAL := REAL(c_wg_ampl) * c_sdp_wpfb_subband_sp_ampl_ratio;
  CONSTANT c_exp_subband_ampl_weighted      : REAL := c_exp_subband_ampl_raw * g_subband_weight_gain;
  CONSTANT c_exp_subband_power_raw          : REAL := c_exp_subband_ampl_raw**2.0;  -- complex, so no divide by 2
  CONSTANT c_exp_subband_power_weighted     : REAL := c_exp_subband_ampl_weighted**2.0;  -- complex, so no divide by 2
  CONSTANT c_exp_subband_sst_raw            : REAL := c_exp_subband_power_raw * REAL(c_nof_block_per_sync);
  CONSTANT c_exp_subband_sst_weighted       : REAL := c_exp_subband_power_weighted * REAL(c_nof_block_per_sync);

  -- . expected limit values, obtained with print_str() for g_subband = 102,
  --   g_wg_ampl = 1.0, g_subband_weight_gain = 1.0, g_subband_weight_phase = 30.0
  CONSTANT c_exp_subband_sst_leakage_snr_dB   : REAL := 70.0;  -- < 74.913
  CONSTANT c_exp_subband_sst_crosstalk_snr_dB : REAL := 90.0;  -- < 96.284

  TYPE t_real_arr IS ARRAY (INTEGER RANGE <>) OF REAL; 
  TyPE t_slv_64_subbands_arr IS ARRAY (INTEGER RANGE <>) OF t_slv_64_arr(0 TO c_sdp_N_sub-1);

  -- . Subband weights for selected g_sp
  CONSTANT c_subband_weight_re   : INTEGER := INTEGER(g_subband_weight_gain * REAL(c_sdp_unit_sub_weight) * COS(g_subband_weight_phase * MATH_2_PI / 360.0));
  CONSTANT c_subband_weight_im   : INTEGER := INTEGER(g_subband_weight_gain * REAL(c_sdp_unit_sub_weight) * SIN(g_subband_weight_phase * MATH_2_PI / 360.0));

  -- MM
  -- . Address widths of a single MM instance
  CONSTANT c_addr_w_reg_diag_wg           : NATURAL := 2;
  -- . Address spans of a single MM instance
  CONSTANT c_mm_span_reg_diag_wg          : NATURAL := 2**c_addr_w_reg_diag_wg;

  CONSTANT c_mm_file_reg_bsn_source_v2    : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "REG_BSN_SOURCE_V2";
  CONSTANT c_mm_file_reg_bsn_scheduler_wg : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "REG_BSN_SCHEDULER";
  CONSTANT c_mm_file_reg_diag_wg          : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "REG_WG";
  CONSTANT c_mm_file_ram_equalizer_gains  : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "RAM_EQUALIZER_GAINS";
  CONSTANT c_mm_file_reg_dp_selector      : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "REG_DP_SELECTOR";
  CONSTANT c_mm_file_ram_st_sst           : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "RAM_ST_SST";

  -- Tb
  SIGNAL tb_end              : STD_LOGIC := '0';
  SIGNAL sim_done            : STD_LOGIC := '0';
  SIGNAL tb_clk              : STD_LOGIC := '0';  
  SIGNAL rd_data             : STD_LOGIC_VECTOR(c_32-1 DOWNTO 0) := (OTHERS => '0');

  SIGNAL pps_rst             : STD_LOGIC := '1';
  SIGNAL gen_pps             : STD_LOGIC := '0';

  -- WG
  SIGNAL current_bsn_wg      : STD_LOGIC_VECTOR(c_dp_stream_bsn_w-1 DOWNTO 0);

  -- FSUB
  -- . WPFB
  SIGNAL sp_subband_ssts_arr2            : t_slv_64_subbands_arr(c_sdp_N_pol-1 DOWNTO 0);   -- [pol][sub]
  SIGNAL sp_subband_sst_sum_arr          : t_real_arr(c_sdp_N_pol-1 DOWNTO 0) := (OTHERS => 0.0);
  SIGNAL sp_subband_sst                  : REAL := 0.0;
  SIGNAL sp_subband_sst_leakage          : REAL := 0.0;
  SIGNAL sp_subband_sst_leakage_snr_dB   : REAL := 0.0;  -- signal to noise (leakage) ratio
  SIGNAL sp_subband_sst_crosstalk        : REAL := 0.0;
  SIGNAL sp_subband_sst_crosstalk_snr_dB : REAL := 0.0;  -- signal to noise (crosstalk) ration

  SIGNAL exp_subband_ampl    : REAL := 0.0;
  SIGNAL exp_subband_power   : REAL := 0.0;
  SIGNAL exp_subband_sst     : REAL := 0.0;
  SIGNAL stat_data           : STD_LOGIC_VECTOR(c_longword_w-1 DOWNTO 0);

  -- . Selector
  SIGNAL sst_offload_weighted_subbands     : STD_LOGIC;

  -- . Subband equalizer
  SIGNAL sp_subband_weight_re    : INTEGER := 0;
  SIGNAL sp_subband_weight_im    : INTEGER := 0;
  SIGNAL sp_subband_weight_gain  : REAL := 0.0;
  SIGNAL sp_subband_weight_phase : REAL := 0.0;

  -- DUT
  SIGNAL ext_clk             : STD_LOGIC := '0';
  SIGNAL ext_pps             : STD_LOGIC := '0';

  SIGNAL WDI                 : STD_LOGIC;
  SIGNAL INTA                : STD_LOGIC;
  SIGNAL INTB                : STD_LOGIC;

  SIGNAL eth_clk             : STD_LOGIC := '0';
  SIGNAL eth_txp             : STD_LOGIC_VECTOR(c_unb2b_board_nof_eth-1 downto 0) := (OTHERS => '0');
  SIGNAL eth_rxp             : STD_LOGIC_VECTOR(c_unb2b_board_nof_eth-1 downto 0) := (OTHERS => '0');

  SIGNAL sens_scl            : STD_LOGIC;
  SIGNAL sens_sda            : STD_LOGIC;
  SIGNAL pmbus_scl           : STD_LOGIC;
  SIGNAL pmbus_sda           : STD_LOGIC;

  -- back transceivers
  SIGNAL JESD204B_SERIAL_DATA : STD_LOGIC_VECTOR(c_sdp_S_pn-1 downto 0);
  SIGNAL JESD204B_REFCLK      : STD_LOGIC := '1';

  -- jesd204b syncronization signals
  SIGNAL jesd204b_sysref     : STD_LOGIC;
  SIGNAL jesd204b_sync_n     : STD_LOGIC_VECTOR(c_sdp_N_sync_jesd-1 DOWNTO 0);

BEGIN

  ----------------------------------------------------------------------------
  -- System setup
  ----------------------------------------------------------------------------
  ext_clk <= NOT ext_clk AFTER c_ext_clk_period/2;  -- External clock (200 MHz)
  JESD204B_REFCLK <= NOT JESD204B_REFCLK AFTER c_bck_ref_clk_period/2;  -- JESD sample clock (200MHz) 

  INTA <= 'H';  -- pull up
  INTB <= 'H';  -- pull up

  sens_scl <= 'H';  -- pull up
  sens_sda <= 'H';  -- pull up
  pmbus_scl <= 'H';  -- pull up
  pmbus_sda <= 'H';  -- pull up

  ------------------------------------------------------------------------------
  -- External PPS
  ------------------------------------------------------------------------------  
  proc_common_gen_pulse(5, c_pps_period, '1', pps_rst, ext_clk, gen_pps);
  jesd204b_sysref <= gen_pps;
  ext_pps <= gen_pps;

  ------------------------------------------------------------------------------
  -- DUT
  ------------------------------------------------------------------------------
  u_lofar_unb2b_sdp_station_fsub : ENTITY lofar2_unb2b_sdp_station_lib.lofar2_unb2b_sdp_station
  GENERIC MAP (
    g_design_name            => "lofar2_unb2b_sdp_station_fsub",
    g_design_note            => "",
    g_sim                    => c_sim,
    g_sim_unb_nr             => c_unb_nr,
    g_sim_node_nr            => c_node_nr,
    g_wpfb                   => c_wpfb_sim,
    g_bsn_nof_clk_per_sync   => c_nof_clk_per_sync,
    g_scope_selected_subband => g_subband
  )
  PORT MAP (
    -- GENERAL
    CLK          => ext_clk,
    PPS          => ext_pps,
    WDI          => WDI,
    INTA         => INTA,
    INTB         => INTB,

    -- Others
    VERSION      => c_version,
    ID           => c_id,
    TESTIO       => open,

    -- I2C Interface to Sensors
    SENS_SC      => sens_scl,
    SENS_SD      => sens_sda,

    PMBUS_SC     => pmbus_scl,
    PMBUS_SD     => pmbus_sda,
    PMBUS_ALERT  => open,

    -- 1GbE Control Interface
    ETH_CLK      => eth_clk,
    ETH_SGIN     => eth_rxp,
    ETH_SGOUT    => eth_txp,

    -- LEDs
    QSFP_LED     => open,

    -- back transceivers
    JESD204B_SERIAL_DATA => JESD204B_SERIAL_DATA,
    JESD204B_REFCLK      => JESD204B_REFCLK,
  
    -- jesd204b syncronization signals
    JESD204B_SYSREF => jesd204b_sysref,
    JESD204B_SYNC_N => jesd204b_sync_n
  );

  -- Raw or weighted subbands
  exp_subband_ampl  <= sel_a_b(sst_offload_weighted_subbands = '0', c_exp_subband_ampl_raw, c_exp_subband_ampl_weighted);
  exp_subband_power <= sel_a_b(sst_offload_weighted_subbands = '0', c_exp_subband_power_raw, c_exp_subband_power_weighted);
  exp_subband_sst   <= sel_a_b(sst_offload_weighted_subbands = '0', c_exp_subband_sst_raw, c_exp_subband_sst_weighted);

  ------------------------------------------------------------------------------
  -- MM slave accesses via file IO
  ------------------------------------------------------------------------------
  tb_clk <= NOT tb_clk AFTER c_tb_clk_period/2;    -- Testbench MM clock
  
  p_mm_stimuli : PROCESS
    VARIABLE v_bsn                           : NATURAL;
    VARIABLE v_data_lo, v_data_hi            : STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
    VARIABLE v_stat_data                     : STD_LOGIC_VECTOR(c_longword_w-1 DOWNTO 0);
    VARIABLE v_len, v_span, v_offset, v_addr : NATURAL;  -- address ranges, indices
    VARIABLE v_W, v_P, v_U, v_S, v_B         : NATURAL;  -- array indicies
    VARIABLE v_re, v_im, v_weight            : INTEGER;
    VARIABLE v_power                         : REAL;
  BEGIN
    -- Wait for DUT power up after reset
    WAIT FOR 1 us;

    ----------------------------------------------------------------------------
    -- Enable BSN
    ----------------------------------------------------------------------------
    mmf_mm_bus_wr(c_mm_file_reg_bsn_source_v2, 2,         c_init_bsn, tb_clk);  -- Init BSN
    mmf_mm_bus_wr(c_mm_file_reg_bsn_source_v2, 3,                  0, tb_clk);  -- Write high part activates the init BSN
    mmf_mm_bus_wr(c_mm_file_reg_bsn_source_v2, 1, c_nof_clk_per_sync, tb_clk);  -- nof_block_per_sync
    mmf_mm_bus_wr(c_mm_file_reg_bsn_source_v2, 0,       16#00000003#, tb_clk);  -- Enable BSN at PPS

    -- Release PPS pulser, to get first PPS now and to start BSN source
    WAIT FOR 1 us;
    pps_rst <= '0';

    ----------------------------------------------------------------------------
    -- Read weighted subband selector
    ----------------------------------------------------------------------------
    mmf_mm_bus_rd(c_mm_file_reg_dp_selector, 0, rd_data, tb_clk);
    proc_common_wait_some_cycles(tb_clk, 1);
    sst_offload_weighted_subbands <= NOT rd_data(0);

    ----------------------------------------------------------------------------
    -- Enable and start WG
    ----------------------------------------------------------------------------
    --   0 : mode[7:0]           --> off=0, calc=1, repeat=2, single=3)
    --       nof_samples[31:16]  --> <= c_ram_wg_size=1024
    --   1 : phase[15:0]
    --   2 : freq[30:0]
    --   3 : ampl[16:0]
    v_offset := g_sp * c_mm_span_reg_diag_wg;
    mmf_mm_bus_wr(c_mm_file_reg_diag_wg, v_offset + 0, 1024*2**16 + 1, tb_clk);  -- nof_samples, mode calc
    mmf_mm_bus_wr(c_mm_file_reg_diag_wg, v_offset + 1, INTEGER(c_wg_phase * c_diag_wg_phase_unit), tb_clk);  -- phase offset in degrees
    mmf_mm_bus_wr(c_mm_file_reg_diag_wg, v_offset + 2, INTEGER(REAL(g_subband) * c_sdp_wg_subband_freq_unit), tb_clk);  -- freq
    mmf_mm_bus_wr(c_mm_file_reg_diag_wg, v_offset + 3, INTEGER(REAL(c_wg_ampl) * c_sdp_wg_ampl_lsb), tb_clk);  -- ampl

    -- Read current BSN
    mmf_mm_bus_rd(c_mm_file_reg_bsn_scheduler_wg, 0, current_bsn_wg(31 DOWNTO  0), tb_clk);
    mmf_mm_bus_rd(c_mm_file_reg_bsn_scheduler_wg, 1, current_bsn_wg(63 DOWNTO 32), tb_clk);
    proc_common_wait_some_cycles(tb_clk, 1);
    
    -- Write scheduler BSN to trigger start of WG at next block
    v_bsn := TO_UINT(current_bsn_wg) + 2;
    ASSERT v_bsn <= c_bsn_start_wg REPORT "Too late to start WG: " & int_to_str(v_bsn) & " > " & int_to_str(c_bsn_start_wg) SEVERITY ERROR;
    mmf_mm_bus_wr(c_mm_file_reg_bsn_scheduler_wg, 0, c_bsn_start_wg, tb_clk);  -- first write low then high part
    mmf_mm_bus_wr(c_mm_file_reg_bsn_scheduler_wg, 1,              0, tb_clk);  -- assume v_bsn < 2**31-1

    ----------------------------------------------------------------------------
    -- Write subband weight for selected g_sp and g_subband
    ----------------------------------------------------------------------------
    -- . MM format: (cint16)RAM_EQUALIZER_GAINS[S_pn/Q_fft]_[Q_fft][N_sub] = [S_pn][N_sub]
    v_addr := g_sp * c_sdp_N_sub + g_subband;
    -- . read
    mmf_mm_bus_rd(c_mm_file_ram_equalizer_gains, v_addr, rd_data, tb_clk);
    v_re := unpack_complex_re(rd_data, c_sdp_W_sub_weight);
    v_im := unpack_complex_im(rd_data, c_sdp_W_sub_weight);
    sp_subband_weight_re <= v_re;
    sp_subband_weight_im <= v_im;
    sp_subband_weight_gain <= SQRT(REAL(v_re)**2.0 + REAL(v_im)**2.0) / REAL(c_sdp_unit_sub_weight);
    sp_subband_weight_phase <= atan2(Y => REAL(v_im), X => REAL(v_re)) * 360.0 / MATH_2_PI;
    -- . write
    v_weight := pack_complex(re => c_subband_weight_re, im => c_subband_weight_im, w => c_sdp_W_sub_weight);  -- c_sdp_W_sub_weight = 16 bit
    mmf_mm_bus_wr(c_mm_file_ram_equalizer_gains, v_addr, v_weight, tb_clk);
    -- . read back
    mmf_mm_bus_rd(c_mm_file_ram_equalizer_gains, v_addr, rd_data, tb_clk);
    v_re := unpack_complex_re(rd_data, c_sdp_W_sub_weight);
    v_im := unpack_complex_im(rd_data, c_sdp_W_sub_weight);
    sp_subband_weight_re <= v_re;
    sp_subband_weight_im <= v_im;
    sp_subband_weight_gain <= SQRT(REAL(v_re)**2.0 + REAL(v_im)**2.0) / REAL(c_sdp_unit_sub_weight);
    sp_subband_weight_phase <= atan2(Y => REAL(v_im), X => REAL(v_re)) * 360.0 / MATH_2_PI;

    ----------------------------------------------------------------------------
    -- Wait for enough WG data and start of sync interval
    ----------------------------------------------------------------------------
    mmf_mm_wait_until_value(c_mm_file_reg_bsn_scheduler_wg, 0,                                 -- read BSN low
                            "UNSIGNED", rd_data, ">=", c_init_bsn + c_nof_block_per_sync * 3,  -- this is the wait until condition
                            c_sdp_T_sub, tb_clk);

    ---------------------------------------------------------------------------
    -- Read subband statistics
    ---------------------------------------------------------------------------
    -- . the subband statistics are c_stat_data_sz = 2 word power values.
    -- . there are c_sdp_S_pn = 12 signal inputs A, B, C, D, E, F, G, H, I, J, K, L
    -- . there are c_sdp_N_sub = 512 subbands per signal input (SI, = signal path, SP)
    -- . one complex WPFB can process two real inputs A, B, so there are c_sdp_P_pfb = 6 WPFB units,
    --   but only read for the 1 WPFB unit of the selected g_sp, to save sim time
    -- . the outputs for A, B are time multiplexed, c_sdp_Q_fft = 2, assume that they
    --   correspond to the c_sdp_N_pol = 2 signal polarizations
    -- . the subbands are output alternately so A0 B0 A1 B1 ... A511 B511 for input A, B
    -- . the subband statistics multiple WPFB units appear in order in the ram_st_sst address map
    -- . the subband statistics are stored first lo word 0 then hi word 1
    v_len := c_sdp_N_sub * c_sdp_N_pol * c_stat_data_sz;  -- 2048 = 512 * 2 * 64/32
    v_span := true_log_pow2(v_len);                       -- = 2048
    FOR I IN 0 TO v_len-1 LOOP
      v_W := I MOD c_stat_data_sz;                   -- 0, 1 per statistics word, word index
      v_P := (I / c_stat_data_sz) MOD c_sdp_N_pol;   -- 0, 1 per SP pol, polarization index
      v_B := I / (c_sdp_N_pol * c_stat_data_sz);     -- subband index, range(N_sub = 512) per dual pol
      v_addr := I + c_pfb_index * v_span;            -- MM address for WPFB unit of selected g_sp
      -- Only read SST for g_subband for dual pol SP, to save sim time
      IF g_read_all_SST = TRUE OR v_B = g_subband THEN
        IF v_W = 0 THEN
          -- low part
          mmf_mm_bus_rd(c_mm_file_ram_st_sst, v_addr, rd_data, tb_clk);
          v_data_lo := rd_data;
        ELSE
          -- high part
          mmf_mm_bus_rd(c_mm_file_ram_st_sst, v_addr, rd_data, tb_clk);
          v_data_hi := rd_data;
          v_stat_data := v_data_hi & v_data_lo;

          sp_subband_ssts_arr2(v_P)(v_B) <= v_stat_data;
          stat_data <= v_stat_data;  -- for time series view in Wave window

          -- sum of all subband powers per pol
          sp_subband_sst_sum_arr(v_P) <= sp_subband_sst_sum_arr(v_P) + TO_UREAL(v_stat_data);
        END IF;
      END IF;
    END LOOP;
    proc_common_wait_some_cycles(tb_clk, 1);

    -- Subband power of g_subband in g_sp
    -- . For the selected g_subband in g_sp the sp_subband_sst will be close
    --   to sp_subband_sst_sum_arr(c_pol_index), because the input is a
    --   sinus, so most power will be in 1 subband.
    sp_subband_sst <= TO_UREAL(sp_subband_ssts_arr2(c_pol_index)(g_subband));
    proc_common_wait_some_cycles(tb_clk, 1);

    -- The sp_subband_sst_leakage shows how much power from the input sinus at a specific
    -- subband has leaked into the N_sub-1 = 511 other subbands. The power ratio yields an
    -- indication of the SNR, although that also depends on the SNR of the WG sinus.
    v_power := sp_subband_sst_sum_arr(c_pol_index) - sp_subband_sst;
    sp_subband_sst_leakage <= v_power;
    IF sp_subband_sst > c_eps AND v_power > c_eps THEN
      sp_subband_sst_leakage_snr_dB <= 10.0 * LOG10(sp_subband_sst / v_power);
    ELSIF g_read_all_SST THEN
      REPORT "Wrong, zero leakage is unexpected for SP-" & NATURAL'IMAGE(g_sp) SEVERITY ERROR;
    END IF;

    -- The sp_subband_sst_crosstalk shows how much power from one WPFB input cross talks
    -- into the other output, due to quantization cross talk in the complex FFT. The power
    -- ration indicates the suppression, provided that the other input was zero.
    v_power := sp_subband_sst_sum_arr(not_int(c_pol_index));
    sp_subband_sst_crosstalk <= v_power;
    IF sp_subband_sst > c_eps AND v_power > c_eps THEN
      sp_subband_sst_crosstalk_snr_dB <= 10.0 * LOG10(sp_subband_sst / v_power);
    ELSIF g_read_all_SST THEN
      REPORT "Zero crosstalk for SP-" & NATURAL'IMAGE(g_sp) SEVERITY NOTE;
    END IF;

    proc_common_wait_some_cycles(tb_clk, 10);

    ---------------------------------------------------------------------------
    -- Log WG, subband statistics
    ---------------------------------------------------------------------------
    print_str("");
    print_str("WG:");
    print_str(". c_wg_ampl                            = " & int_to_str(c_wg_ampl));
    print_str(". c_exp_sp_power                       = " & real_to_str(c_exp_sp_power, 20, 1));
    print_str(". c_exp_sp_ast                         = " & real_to_str(c_exp_sp_ast, 20, 1));

    print_str("");
    print_str("Subband selector:");
    print_str(". sst_offload_weighted_subbands        = " & sl_to_str(sst_offload_weighted_subbands));

    print_str("");
    print_str("Subband weight:");
    print_str(". sp_subband_weight_gain               = " & real_to_str(sp_subband_weight_gain, 20, 6));
    print_str(". sp_subband_weight_phase              = " & real_to_str(sp_subband_weight_phase, 20, 6));

    print_str("");
    print_str("SST results:");
    print_str(". exp_subband_ampl                     = " & int_to_str(NATURAL(exp_subband_ampl)));
    print_str(". exp_subband_power                    = " & real_to_str(exp_subband_power, 20, 1));
    print_str(". exp_subband_sst                      = " & real_to_str(exp_subband_sst, 20, 1));
    print_str("");
    print_str(". sp_subband_sst                       = " & real_to_str(sp_subband_sst, 20, 1));
    print_str(". sp_subband_sst / exp_subband_sst     = " & real_to_str(sp_subband_sst / exp_subband_sst, 20, 6));

    IF g_read_all_SST THEN
      -- Log WPFB details, these are allready verified in tb of wpfb_unit_dev.vhd, so here
      -- quality indicators like leakage and crosstalk are also reported out of interest.
      print_str("");
      print_str("SST quality indicators");
      print_str(". sp_subband_sst_leakage               = " & real_to_str(sp_subband_sst_leakage, 20, 0));
      print_str(". sp_subband_sst_leakage_snr_dB        = " & real_to_str(sp_subband_sst_leakage_snr_dB, 20, 3));
      print_str(". sp_subband_sst_crosstalk             = " & real_to_str(sp_subband_sst_crosstalk, 20, 0));
      print_str(". sp_subband_sst_crosstalk_snr_db      = " & real_to_str(sp_subband_sst_crosstalk_snr_db, 20, 3));
    END IF;

    ---------------------------------------------------------------------------
    -- Verify SST
    ---------------------------------------------------------------------------
    -- verify expected subband power based on WG power
    ASSERT sp_subband_sst > c_lo_factor * exp_subband_sst REPORT "Wrong subband power for SP " & NATURAL'IMAGE(g_sp) SEVERITY ERROR;
    ASSERT sp_subband_sst < c_hi_factor * exp_subband_sst REPORT "Wrong subband power for SP " & NATURAL'IMAGE(g_sp) SEVERITY ERROR;

    IF g_read_all_SST THEN
      -- Verify expected SNR quality measures
      ASSERT sp_subband_sst_leakage = 0.0 OR sp_subband_sst_leakage_snr_dB > c_exp_subband_sst_leakage_snr_dB REPORT "Wrong too much leakage for SP " & NATURAL'IMAGE(g_sp) SEVERITY ERROR;
      ASSERT sp_subband_sst_crosstalk = 0.0 OR sp_subband_sst_crosstalk_snr_dB > c_exp_subband_sst_crosstalk_snr_dB REPORT "Wrong too much crosstalk for SP " & NATURAL'IMAGE(g_sp) SEVERITY ERROR;
    END IF;

    ---------------------------------------------------------------------------
    -- End Simulation 
    ---------------------------------------------------------------------------  
    sim_done <= '1';
    proc_common_wait_some_cycles(ext_clk, 100);
    proc_common_stop_simulation(TRUE, ext_clk, sim_done, tb_end);
    WAIT;
  END PROCESS;

END tb;
