-------------------------------------------------------------------------------
--
-- Copyright 2020
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--
-- Author: R. van der Walle
-- Purpose: Self-checking testbench for simulating lofar2_unb2b_sdp_station_xsub_one using WG data.
--
-- Description:
--   MM control actions:
--
--   1) Enable calc mode for WG via reg_diag_wg with:
--        freq = 19.921875MHz = subband index 102
--        ampl = 0.5 * 2**13, full scale amplitude is 2**13
--   
--   2) Read current BSN from reg_bsn_scheduler_wg and write reg_bsn_scheduler_wg 
--      to trigger start of WG at BSN.
--     
--   3) Read crosslets statistics (XST) via ram_st_xsq and verify that the values
--      are as expected. This is done by comparing the values in the outgoing square
--      correlation matrix.
--         
--
-- Usage:
--   > as 7    # default
--   > as 12   # for detailed debugging
--   > run -a  
--   Takes about 40 m
--
-------------------------------------------------------------------------------
LIBRARY IEEE, common_lib, unb2b_board_lib, i2c_lib, mm_lib, dp_lib, diag_lib, lofar2_sdp_lib, wpfb_lib, lofar2_unb2b_sdp_station_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE IEEE.MATH_REAL.ALL;
USE common_lib.common_pkg.ALL;
USE unb2b_board_lib.unb2b_board_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE common_lib.common_str_pkg.ALL;
USE mm_lib.mm_file_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE mm_lib.mm_file_unb_pkg.ALL;
USE diag_lib.diag_pkg.ALL;
USE wpfb_lib.wpfb_pkg.ALL;
USE lofar2_sdp_lib.sdp_pkg.ALL;

ENTITY tb_lofar2_unb2b_sdp_station_xsub_one IS
END tb_lofar2_unb2b_sdp_station_xsub_one;

ARCHITECTURE tb OF tb_lofar2_unb2b_sdp_station_xsub_one IS

  CONSTANT c_sim             : BOOLEAN := TRUE;
  CONSTANT c_unb_nr          : NATURAL := 0; -- UniBoard 0
  CONSTANT c_node_nr         : NATURAL := 0; 
  CONSTANT c_id              : STD_LOGIC_VECTOR(7 DOWNTO 0) := "00000000";
  CONSTANT c_version         : STD_LOGIC_VECTOR(1 DOWNTO 0) := "00";
  CONSTANT c_fw_version      : t_unb2b_board_fw_version := (1, 0);

  CONSTANT c_eth_clk_period      : TIME := 8 ns;  -- 125 MHz XO on UniBoard
  CONSTANT c_ext_clk_period      : TIME := 5 ns;
  CONSTANT c_bck_ref_clk_period  : TIME := 5 ns;

  CONSTANT c_tb_clk_period       : TIME := 100 ps; -- use fast tb_clk to speed up M&C

  CONSTANT c_nof_block_per_sync  : NATURAL := 24;
  CONSTANT c_nof_clk_per_sync    : NATURAL := c_nof_block_per_sync*c_sdp_N_fft; 
  CONSTANT c_pps_period          : NATURAL := c_nof_clk_per_sync;
  CONSTANT c_wpfb_sim            : t_wpfb := func_wpfb_set_nof_block_per_sync(c_sdp_wpfb_subbands, c_nof_block_per_sync);
  CONSTANT c_ctrl_interval_size  : NATURAL := c_nof_clk_per_sync;

  CONSTANT c_percentage          : REAL := 0.05;  -- percentage that actual value may differ from expected value
  CONSTANT c_lo_factor           : REAL := 1.0 - c_percentage;  -- lower boundary  
  CONSTANT c_hi_factor           : REAL := 1.0 + c_percentage;  -- higher boundary

  -- WG
  CONSTANT c_bsn_start_wg         : NATURAL := 2;  -- start WG at this BSN to instead of some BSN, to avoid mismatches in exact expected data values
  CONSTANT c_ampl_sp_0            : NATURAL := c_sdp_FS_adc/2;  -- = 0.5 * FS, so in number of lsb
  CONSTANT c_wg_freq_offset       : REAL := 0.0/11.0; -- in freq_unit
  CONSTANT c_subband_sp_0         : REAL := 102.0;  -- Select subband at index 102 = 102/1024 * 200MHz = 19.921875 MHz 
  CONSTANT c_exp_wg_power_sp_0    : REAL := REAL(c_ampl_sp_0**2)/2.0 * REAL(c_nof_clk_per_sync);

  -- WPFB
  CONSTANT c_nof_pfb                        : NATURAL := 1; -- Verifying 1 of c_sdp_P_pfb = 6 pfb to speed up simulation.
  CONSTANT c_wb_leakage_bin                 : NATURAL := c_wpfb_sim.nof_points / c_wpfb_sim.wb_factor;   -- = 256, leakage will occur in this bin if FIR wb_factor is reversed 
  CONSTANT c_exp_sp_subband_power_ratio     : REAL := 1.0/8.0;   -- depends on internal WPFB quantization and FIR coefficients
  CONSTANT c_exp_sp_subband_power_sum_ratio : REAL := c_exp_sp_subband_power_ratio;   -- because all sinus power is expected in one subband
  CONSTANT c_exp_subband_power_sp_0         : REAL := c_exp_wg_power_sp_0 * c_exp_sp_subband_power_ratio;

  TYPE t_real_arr IS ARRAY (INTEGER RANGE <>) OF REAL; 

  -- MM  
  CONSTANT c_mm_file_reg_bsn_source_v2           : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "REG_BSN_SOURCE_V2";
  CONSTANT c_mm_file_reg_bsn_scheduler_wg        : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "REG_BSN_SCHEDULER";
  CONSTANT c_mm_file_reg_diag_wg                 : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "REG_WG";
  CONSTANT c_mm_file_ram_st_sst                  : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "RAM_ST_SST";
  CONSTANT c_mm_file_reg_crosslets_info          : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "REG_CROSSLETS_INFO";
  CONSTANT c_mm_file_reg_bsn_sync_scheduler_xsub : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "REG_BSN_SYNC_SCHEDULER_XSUB";
  CONSTANT c_mm_file_ram_st_xsq                  : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "RAM_ST_XSQ";

  -- Tb
  SIGNAL tb_end              : STD_LOGIC := '0';
  SIGNAL sim_done            : STD_LOGIC := '0';
  SIGNAL tb_clk              : STD_LOGIC := '0';  
  SIGNAL rd_data             : STD_LOGIC_VECTOR(c_32-1 DOWNTO 0) := (OTHERS => '0');

  -- WG
  SIGNAL current_bsn_wg          : STD_LOGIC_VECTOR(c_dp_stream_bsn_w-1 DOWNTO 0);

  -- WPFB
  SIGNAL xsub_stats_arr         : t_slv_64_arr(0 TO c_nof_complex * c_sdp_X_sq -1);

  
  -- DUT
  SIGNAL ext_clk             : STD_LOGIC := '0';
  SIGNAL pps                 : STD_LOGIC := '0';
  SIGNAL ext_pps             : STD_LOGIC := '0'; 
  SIGNAL pps_rst             : STD_LOGIC := '1';

  SIGNAL WDI                 : STD_LOGIC;
  SIGNAL INTA                : STD_LOGIC;
  SIGNAL INTB                : STD_LOGIC;

  SIGNAL eth_clk             : STD_LOGIC := '0';
  SIGNAL eth_txp             : STD_LOGIC_VECTOR(c_unb2b_board_nof_eth-1 downto 0) := (OTHERS => '0');
  SIGNAL eth_rxp             : STD_LOGIC_VECTOR(c_unb2b_board_nof_eth-1 downto 0) := (OTHERS => '0');

  SIGNAL sens_scl            : STD_LOGIC;
  SIGNAL sens_sda            : STD_LOGIC;
  SIGNAL pmbus_scl           : STD_LOGIC;
  SIGNAL pmbus_sda           : STD_LOGIC;

  -- back transceivers
  SIGNAL JESD204B_SERIAL_DATA : STD_LOGIC_VECTOR(c_sdp_S_pn-1 downto 0);
  SIGNAL JESD204B_REFCLK      : STD_LOGIC := '1';

  -- jesd204b syncronization signals
  SIGNAL jesd204b_sysref     : STD_LOGIC;
  SIGNAL jesd204b_sync_n     : STD_LOGIC_VECTOR(c_sdp_N_sync_jesd-1 DOWNTO 0);

BEGIN

  ----------------------------------------------------------------------------
  -- System setup
  ----------------------------------------------------------------------------
  ext_clk <= NOT ext_clk AFTER c_ext_clk_period/2;  -- External clock (200 MHz)
  JESD204B_REFCLK <= NOT JESD204B_REFCLK AFTER c_bck_ref_clk_period/2;  -- JESD sample clock (200MHz) 

  INTA <= 'H';  -- pull up
  INTB <= 'H';  -- pull up

  sens_scl <= 'H';  -- pull up
  sens_sda <= 'H';  -- pull up
  pmbus_scl <= 'H';  -- pull up
  pmbus_sda <= 'H';  -- pull up

  ------------------------------------------------------------------------------
  -- External PPS
  ------------------------------------------------------------------------------  
  proc_common_gen_pulse(5, c_pps_period, '1', pps_rst, ext_clk, pps);
  jesd204b_sysref <= pps;
  ext_pps <= pps;

  ------------------------------------------------------------------------------
  -- DUT
  ------------------------------------------------------------------------------
  u_lofar_unb2b_sdp_station_xsub_one : ENTITY lofar2_unb2b_sdp_station_lib.lofar2_unb2b_sdp_station
  GENERIC MAP (
    g_design_name            => "lofar2_unb2b_sdp_station_xsub_one",
    g_design_note            => "",
    g_sim                    => c_sim,
    g_sim_unb_nr             => c_unb_nr,
    g_sim_node_nr            => c_node_nr,
    g_wpfb                   => c_wpfb_sim,
    g_bsn_nof_clk_per_sync   => c_nof_clk_per_sync,
    g_scope_selected_subband => NATURAL(c_subband_sp_0)
  )
  PORT MAP (
    -- GENERAL
    CLK          => ext_clk,
    PPS          => pps,
    WDI          => WDI,
    INTA         => INTA,
    INTB         => INTB,

    -- Others
    VERSION      => c_version,
    ID           => c_id,
    TESTIO       => open,

    -- I2C Interface to Sensors
    SENS_SC      => sens_scl,
    SENS_SD      => sens_sda,

    PMBUS_SC     => pmbus_scl,
    PMBUS_SD     => pmbus_sda,
    PMBUS_ALERT  => open,

    -- 1GbE Control Interface
    ETH_CLK      => eth_clk,
    ETH_SGIN     => eth_rxp,
    ETH_SGOUT    => eth_txp,

    -- LEDs
    QSFP_LED     => open,

    -- back transceivers
    JESD204B_SERIAL_DATA => JESD204B_SERIAL_DATA,
    JESD204B_REFCLK      => JESD204B_REFCLK,
  
    -- jesd204b syncronization signals
    JESD204B_SYSREF => jesd204b_sysref,
    JESD204B_SYNC_N => jesd204b_sync_n
  );

  ------------------------------------------------------------------------------
  -- MM slave accesses via file IO
  ------------------------------------------------------------------------------
  tb_clk  <= NOT tb_clk AFTER c_tb_clk_period/2;    -- Testbench MM clock
  
  p_mm_stimuli : PROCESS
    VARIABLE v_bsn                   : NATURAL;
    VARIABLE v_sp_subband_power      : REAL;
    VARIABLE v_W, v_C, v_A, v_X, v_B, v_A_even, v_B_even : NATURAL;  -- array indicies
  BEGIN
    -- Wait for DUT power up after reset
    WAIT FOR 1 us;
 
    ----------------------------------------------------------------------------
    -- Enable BSN
    ----------------------------------------------------------------------------
    mmf_mm_bus_wr(c_mm_file_reg_bsn_source_v2, 3,                    0, tb_clk);
    mmf_mm_bus_wr(c_mm_file_reg_bsn_source_v2, 2,                    0, tb_clk);  -- Init BSN = 0
    mmf_mm_bus_wr(c_mm_file_reg_bsn_source_v2, 1,   c_nof_clk_per_sync, tb_clk);  -- nof_block_per_sync
    mmf_mm_bus_wr(c_mm_file_reg_bsn_source_v2, 0,         16#00000003#, tb_clk);  -- Enable BSN at PPS

    -- Release PPS pulser, to get first PPS now and to start BSN source
    WAIT FOR 1 us;
    pps_rst <= '0';

    ----------------------------------------------------------------------------
    -- Crosslets Info
    ----------------------------------------------------------------------------  
    mmf_mm_bus_wr(c_mm_file_reg_crosslets_info, 0,  INTEGER(c_subband_sp_0), tb_clk); -- offset
    mmf_mm_bus_wr(c_mm_file_reg_crosslets_info, 15, 0                      , tb_clk); -- stepsize

    ----------------------------------------------------------------------------
    -- Enable WG
    ----------------------------------------------------------------------------
    --   0 : mode[7:0]           --> off=0, calc=1, repeat=2, single=3)
    --       nof_samples[31:16]  --> <= c_ram_wg_size=1024
    --   1 : phase[15:0]
    --   2 : freq[30:0]
    --   3 : ampl[16:0]
    FOR I IN 0 TO c_sdp_S_pn-1 LOOP
      mmf_mm_bus_wr(c_mm_file_reg_diag_wg, I*4 + 0, 1024*2**16 + 1, tb_clk);  -- nof_samples, mode calc
      mmf_mm_bus_wr(c_mm_file_reg_diag_wg, I*4 + 1, INTEGER(  0.0 * c_diag_wg_phase_unit), tb_clk);  -- phase offset in degrees
      mmf_mm_bus_wr(c_mm_file_reg_diag_wg, I*4 + 2, INTEGER((c_subband_sp_0 + c_wg_freq_offset) * c_sdp_wg_subband_freq_unit), tb_clk);  -- freq
      mmf_mm_bus_wr(c_mm_file_reg_diag_wg, I*4 + 3, INTEGER(REAL(c_ampl_sp_0) * c_sdp_wg_ampl_lsb), tb_clk);  -- ampl
    END LOOP;

    -- Read current BSN
    mmf_mm_bus_rd(c_mm_file_reg_bsn_scheduler_wg, 0, current_bsn_wg(31 DOWNTO  0), tb_clk);
    mmf_mm_bus_rd(c_mm_file_reg_bsn_scheduler_wg, 1, current_bsn_wg(63 DOWNTO 32), tb_clk);
    proc_common_wait_some_cycles(tb_clk, 1);
    
    -- Write scheduler BSN to trigger start of WG at next block
    v_bsn := TO_UINT(current_bsn_wg) + 2;
    ASSERT v_bsn <= c_bsn_start_wg REPORT "Too late to start WG: " & int_to_str(v_bsn) & " > " & int_to_str(c_bsn_start_wg) SEVERITY ERROR;
    mmf_mm_bus_wr(c_mm_file_reg_bsn_scheduler_wg, 0, c_bsn_start_wg, tb_clk);  -- first write low then high part
    mmf_mm_bus_wr(c_mm_file_reg_bsn_scheduler_wg, 1,              0, tb_clk);  -- assume v_bsn < 2**31-1

    -- bsn_scheduler_xsub
    mmf_mm_bus_wr(c_mm_file_reg_bsn_sync_scheduler_xsub, 1, c_ctrl_interval_size, tb_clk);  -- Interval size
    mmf_mm_bus_wr(c_mm_file_reg_bsn_sync_scheduler_xsub, 2,       c_bsn_start_wg, tb_clk);  -- first write low then high part
    mmf_mm_bus_wr(c_mm_file_reg_bsn_sync_scheduler_xsub, 3,                    0, tb_clk);  -- assume v_bsn < 2**31-1
    mmf_mm_bus_wr(c_mm_file_reg_bsn_sync_scheduler_xsub, 0,                    1, tb_clk);  -- enable

    -- Wait for enough WG data and start of sync interval
    mmf_mm_wait_until_value(c_mm_file_reg_bsn_scheduler_wg, 0,                    -- read BSN low
                            "UNSIGNED", rd_data, ">=", c_nof_block_per_sync * 2,  -- this is the wait until condition
                            c_sdp_T_sub, tb_clk);

    ---------------------------------------------------------------------------
    -- Read crosslet statistics
    ---------------------------------------------------------------------------
    FOR I IN 0 TO c_nof_complex * c_sdp_X_sq * (c_longword_sz/c_word_sz) -1 LOOP
      v_W := I MOD 2;
      v_B := I / 2;
      IF v_W=0 THEN
        -- low part
        mmf_mm_bus_rd(c_mm_file_ram_st_xsq, I, rd_data, tb_clk);
        xsub_stats_arr(v_B)(31 DOWNTO 0) <= rd_data;
      ELSE      
        -- high part
        mmf_mm_bus_rd(c_mm_file_ram_st_xsq, I, rd_data, tb_clk);
        xsub_stats_arr(v_B)(63 DOWNTO 32) <= rd_data;
      END IF;
    END LOOP;
    proc_common_wait_some_cycles(tb_clk, 1);

    ---------------------------------------------------------------------------
    -- Verify crosslet statistics
    --------------------------------------------------------------------------- 
    -- With all WGs having the same input all crosslets should be identical. Due to quantization cross talk
    -- between the two real inputs of the filterbank the two signals in the output pairs per P_pfb differ 
    -- slightly, therefore 3 slightly different correlation values are expected. 1 for each correlation 
    -- between even indexed signals, 1 for odd indexed signals and 1 for correlations between even and odd 
    -- indexed signals. This is verified by checking if these values are the same.
    FOR I IN 0 TO c_nof_complex * c_sdp_X_sq -1 LOOP
      v_C := I MOD 2;
      v_X := I /c_nof_complex;
      v_A := v_X MOD c_sdp_S_pn;
      v_B := v_X / c_sdp_S_pn;
      v_A_even := v_A MOD 2;
      v_B_even := v_B MOD 2;
      
      -- Check real values of even indices
      IF v_C=0 AND v_A_even=0 AND v_B_even=0 THEN 
        ASSERT SIGNED(xsub_stats_arr(I)) = SIGNED(xsub_stats_arr(0)) REPORT "correlation between even indexed signals (re) is wrong at I = " & int_to_str(I) SEVERITY ERROR; END IF;

      -- Check real values of odd indices
      IF v_C=0 AND v_A_even=1 AND v_B_even=1 THEN 
        ASSERT SIGNED(xsub_stats_arr(I)) = SIGNED(xsub_stats_arr((c_sdp_S_pn + 1) * c_nof_complex)) REPORT "correlation between odd indexed signals (re) is wrong at I = " & int_to_str(I)  SEVERITY ERROR; END IF;

      -- Check real values of even correlated with odd indices
      IF v_C=0 AND (v_A_even=0 XOR v_B_even=0) THEN 
        ASSERT SIGNED(xsub_stats_arr(I)) = SIGNED(xsub_stats_arr(1 * c_nof_complex)) REPORT "correlation between even indexed signals (re) is wrong at I = " & int_to_str(I) SEVERITY ERROR; END IF;
     
      -- Using absolute value of the imaginary part (force positive) as the sign can be opposite when comparing A to B vs B to A.
      -- Check im values of even indices
      IF v_C=1 AND v_A_even=0 AND v_B_even=0 THEN 
        ASSERT ABS(SIGNED(xsub_stats_arr(I))) = ABS(SIGNED(xsub_stats_arr(1))) REPORT "correlation between even indexed signals (im) is wrong at I = " & int_to_str(I) SEVERITY ERROR; END IF;

      -- Check im values of odd indices
      IF v_C=1 AND v_A_even=1 AND v_B_even=1 THEN 
        ASSERT ABS(SIGNED(xsub_stats_arr(I))) = ABS(SIGNED(xsub_stats_arr((c_sdp_S_pn + 1) * c_nof_complex + 1))) REPORT "correlation between odd indexed signals (im) is wrong at I = " & int_to_str(I)  SEVERITY ERROR; END IF;

      -- Check im values of even correlated with odd indices
      IF v_C=1 AND (v_A_even=0 XOR v_B_even=0) THEN 
        ASSERT ABS(SIGNED(xsub_stats_arr(I))) = ABS(SIGNED(xsub_stats_arr(1 * c_nof_complex + 1))) REPORT "correlation between even/odd indexed signals (im) is wrong at I = " & int_to_str(I) SEVERITY ERROR; END IF;

      -- Check if values are > 0
      IF v_C=0 THEN ASSERT (SIGNED(xsub_stats_arr(I)) > TO_SIGNED(0, c_longword_w)) REPORT "correlation is 0 which is unexpected! at I = " & int_to_str(I) SEVERITY ERROR; END IF;
    END LOOP; 
 
    ---------------------------------------------------------------------------
    -- End Simulation 
    ---------------------------------------------------------------------------  
    sim_done <= '1';
    proc_common_wait_some_cycles(ext_clk, 100);
    proc_common_stop_simulation(TRUE, ext_clk, sim_done, tb_end);
    WAIT;
  END PROCESS;

END tb;
