-------------------------------------------------------------------------------
--
-- Copyright (C) 2015
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, unb2c_board_lib, technology_lib, diag_lib, dp_lib, lofar2_ddrctrl_lib, tech_ddr_lib, lofar2_sdp_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE technology_lib.technology_pkg.ALL;
USE unb2c_board_lib.unb2c_board_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE tech_ddr_lib.tech_ddr_pkg.ALL;
USE lofar2_sdp_lib.sdp_pkg.ALL;
USE diag_lib.diag_pkg.ALL;



ENTITY lofar2_unb2c_ddrctrl IS
  GENERIC (
    g_design_name       : STRING  := "lofar2_unb2c_ddrctrl";
    g_design_note       : STRING  := "UNUSED";
    g_technology        : NATURAL := c_tech_arria10_e2sg;
    g_sim               : BOOLEAN := FALSE; --Overridden by TB
    g_sim_unb_nr        : NATURAL := 0;
    g_sim_node_nr       : NATURAL := 0;
    g_stamp_date        : NATURAL := 0;  -- Date (YYYYMMDD) -- set by QSF
    g_stamp_time        : NATURAL := 0;  -- Time (HHMMSS)   -- set by QSF
    g_revision_id       : STRING  := "";  -- revision id    -- set by QSF
    g_factory_image     : BOOLEAN := TRUE;
    g_protect_addr_range: BOOLEAN := FALSE
  );
  PORT (
    -- GENERAL
    CLK          : IN    STD_LOGIC; -- System Clock
    tb_clk       : IN    STD_LOGIC;
    PPS          : IN    STD_LOGIC; -- System Sync
    WDI          : OUT   STD_LOGIC; -- Watchdog Clear
    INTA         : INOUT STD_LOGIC; -- FPGA interconnect line
    INTB         : INOUT STD_LOGIC; -- FPGA interconnect line

    -- Others
    VERSION      : IN    STD_LOGIC_VECTOR(c_unb2c_board_aux.version_w-1 DOWNTO 0);
    ID           : IN    STD_LOGIC_VECTOR(c_unb2c_board_aux.id_w-1 DOWNTO 0);
    TESTIO       : INOUT STD_LOGIC_VECTOR(c_unb2c_board_aux.testio_w-1 DOWNTO 0);
    
    -- 1GbE Control Interface
    ETH_CLK      : IN    STD_LOGIC_VECTOR(c_unb2c_board_nof_eth-1 DOWNTO 0);
    ETH_SGIN     : IN    STD_LOGIC_VECTOR(c_unb2c_board_nof_eth-1 DOWNTO 0);
    ETH_SGOUT    : OUT   STD_LOGIC_VECTOR(c_unb2c_board_nof_eth-1 DOWNTO 0);

    QSFP_LED     : OUT   STD_LOGIC_VECTOR(c_unb2c_board_tr_qsfp_nof_leds-1 DOWNTO 0)
  );
END lofar2_unb2c_ddrctrl;


ARCHITECTURE str OF lofar2_unb2c_ddrctrl IS

  -- Firmware version x.y
  -- If x >= 2, rom_info starts on 0x10000 and max size = 0x8192 words                                                                                
  CONSTANT c_fw_version             : t_unb2c_board_fw_version := (2, 0);
  CONSTANT c_mm_clk_freq            : NATURAL := c_unb2c_board_mm_clk_freq_125M;
  CONSTANT c_bs_block_size          : NATURAL := 1024;
  CONSTANT c_bs_bsn_w               : NATURAL := 64;
  CONSTANT c_nof_streams            : NATURAL := 12;
  CONSTANT c_data_w                 : NATURAL := 14;

  -- wg_wideband_arr
  CONSTANT c_wg_buf_directory       : STRING  := "data/";
  CONSTANT c_wg_buf_dat_w           : NATURAL := 18; --default value of WG that fits 14 bits of ADC data
  CONSTANT c_wg_buf_addr_w          : NATURAL := 10; --default value of WG for 1024 samples
  CONSTANT c_bsn_nof_clk_per_sync   : NATURAL := sel_a_b(g_sim, 16*c_bs_block_size, c_sdp_N_clk_per_sync);

  -- ddrctrl
  CONSTANT c_tech_ddr               : t_c_tech_ddr          := func_tech_sel_ddr(g_sim, c_tech_ddr4_sim_16k, c_tech_ddr4_8g_1600m);
  CONSTANT c_stop_percentage        : NATURAL               := 80;

  -- diag_data_buffer
  CONSTANT c_data_type              : t_diag_data_type_enum := e_data;
  CONSTANT c_buf_nof_words          : NATURAL               := c_bs_block_size;

  -- diag_bsn_buffer
  CONSTANT c_bsn_type               : t_diag_data_type_enum := e_real;



  -- System
  SIGNAL cs_sim                     : STD_LOGIC;
  SIGNAL xo_ethclk                  : STD_LOGIC;
  SIGNAL xo_rst                     : STD_LOGIC;
  SIGNAL xo_rst_n                   : STD_LOGIC;
  SIGNAL not_mm_clk                 : STD_LOGIC;
  SIGNAL mm_clk                     : STD_LOGIC;
  SIGNAL mm_rst                     : STD_LOGIC := '0';
  SIGNAL st_pps                     : STD_LOGIC;
  
  SIGNAL st_rst                     : STD_LOGIC := '0';
  SIGNAL st_clk                     : STD_LOGIC;

  -- PIOs
  SIGNAL pout_wdi                   : STD_LOGIC;

  -- WDI override
  SIGNAL reg_wdi_mosi               : t_mem_mosi;
  SIGNAL reg_wdi_miso               : t_mem_miso;

  -- PPSH
  SIGNAL reg_ppsh_mosi              : t_mem_mosi;
  SIGNAL reg_ppsh_miso              : t_mem_miso;
  
  -- UniBoard system info
  SIGNAL reg_unb_system_info_mosi   : t_mem_mosi;
  SIGNAL reg_unb_system_info_miso   : t_mem_miso;
  SIGNAL rom_unb_system_info_mosi   : t_mem_mosi;
  SIGNAL rom_unb_system_info_miso   : t_mem_miso;

  -- FPGA sensors
  SIGNAL reg_fpga_temp_sens_mosi     : t_mem_mosi;
  SIGNAL reg_fpga_temp_sens_miso     : t_mem_miso;
  SIGNAL reg_fpga_voltage_sens_mosi  : t_mem_mosi;
  SIGNAL reg_fpga_voltage_sens_miso  : t_mem_miso;

  -- eth1g
  SIGNAL eth1g_mm_rst               : STD_LOGIC;
  SIGNAL eth1g_tse_mosi             : t_mem_mosi;  -- ETH TSE MAC registers
  SIGNAL eth1g_tse_miso             : t_mem_miso;
  SIGNAL eth1g_reg_mosi             : t_mem_mosi;  -- ETH control and status registers
  SIGNAL eth1g_reg_miso             : t_mem_miso;
  SIGNAL eth1g_reg_interrupt        : STD_LOGIC;   -- Interrupt
  SIGNAL eth1g_ram_mosi             : t_mem_mosi;  -- ETH rx frame and tx frame memory
  SIGNAL eth1g_ram_miso             : t_mem_miso;

  -- EPCS read
  SIGNAL reg_dpmm_data_mosi         : t_mem_mosi;
  SIGNAL reg_dpmm_data_miso         : t_mem_miso;
  SIGNAL reg_dpmm_ctrl_mosi         : t_mem_mosi;
  SIGNAL reg_dpmm_ctrl_miso         : t_mem_miso;

  -- EPCS write
  SIGNAL reg_mmdp_data_mosi         : t_mem_mosi;
  SIGNAL reg_mmdp_data_miso         : t_mem_miso;
  SIGNAL reg_mmdp_ctrl_mosi         : t_mem_mosi;
  SIGNAL reg_mmdp_ctrl_miso         : t_mem_miso;

  -- EPCS status/control
  SIGNAL reg_epcs_mosi              : t_mem_mosi;
  SIGNAL reg_epcs_miso              : t_mem_miso;

  -- Remote Update
  SIGNAL reg_remu_mosi              : t_mem_mosi;
  SIGNAL reg_remu_miso              : t_mem_miso;

  -- Scrap RAM
  SIGNAL ram_scrap_mosi             : t_mem_mosi;
  SIGNAL ram_scrap_miso             : t_mem_miso;

  -- QSFP leds
  SIGNAL qsfp_green_led_arr         : STD_LOGIC_VECTOR(c_unb2c_board_tr_qsfp.nof_bus-1 DOWNTO 0);
  SIGNAL qsfp_red_led_arr           : STD_LOGIC_VECTOR(c_unb2c_board_tr_qsfp.nof_bus-1 DOWNTO 0);

  -- bsn_source_v2
  SIGNAL  reg_bsn_source_v2_mosi    : t_mem_mosi;
  SIGNAL  reg_bsn_source_v2_miso    : t_mem_miso;
  SIGNAL  bs_sosi                   : t_dp_sosi;

  -- bsn_scheduler
  SIGNAL  reg_bsn_scheduler_wg_mosi : t_mem_mosi;
  SIGNAL  reg_bsn_scheduler_wg_miso : t_mem_miso;
  SIGNAL  trigger_wg                : STD_LOGIC;

  -- diag_wg_wideband_arr
  SIGNAL  reg_wg_mosi               : t_mem_mosi;
  SIGNAL  reg_wg_miso               : t_mem_miso;
  SIGNAL  ram_wg_mosi               : t_mem_mosi;
  SIGNAL  ram_wg_miso               : t_mem_miso;
  SIGNAL  wg_sosi_arr               : t_dp_sosi_arr(c_nof_streams-1 DOWNTO 0);

  -- stop_in
  SIGNAL  reg_stop_in_mosi          : t_mem_mosi;
  SIGNAL  reg_stop_in_miso          : t_mem_miso;
  SIGNAL  stop_in_arr               : STD_LOGIC_VECTOR(32-1 DOWNTO 0);

  -- ddrctrl
  SIGNAL  st_sosi_arr               : t_dp_sosi_arr(c_nof_streams-1 DOWNTO 0);
  SIGNAL  out_siso                  : t_dp_siso                               := c_dp_siso_rst;
  SIGNAL  out_sosi_arr              : t_dp_sosi_arr(c_nof_streams-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);
  SIGNAL  phy3_io                   : t_tech_ddr3_phy_io;
  SIGNAL  phy3_ou                   : t_tech_ddr3_phy_ou;
  SIGNAL  phy4_io                   : t_tech_ddr4_phy_io;
  SIGNAL  phy4_ou                   : t_tech_ddr4_phy_ou;

  -- data_buffer
  SIGNAL  reg_data_buf_mosi         : t_mem_mosi;
  SIGNAL  reg_data_buf_miso         : t_mem_miso;
  SIGNAL  ram_data_buf_mosi         : t_mem_mosi;
  SIGNAL  ram_data_buf_miso         : t_mem_miso;
  SIGNAL  reg_rx_seq_data_mosi      : t_mem_mosi;
  SIGNAL  reg_rx_seq_data_miso      : t_mem_miso;
  SIGNAL  out_wr_data_done_arr      : STD_LOGIC_VECTOR(c_nof_streams-1 DOWNTO 0);

  -- bsn_buffer
  SIGNAL  reg_bsn_buf_mosi          : t_mem_mosi;
  SIGNAL  reg_bsn_buf_miso          : t_mem_miso;
  SIGNAL  ram_bsn_buf_mosi          : t_mem_mosi;
  SIGNAL  ram_bsn_buf_miso          : t_mem_miso;
  SIGNAL  reg_rx_seq_bsn_mosi       : t_mem_mosi;
  SIGNAL  reg_rx_seq_bsn_miso       : t_mem_miso;
  SIGNAL  out_wr_bsn_done_arr       : STD_LOGIC_VECTOR(c_nof_streams-1 DOWNTO 0);




BEGIN

  mm_clk <= tb_clk;

  out_siso.ready        <= vector_and(NOT out_wr_data_done_arr) AND vector_and(NOT out_wr_bsn_done_arr);


  store_bsn_in_re : FOR I IN 0 TO c_nof_streams-1 GENERATE
    out_sosi_arr(I).re  <= out_sosi_arr(I).bsn;
  END GENERATE;

  u_bsn_source_v2 : ENTITY dp_lib.mms_dp_bsn_source_v2
  GENERIC MAP (
    g_cross_clock_domain     => TRUE,
    g_block_size             => c_bs_block_size,
    g_nof_clk_per_sync       => c_bsn_nof_clk_per_sync,
    g_bsn_w                  => c_bs_bsn_w
  )
  PORT MAP (
    -- Clocks and reset
    mm_rst            => mm_rst,
    mm_clk            => mm_clk,
    dp_rst            => st_rst,
    dp_clk            => st_clk,
    dp_pps            => st_pps,
    
    -- Memory-mapped clock domain
    reg_mosi          => reg_bsn_source_v2_mosi,
    reg_miso          => reg_bsn_source_v2_miso,
    
    -- Streaming clock domain
    bs_sosi           => bs_sosi,

    bs_restart        => open 
  );

  u_bsn_trigger_wg : ENTITY dp_lib.mms_dp_bsn_scheduler
  GENERIC MAP (
    g_cross_clock_domain => TRUE,
    g_bsn_w              => c_bs_bsn_w
  )
  PORT MAP (
    -- Memory-mapped clock domain
    mm_rst      => mm_rst,
    mm_clk      => mm_clk,

    reg_mosi    => reg_bsn_scheduler_wg_mosi,
    reg_miso    => reg_bsn_scheduler_wg_miso,

    -- Streaming clock domain
    dp_rst      => st_rst,
    dp_clk      => st_clk,

    snk_in      => bs_sosi,  -- only uses eop (= block sync), bsn[]
    trigger_out => trigger_wg
  );


  u_wg_arr : ENTITY diag_lib.mms_diag_wg_wideband_arr
  GENERIC MAP (
    g_nof_streams        => c_nof_streams, 
    g_cross_clock_domain => TRUE,
    g_buf_dir            => c_wg_buf_directory,

    -- Wideband parameters
    g_wideband_factor    => 1, 

    -- Basic WG parameters, see diag_wg.vhd for their meaning
    g_buf_dat_w          => c_wg_buf_dat_w,
    g_buf_addr_w         => c_wg_buf_addr_w,
    g_calc_support       => TRUE,
    g_calc_gain_w        => 1,
    g_calc_dat_w         => c_sdp_W_adc
  )
  PORT MAP (
    -- Memory-mapped clock domain
    mm_rst              => mm_rst,
    mm_clk              => mm_clk,

    reg_mosi            => reg_wg_mosi,
    reg_miso            => reg_wg_miso,

    buf_mosi            => ram_wg_mosi,
    buf_miso            => ram_wg_miso,

    -- Streaming clock domain
    st_rst              => st_rst,
    st_clk              => st_clk,
    st_restart          => trigger_wg,

    out_sosi_arr        => wg_sosi_arr
  );


  gen_concat : FOR I IN 0 TO c_sdp_S_pn-1 GENERATE
    p_sosi : PROCESS(wg_sosi_arr(I), bs_sosi)
    BEGIN
      st_sosi_arr(I)       <= bs_sosi;
      st_sosi_arr(I).data  <= wg_sosi_arr(I).data;
    END PROCESS;
  END GENERATE;


  u_stop_in_reg : ENTITY common_lib.mms_common_reg
  PORT MAP (
    -- Memory-mapped clock domain
    mm_rst              => mm_rst,
    mm_clk              => mm_clk,

    reg_mosi            => reg_stop_in_mosi,
    reg_miso            => reg_stop_in_miso,

    -- Streaming clock domain
    st_rst              => st_rst,
    st_clk              => st_clk,

    in_reg              => stop_in_arr,
    out_reg             => stop_in_arr
  );


  u_ddrctrl : ENTITY lofar2_ddrctrl_lib.ddrctrl
  GENERIC MAP (
    g_tech_ddr        => c_tech_ddr,
    g_sim_model       => g_sim,
    g_technology      => g_technology,
    g_nof_streams     => c_nof_streams,
    g_data_w          => c_data_w,
    g_stop_percentage => c_stop_percentage,
    g_block_size      => c_bs_block_size
  )
  PORT MAP (
    clk               => st_clk,
    rst               => st_rst,
    mm_clk            => mm_clk,
    mm_rst            => mm_rst,
    in_sosi_arr       => st_sosi_arr,
    stop_in           => stop_in_arr(0),
    out_sosi_arr      => out_sosi_arr,
    out_siso          => out_siso,

    --PHY
    phy3_io           => phy3_io,
    phy3_ou           => phy3_ou,
    phy4_io           => phy4_io,
    phy4_ou           => phy4_ou
  );

  u_diag_data_buffer : ENTITY diag_lib.mms_diag_data_buffer_dev
  GENERIC MAP (    
    g_technology      => g_technology,

    -- General
    g_nof_streams     => c_nof_streams,

    -- DB settings
    g_data_type       => c_data_type,
    g_data_w          => c_word_w,
    g_buf_nof_data    => c_buf_nof_words,
    g_buf_use_sync    => FALSE,
    g_use_steps       => FALSE,
    g_nof_steps       => c_diag_seq_rx_reg_nof_steps
  )
  PORT MAP (
    -- Memory-mapped clock domain
    mm_rst            => mm_rst,
    mm_clk            => mm_clk,

    reg_data_buf_mosi => reg_data_buf_mosi,
    reg_data_buf_miso => reg_data_buf_miso,

    ram_data_buf_mosi => ram_data_buf_mosi,
    ram_data_buf_miso => ram_data_buf_miso,

    -- Streaming clock domain
    dp_rst            => st_rst,
    dp_clk            => st_clk,
    
    in_sync           => st_pps,
    in_sosi_arr       => out_sosi_arr,
    out_wr_done_arr   => out_wr_data_done_arr
  );


  u_diag_bsn_buffer : ENTITY diag_lib.mms_diag_data_buffer_dev
  GENERIC MAP (    
    g_technology      => g_technology,

    -- General
    g_nof_streams     => c_nof_streams,

    -- DB settings
    g_data_type       => c_bsn_type,
    g_data_w          => c_dp_stream_bsn_w,
    g_buf_nof_data    => c_buf_nof_words,
    g_buf_use_sync    => FALSE,
    g_use_steps       => FALSE,
    g_nof_steps       => c_diag_seq_rx_reg_nof_steps
  )
  PORT MAP (
    -- Memory-mapped clock domain
    mm_rst            => mm_rst,
    mm_clk            => mm_clk,

    reg_data_buf_mosi => reg_bsn_buf_mosi,
    reg_data_buf_miso => reg_bsn_buf_miso,

    ram_data_buf_mosi => ram_bsn_buf_mosi,
    ram_data_buf_miso => ram_bsn_buf_miso,

    -- Streaming clock domain
    dp_rst            => st_rst,
    dp_clk            => st_clk,
    
    in_sync           => st_pps,
    in_sosi_arr       => out_sosi_arr,
    out_wr_done_arr   => out_wr_bsn_done_arr
  );


  -----------------------------------------------------------------------------
  -- General control function
  -----------------------------------------------------------------------------
  u_ctrl : ENTITY unb2c_board_lib.ctrl_unb2c_board
  GENERIC MAP (
    g_sim                => g_sim,
    g_technology         => g_technology,
    g_design_name        => g_design_name,
    g_design_note        => g_design_note,
    g_stamp_date         => g_stamp_date,
    g_stamp_time         => g_stamp_time, 
    g_revision_id        => g_revision_id, 
    g_fw_version         => c_fw_version,
    g_mm_clk_freq        => SEL_A_B(g_sim, 10**10, c_mm_clk_freq),
    g_eth_clk_freq       => c_unb2c_board_eth_clk_freq_125M,
    g_aux                => c_unb2c_board_aux,
    g_factory_image      => g_factory_image,
    g_protect_addr_range => g_protect_addr_range
  )
  PORT MAP (
    -- Clock an reset signals
    cs_sim                   => cs_sim,
    xo_ethclk                => xo_ethclk,
    xo_rst                   => xo_rst,
    xo_rst_n                 => xo_rst_n,

    mm_clk                   => not_mm_clk,
    mm_rst                   => mm_rst,

    dp_rst                   => st_rst,
    dp_clk                   => st_clk,
    dp_pps                   => st_pps,
    dp_rst_in                => st_rst,
    dp_clk_in                => st_clk,
    
    -- Toggle WDI
    pout_wdi                 => pout_wdi,

    -- MM buses
    -- REMU
    reg_remu_mosi            => reg_remu_mosi,
    reg_remu_miso            => reg_remu_miso,

    -- EPCS read
    reg_dpmm_data_mosi       => reg_dpmm_data_mosi,
    reg_dpmm_data_miso       => reg_dpmm_data_miso,
    reg_dpmm_ctrl_mosi       => reg_dpmm_ctrl_mosi,
    reg_dpmm_ctrl_miso       => reg_dpmm_ctrl_miso,

    -- EPCS write
    reg_mmdp_data_mosi       => reg_mmdp_data_mosi,
    reg_mmdp_data_miso       => reg_mmdp_data_miso,
    reg_mmdp_ctrl_mosi       => reg_mmdp_ctrl_mosi,
    reg_mmdp_ctrl_miso       => reg_mmdp_ctrl_miso,

    -- EPCS status/control
    reg_epcs_mosi            => reg_epcs_mosi,
    reg_epcs_miso            => reg_epcs_miso,

    -- . Manual WDI override
    reg_wdi_mosi             => reg_wdi_mosi,
    reg_wdi_miso             => reg_wdi_miso,
    
    -- . System_info
    reg_unb_system_info_mosi => reg_unb_system_info_mosi,
    reg_unb_system_info_miso => reg_unb_system_info_miso, 
    rom_unb_system_info_mosi => rom_unb_system_info_mosi,
    rom_unb_system_info_miso => rom_unb_system_info_miso, 
    
    -- . FPGA sensors
    reg_fpga_temp_sens_mosi  => reg_fpga_temp_sens_mosi,
    reg_fpga_temp_sens_miso  => reg_fpga_temp_sens_miso,
    reg_fpga_voltage_sens_mosi  => reg_fpga_voltage_sens_mosi,
    reg_fpga_voltage_sens_miso  => reg_fpga_voltage_sens_miso,

    -- . PPSH
    reg_ppsh_mosi            => reg_ppsh_mosi,
    reg_ppsh_miso            => reg_ppsh_miso,
    
    -- eth1g
    eth1g_mm_rst             => eth1g_mm_rst,
    eth1g_tse_mosi           => eth1g_tse_mosi,
    eth1g_tse_miso           => eth1g_tse_miso,
    eth1g_reg_mosi           => eth1g_reg_mosi,
    eth1g_reg_miso           => eth1g_reg_miso,
    eth1g_reg_interrupt      => eth1g_reg_interrupt,
    eth1g_ram_mosi           => eth1g_ram_mosi,
    eth1g_ram_miso           => eth1g_ram_miso,

    -- scrap ram
    ram_scrap_mosi           => ram_scrap_mosi,
    ram_scrap_miso           => ram_scrap_miso,        

    -- FPGA pins
    -- . General
    CLK                      => CLK,
    PPS                      => PPS,
    WDI                      => WDI,
    INTA                     => INTA,
    INTB                     => INTB,
    -- . Others
    VERSION                  => VERSION,
    ID                       => ID,
    TESTIO                   => TESTIO,

    -- . 1GbE Control Interface
--    ETH_clk                  => ETH_CLK(0),
--    ETH_SGIN                 => ETH_SGIN(0),
--    ETH_SGOUT                => ETH_SGOUT(0)

    ETH_clk                  => ETH_CLK(1),
    ETH_SGIN                 => ETH_SGIN(1),
    ETH_SGOUT                => ETH_SGOUT(1)
  );

  -----------------------------------------------------------------------------
  -- MM master
  -----------------------------------------------------------------------------
  u_mmm : ENTITY work.mmm_lofar2_unb2c_ddrctrl
  GENERIC MAP (
    g_sim         => g_sim,
    g_sim_unb_nr  => g_sim_unb_nr,
    g_sim_node_nr => g_sim_node_nr
   )
  PORT MAP(  
    mm_rst                   => mm_rst,
    mm_clk                   => mm_clk,

    -- PIOs
    pout_wdi                 => pout_wdi,

    -- Manual WDI override
    reg_wdi_mosi             => reg_wdi_mosi,
    reg_wdi_miso             => reg_wdi_miso,

    -- system_info
    reg_unb_system_info_mosi => reg_unb_system_info_mosi,
    reg_unb_system_info_miso => reg_unb_system_info_miso,
    rom_unb_system_info_mosi => rom_unb_system_info_mosi,
    rom_unb_system_info_miso => rom_unb_system_info_miso, 

    -- FPGA sensors
    reg_fpga_temp_sens_mosi  => reg_fpga_temp_sens_mosi,
    reg_fpga_temp_sens_miso  => reg_fpga_temp_sens_miso,
    reg_fpga_voltage_sens_mosi  => reg_fpga_voltage_sens_mosi,
    reg_fpga_voltage_sens_miso  => reg_fpga_voltage_sens_miso,

    -- PPSH
    reg_ppsh_mosi            => reg_ppsh_mosi,
    reg_ppsh_miso            => reg_ppsh_miso, 
  
    -- eth1g
    eth1g_mm_rst             => eth1g_mm_rst,
    eth1g_tse_mosi           => eth1g_tse_mosi,
    eth1g_tse_miso           => eth1g_tse_miso,
    eth1g_reg_mosi           => eth1g_reg_mosi,
    eth1g_reg_miso           => eth1g_reg_miso,
    eth1g_reg_interrupt      => eth1g_reg_interrupt,
    eth1g_ram_mosi           => eth1g_ram_mosi,
    eth1g_ram_miso           => eth1g_ram_miso,

    -- EPCS read
    reg_dpmm_data_mosi       => reg_dpmm_data_mosi,
    reg_dpmm_data_miso       => reg_dpmm_data_miso,
    reg_dpmm_ctrl_mosi       => reg_dpmm_ctrl_mosi,
    reg_dpmm_ctrl_miso       => reg_dpmm_ctrl_miso,

    -- EPCS write
    reg_mmdp_data_mosi       => reg_mmdp_data_mosi,
    reg_mmdp_data_miso       => reg_mmdp_data_miso,
    reg_mmdp_ctrl_mosi       => reg_mmdp_ctrl_mosi,
    reg_mmdp_ctrl_miso       => reg_mmdp_ctrl_miso,

    -- EPCS status/control
    reg_epcs_mosi            => reg_epcs_mosi,
    reg_epcs_miso            => reg_epcs_miso,

    -- Remote Update
    reg_remu_mosi            => reg_remu_mosi,
    reg_remu_miso            => reg_remu_miso,   

    -- Scrap RAM
    ram_scrap_mosi           => ram_scrap_mosi,
    ram_scrap_miso           => ram_scrap_miso,

    -- bsn_source_v2
    reg_bsn_source_v2_mosi   => reg_bsn_source_v2_mosi,
    reg_bsn_source_v2_miso   => reg_bsn_source_v2_miso,

    -- bsn_scheduler
    reg_bsn_scheduler_wg_mosi => reg_bsn_scheduler_wg_mosi,
    reg_bsn_scheduler_wg_miso => reg_bsn_scheduler_wg_miso,

    -- wg_wideband_arr
    reg_wg_mosi             => reg_wg_mosi,
    reg_wg_miso             => reg_wg_miso,
    ram_wg_mosi             => ram_wg_mosi,
    ram_wg_miso             => ram_wg_miso,

    -- stop_in
    reg_stop_in_mosi        => reg_stop_in_mosi,
    reg_stop_in_miso        => reg_stop_in_miso,

    -- data_buffer
    reg_data_buf_mosi       => reg_data_buf_mosi,
    reg_data_buf_miso       => reg_data_buf_miso,
    ram_data_buf_mosi       => ram_data_buf_mosi,
    ram_data_buf_miso       => ram_data_buf_miso,
    reg_rx_seq_data_mosi    => reg_rx_seq_data_mosi,
    reg_rx_seq_data_miso    => reg_rx_seq_data_miso,

    -- bsn_buffer
    reg_bsn_buf_mosi        => reg_bsn_buf_mosi,
    reg_bsn_buf_miso        => reg_bsn_buf_miso,
    ram_bsn_buf_mosi        => ram_bsn_buf_mosi,
    ram_bsn_buf_miso        => ram_bsn_buf_miso,
    reg_rx_seq_bsn_mosi     => reg_rx_seq_bsn_mosi,
    reg_rx_seq_bsn_miso     => reg_rx_seq_bsn_miso

  );

  u_front_led : ENTITY unb2c_board_lib.unb2c_board_qsfp_leds
  GENERIC MAP (
    g_sim           => g_sim,
    g_factory_image => g_factory_image,
    g_nof_qsfp      => c_unb2c_board_tr_qsfp.nof_bus,
    g_pulse_us      => 1000 / (10**9 / c_mm_clk_freq) -- nof clk cycles to get us period
  )
  PORT MAP (
    rst             => mm_rst,
    clk             => mm_clk,
    green_led_arr   => qsfp_green_led_arr,
    red_led_arr     => qsfp_red_led_arr
  );

  u_front_io : ENTITY unb2c_board_lib.unb2c_board_front_io
  GENERIC MAP (
    g_nof_qsfp_bus => c_unb2c_board_tr_qsfp.nof_bus
  )
  PORT MAP (
    green_led_arr => qsfp_green_led_arr,
    red_led_arr   => qsfp_red_led_arr,
    QSFP_LED      => QSFP_LED
  );

END str;

