-------------------------------------------------------------------------------
--
-- Copyright (C) 2012
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- JIVE (Joint Institute for VLBI in Europe) <http://www.jive.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- This program is free software: you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation, either version 3 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program.  If not, see <http://www.gnu.org/licenses/>.
--
-------------------------------------------------------------------------------

-- Purpose: Test bench for lofar2_unb2c_ddrctrl.
-- Description:
--   The DUT can be targeted at unb 0, node 3 with the same Python scripts 
--   that are used on hardware. 
-- Usage:
--   On command line do:
--     > run_modelsim & (to start Modeslim)
--
--   In Modelsim do:
--     > lp lofar2_unb2c_ddrctrl
--     > mk clean all (only first time to clean all libraries)
--     > mk all (to compile all libraries that are needed for lofar2_unb2c_ddrctrl)
--     . load tb_unb1_minimal simulation by double clicking the tb_lofar2_unb2c_ddrctrl icon
--     > as 10 (to view signals in Wave Window)
--     > run 100 us (or run -all)
--
--   On command line do:
--     > python $UPE_GEAR/peripherals/util_system_info.py --gn 3 -n 0 -v 5 --sim
--     > python $UPE_GEAR/peripherals/util_ppsh.py --gn 3 -n 1 -v 5 --sim
--

LIBRARY IEEE, common_lib, unb2c_board_lib, i2c_lib, technology_lib, mm_lib, dp_lib, tech_ddr_lib, lofar2_sdp_lib, diag_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE common_lib.common_pkg.ALL;
USE unb2c_board_lib.unb2c_board_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE i2c_lib.i2c_dev_unb2_pkg.ALL;
USE i2c_lib.i2c_commander_unb2_pmbus_pkg.ALL;
USE technology_lib.technology_select_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE tech_ddr_lib.tech_ddr_pkg.ALL;
USE lofar2_sdp_lib.sdp_pkg.ALL;
USE diag_lib.diag_pkg.ALL;
USE mm_lib.mm_file_pkg.ALL;
USE mm_lib.mm_file_unb_pkg.ALL;


ENTITY tb_lofar2_unb2c_ddrctrl IS
    GENERIC (
      g_design_name : STRING  := "lofar2_unb2c_ddrctrl";
      g_sim_unb_nr  : NATURAL := 0; -- UniBoard 0
      g_sim_node_nr : NATURAL := 3  -- Node 3
    );
END tb_lofar2_unb2c_ddrctrl;

ARCHITECTURE tb OF tb_lofar2_unb2c_ddrctrl IS

  CONSTANT c_sim             : BOOLEAN := TRUE;
  CONSTANT c_rd_data_w       : NATURAL := 32;

  CONSTANT c_unb_nr          : NATURAL := 0; -- UniBoard 0
  CONSTANT c_node_nr         : NATURAL := 3; -- Node 3
  CONSTANT c_id              : STD_LOGIC_VECTOR(7 DOWNTO 0) := TO_UVEC(c_unb_nr, c_unb2c_board_nof_uniboard_w) & TO_UVEC(c_node_nr, c_unb2c_board_nof_chip_w);

  CONSTANT c_version         : STD_LOGIC_VECTOR(1 DOWNTO 0) := "00";
  CONSTANT c_fw_version      : t_unb2c_board_fw_version := (1, 0);

  CONSTANT c_cable_delay     : TIME := 12 ns;
  CONSTANT c_eth_clk_period  : TIME := 8 ns;  -- 125 MHz XO on UniBoardi
  CONSTANT c_st_clk_period   : TIME := 5 ns;  -- 200 MHz st clk
  CONSTANT c_tb_clk_period   : TIME := 1 ns;  -- 1000 MHz tb clk
  CONSTANT c_mm_clk_period   : TIME := 10 ns; -- 100 MHz mm clk
  CONSTANT c_pps_period      : NATURAL := 1000; 

  CONSTANT c_mm_file_reg_bsn_source_v2        : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "REG_BSN_SOURCE_V2";
  CONSTANT c_mm_file_reg_bsn_scheduler_wg     : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "REG_BSN_SCHEDULER";
  CONSTANT c_mm_file_reg_diag_wg_wideband_arr : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "REG_WG_WIDEBAND_ARR";
  CONSTANT c_mm_file_reg_stop_in              : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "REG_STOP_IN";
  CONSTANT c_mm_file_reg_data_buf             : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "REG_DATA_BUF";
  CONSTANT c_mm_file_ram_data_buf             : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "RAM_DATA_BUF";
  CONSTANT c_mm_file_reg_rx_seq_data          : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "REG_RX_SEQ_DATA";
  CONSTANT c_mm_file_reg_bsn_buf              : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "REG_BSN_BUF";
  CONSTANT c_mm_file_ram_bsn_buf              : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "RAM_BSN_BUF";
  CONSTANT c_mm_file_reg_rx_seq_bsn           : STRING := mmf_unb_file_prefix(c_unb_nr, c_node_nr) & "REG_RX_SEQ_BSN";

  -- c_check_vector
  CONSTANT  c_tech_ddr        : t_c_tech_ddr            := c_tech_ddr4_sim_16k;
  CONSTANT  c_ctrl_data_w     : NATURAL                 := func_tech_ddr_ctlr_data_w(c_tech_ddr);                             -- 576
  CONSTANT  c_adr_w           : NATURAL                 := func_tech_ddr_ctlr_address_w(c_tech_ddr);                          -- the lengt of the address vector, for simulation this is smaller, otherwise the simulation would take to long, 27
  CONSTANT  c_max_adr         : NATURAL                 := 2**(c_adr_w)-1;                                                    -- the maximal address that is possible within the address vector length
  CONSTANT  c_block_size      : NATURAL                 := 1024;
  CONSTANT  c_nof_streams     : NATURAL                 := 12;
  CONSTANT  c_data_w          : NATURAL                 := 14;
  CONSTANT  c_bim             : NATURAL                 := (c_max_adr*c_ctrl_data_w)/(c_block_size*c_nof_streams*c_data_w);   -- the amount of whole blocks that fit in memory.
  CONSTANT  c_nof_st_in_mem   : NATURAL                 := c_bim*c_block_size;


  CONSTANT c_check_vector     : STD_LOGIC_VECTOR(c_ctrl_data_w*c_bim*c_block_size-1 DOWNTO 0) := (OTHERS => '0'); -- the sinewave of one stream for c_bim length

  -- BSN
  CONSTANT c_init_bsn             : NATURAL := 17;  -- some recognizable value >= 0
  CONSTANT c_nof_block_per_sync   : NATURAL := 16;
  CONSTANT c_nof_clk_per_sync     : NATURAL := c_nof_block_per_sync * c_sdp_N_fft;

  CONSTANT c_speed            : NATURAL   := 4;

  -- WG
  CONSTANT c_wg_phase         : REAL      := 0.0;   -- WG phase in degrees
  CONSTANT c_wg_freq          : REAL      := 160.0; -- WG freq
  CONSTANT c_wg_ampl          : NATURAL   := NATURAL(1.0*REAL(c_sdp_FS_adc));  -- in number of lsb
  CONSTANT c_bsn_start_wg     : NATURAL   := c_init_bsn + 2;  -- start WG at this BSN to instead of some BSN, to avoid mismatches in exact expected data values
  CONSTANT c_subband          : REAL      := 1.0*(2**c_speed);

  -- DUT
  SIGNAL st_clk               : STD_LOGIC := '0';
  SIGNAL tb_clk               : STD_LOGIC := '0';
  SIGNAL tb_end               : STD_LOGIC := '0';
  SIGNAL pps                  : STD_LOGIC := '0';
  SIGNAL pps_rst              : STD_LOGIC := '0';

  SIGNAL rd_data              : STD_LOGIC_VECTOR(c_rd_data_w-1 DOWNTO 0)  := (OTHERS => '0');
  SIGNAL sosi_out_data        : STD_LOGIC_VECTOR(c_rd_data_w-1 DOWNTO 0)  := (OTHERS => '0');
  SIGNAL sosi_out_bsn         : STD_LOGIC_VECTOR(c_rd_data_w-1 DOWNTO 0)  := (OTHERS => '0');
  SIGNAL sosi_out_data_sin    : STD_LOGIC_VECTOR(   c_data_w-1 DOWNTO 0)  := (OTHERS => '0');
  SIGNAL sosi_out_not_bsn     : STD_LOGIC_VECTOR(c_rd_data_w-1 DOWNTO 0)  := (OTHERS => '0');

  SIGNAL WDI                  : STD_LOGIC;
  SIGNAL INTA                 : STD_LOGIC;
  SIGNAL INTB                 : STD_LOGIC;

  SIGNAL eth_clk              : STD_LOGIC_VECTOR(1 DOWNTO 0);
  SIGNAL eth_txp              : STD_LOGIC_VECTOR(1 DOWNTO 0);
  SIGNAL eth_rxp              : STD_LOGIC_VECTOR(1 DOWNTO 0);
  
  SIGNAL VERSION              : STD_LOGIC_VECTOR(c_unb2c_board_aux.version_w-1 DOWNTO 0) := c_version; 
  SIGNAL ID                   : STD_LOGIC_VECTOR(c_unb2c_board_aux.id_w-1 DOWNTO 0)      := c_id;
  SIGNAL TESTIO               : STD_LOGIC_VECTOR(c_unb2c_board_aux.testio_w-1 DOWNTO 0);

  SIGNAL qsfp_led             : STD_LOGIC_VECTOR(c_unb2c_board_tr_qsfp_nof_leds-1 DOWNTO 0);

  SIGNAL current_bsn_wg       : STD_LOGIC_VECTOR(c_dp_stream_bsn_w-1 DOWNTO 0);


BEGIN

  ----------------------------------------------------------------------------
  -- System setup
  ----------------------------------------------------------------------------
  st_clk     <= NOT st_clk OR tb_end AFTER c_st_clk_period/2;        -- External clock (200 MHz)
  eth_clk(0) <= NOT eth_clk(0) OR tb_end AFTER c_eth_clk_period/2;  -- Ethernet ref clock (25 MHz)
  
  INTA <= 'H';  -- pull up
  INTB <= 'H';  -- pull up

  ------------------------------------------------------------------------------
  -- External PPS
  ------------------------------------------------------------------------------  
  proc_common_gen_pulse(1, c_pps_period, '1', pps_rst, st_clk, pps);

  ------------------------------------------------------------------------------
  -- 1GbE Loopback model
  ------------------------------------------------------------------------------  
  eth_rxp(0) <= TRANSPORT eth_txp(0) AFTER c_cable_delay;
  
  eth_rxp(1) <= '0';
  eth_txp(1) <= '0';

  ------------------------------------------------------------------------------
  -- DUT
  ------------------------------------------------------------------------------
   u_lofar2_unb2c_ddrctrl : ENTITY work.lofar2_unb2c_ddrctrl
    GENERIC MAP (
      g_sim         => c_sim,
      g_sim_unb_nr  => c_unb_nr,
      g_sim_node_nr => c_node_nr,
      g_design_name => g_design_name
    )
    PORT MAP (
      -- GENERAL
      CLK         => st_clk,
      tb_clk      => tb_clk,
      PPS         => pps,
      WDI         => WDI,
      INTA        => INTA,
      INTB        => INTB,
      
      -- Others
      VERSION     => VERSION,
      ID          => ID,
      TESTIO      => TESTIO,

      -- 1GbE Control Interface
      ETH_clk     => eth_clk,
      ETH_SGIN    => eth_rxp,
      ETH_SGOUT   => eth_txp,

      QSFP_LED    => qsfp_led
    );


    -- WG


  ------------------------------------------------------------------------------
  -- MM slave accesses via file IO
  ------------------------------------------------------------------------------
  tb_clk <= NOT tb_clk OR tb_end AFTER c_tb_clk_period/2;    -- Testbench MM clock



  p_mm_stimuli : PROCESS

  VARIABLE v_bsn    : NATURAL  := 0;

  BEGIN

    WAIT FOR 1 us;
    ----------------------------------------------------------------------------
    -- Enable BSN
    ----------------------------------------------------------------------------
    mmf_mm_bus_wr(c_mm_file_reg_bsn_source_v2, 2,         c_init_bsn, tb_clk);  -- Init BSN
    mmf_mm_bus_wr(c_mm_file_reg_bsn_source_v2, 3,                  0, tb_clk);  -- Write high part activates the init BSN
    mmf_mm_bus_wr(c_mm_file_reg_bsn_source_v2, 1, c_nof_clk_per_sync, tb_clk);  -- nof_block_per_sync
    mmf_mm_bus_wr(c_mm_file_reg_bsn_source_v2, 0,       16#00000003#, tb_clk);  -- Enable BSN at PPS



    --mmf_mm_bus_rd(c_mm_file_reg_dp_selector, 0, rd_data, tb_clk);

    ----------------------------------------------------------------------------
    -- Enable and start WG
    ----------------------------------------------------------------------------
    --   0 : mode[7:0]           --> off=0, calc=1, repeat=2, single=3)
    --       nof_samples[31:16]  --> <= c_ram_wg_size=1024
    --   1 : phase[15:0]
    --   2 : freq[30:0]
    --   3 : ampl[16:0]

    FOR I IN 0 TO c_nof_streams-1 LOOP
      mmf_mm_bus_wr(c_mm_file_reg_diag_wg_wideband_arr, (I*4)+0, 1024*2**16 + 1, tb_clk);      -- nof_samples, mode calc
      mmf_mm_bus_wr(c_mm_file_reg_diag_wg_wideband_arr, (I*4)+1, INTEGER(c_wg_phase), tb_clk); -- phase offset in degrees
      mmf_mm_bus_wr(c_mm_file_reg_diag_wg_wideband_arr, (I*4)+2, INTEGER(REAL(c_subband)*c_sdp_wg_subband_freq_unit), tb_clk);  -- freq
      mmf_mm_bus_wr(c_mm_file_reg_diag_wg_wideband_arr, (I*4)+3, INTEGER(REAL(c_wg_ampl)*c_sdp_wg_ampl_lsb), tb_clk);  -- ampl
    END LOOP;
    mmf_mm_bus_wr(c_mm_file_reg_stop_in, 0,     0, tb_clk);
    -- Read current BSN
    mmf_mm_bus_rd(c_mm_file_reg_bsn_scheduler_wg, 0, current_bsn_wg(31 DOWNTO  0), tb_clk);
    mmf_mm_bus_rd(c_mm_file_reg_bsn_scheduler_wg, 1, current_bsn_wg(63 DOWNTO 32), tb_clk);
    proc_common_wait_some_cycles(tb_clk, 1);

    -- Write scheduler BSN to trigger start of WG at next block
    v_bsn := TO_UINT(current_bsn_wg) + 2;
    ASSERT v_bsn <= c_bsn_start_wg REPORT "Too late to start WG: " & NATURAL'image(v_bsn) & " > " & NATURAL'image(c_bsn_start_wg) SEVERITY ERROR;
    mmf_mm_bus_wr(c_mm_file_reg_bsn_scheduler_wg, 0, c_bsn_start_wg, tb_clk);  -- first write low then high part
    mmf_mm_bus_wr(c_mm_file_reg_bsn_scheduler_wg, 1,              0, tb_clk);  -- assume v_bsn < 2**31-1


    WAIT FOR c_mm_clk_period*24000;
    mmf_mm_bus_wr(c_mm_file_reg_stop_in, 0, 1, tb_clk);
    WAIT FOR c_mm_clk_period*16;
    mmf_mm_bus_wr(c_mm_file_reg_stop_in, 0, 0, tb_clk);
    WAIT FOR c_mm_clk_period*50000;


    assert false report "1. read whole memory!" severity note;
    FOR I IN 0 TO c_bim-1 LOOP
      FOR J IN 0 TO c_nof_streams-1 LOOP
        FOR K IN c_block_size-(c_block_size/(2**c_speed)) TO c_block_size-1 LOOP
          mmf_mm_bus_rd(c_mm_file_ram_data_buf, (J*c_block_size)+K      ,     sosi_out_data(c_rd_data_w-1 DOWNTO 0), tb_clk);
          sosi_out_data_sin(c_data_w-1 DOWNTO 0) <= sosi_out_data(c_data_w-1 DOWNTO 0);
          mmf_mm_bus_rd( c_mm_file_ram_bsn_buf, ((J*c_block_size)+k)*2  ,      sosi_out_bsn(c_rd_data_w-1 DOWNTO 0), tb_clk);
          mmf_mm_bus_rd( c_mm_file_ram_bsn_buf, ((J*c_block_size)+k)*2+1,  sosi_out_not_bsn(c_rd_data_w-1 DOWNTO 0), tb_clk);
        END LOOP;
      END LOOP;
      WAIT FOR c_st_clk_period*c_block_size;
    END LOOP;


    WAIT FOR c_mm_clk_period*2400;
    mmf_mm_bus_wr(c_mm_file_reg_stop_in, 0, 1, tb_clk);
    WAIT FOR c_mm_clk_period*30;
    mmf_mm_bus_wr(c_mm_file_reg_stop_in, 0, 0, tb_clk);
    WAIT FOR c_mm_clk_period*55000;


    assert false report "2. read whole memory!" severity note;
    FOR I IN 0 TO c_bim-1 LOOP
      FOR J IN 0 TO c_nof_streams-1 LOOP
        FOR K IN c_block_size-(c_block_size/(2**c_speed)) TO c_block_size-1 LOOP
          mmf_mm_bus_rd(c_mm_file_ram_data_buf, (J*c_block_size)+K      ,     sosi_out_data(c_rd_data_w-1 DOWNTO 0), tb_clk);
          sosi_out_data_sin(c_data_w-1 DOWNTO 0) <= sosi_out_data(c_data_w-1 DOWNTO 0);
          mmf_mm_bus_rd( c_mm_file_ram_bsn_buf, ((J*c_block_size)+k)*2  ,      sosi_out_bsn(c_rd_data_w-1 DOWNTO 0), tb_clk);
          mmf_mm_bus_rd( c_mm_file_ram_bsn_buf, ((J*c_block_size)+k)*2+1,  sosi_out_not_bsn(c_rd_data_w-1 DOWNTO 0), tb_clk);
        END LOOP;
      END LOOP;
      WAIT FOR c_st_clk_period*c_block_size;
    END LOOP;


    WAIT FOR c_mm_clk_period*2200;
    mmf_mm_bus_wr(c_mm_file_reg_stop_in, 0, 1, tb_clk);
    WAIT FOR c_mm_clk_period*300;
    mmf_mm_bus_wr(c_mm_file_reg_stop_in, 0, 0, tb_clk);
    WAIT FOR c_mm_clk_period*55000;


    assert false report "3. read whole memory!" severity note;
    FOR I IN 0 TO c_bim-1 LOOP
      FOR J IN 0 TO c_nof_streams-1 LOOP
        FOR K IN c_block_size-(c_block_size/(2**c_speed)) TO c_block_size-1 LOOP
          mmf_mm_bus_rd(c_mm_file_ram_data_buf, (J*c_block_size)+K      ,     sosi_out_data(c_rd_data_w-1 DOWNTO 0), tb_clk);
          sosi_out_data_sin(c_data_w-1 DOWNTO 0) <= sosi_out_data(c_data_w-1 DOWNTO 0);
          mmf_mm_bus_rd( c_mm_file_ram_bsn_buf, ((J*c_block_size)+k)*2  ,      sosi_out_bsn(c_rd_data_w-1 DOWNTO 0), tb_clk);
          mmf_mm_bus_rd( c_mm_file_ram_bsn_buf, ((J*c_block_size)+k)*2+1,  sosi_out_not_bsn(c_rd_data_w-1 DOWNTO 0), tb_clk);
        END LOOP;
      END LOOP;
      WAIT FOR c_st_clk_period*c_block_size;
    END LOOP;

    tb_end <= '1';
    ASSERT FALSE  REPORT "Test: OK" SEVERITY FAILURE;
    WAIT;
  END PROCESS;



END tb;
