-------------------------------------------------------------------------------
--
-- Copyright 2021
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--
-- Author: R. van der Walle
-- Purpose: Self-checking testbench for simulating lofar2_unb2c_ring using BG data.
--
-- Description:
-- See, https://support.astron.nl/confluence/x/jyu7Ag
--
-- Usage:
--   > as 7    # default
--   > as 12   # for detailed debugging
--   > run -a  
--
-------------------------------------------------------------------------------
LIBRARY IEEE, common_lib, unb2c_board_lib, i2c_lib, mm_lib, dp_lib, diag_lib, lofar2_sdp_lib, wpfb_lib, tech_pll_lib, tr_10GbE_lib, ring_lib;
USE IEEE.std_logic_1164.ALL;
USE IEEE.numeric_std.ALL;
USE IEEE.MATH_REAL.ALL;
USE common_lib.common_pkg.ALL;
USE unb2c_board_lib.unb2c_board_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE common_lib.common_str_pkg.ALL;
USE mm_lib.mm_file_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE mm_lib.mm_file_unb_pkg.ALL;
USE diag_lib.diag_pkg.ALL;
USE wpfb_lib.wpfb_pkg.ALL;
USE lofar2_sdp_lib.sdp_pkg.ALL;
USE tech_pll_lib.tech_pll_component_pkg.ALL;
USE ring_lib.ring_pkg.ALL;
USE work.lofar2_unb2c_ring_pkg.ALL;

ENTITY tb_lofar2_unb2c_ring IS
  GENERIC (
    g_multi_tb            : BOOLEAN              := FALSE;
    g_unb_nr              : NATURAL              := 0;
    g_design_name         : STRING               := "lofar2_unb2c_ring_one";
    g_nof_rn              : NATURAL              := 3;
    g_nof_block_per_sync  : NATURAL              := 3;
    g_access_scheme       : INTEGER RANGE 1 TO 3 := 1
  );
  PORT (
    tb_end : OUT STD_LOGIC := '0' -- For multi tb 
  );

END tb_lofar2_unb2c_ring;

ARCHITECTURE tb OF tb_lofar2_unb2c_ring IS

  CONSTANT c_sim             : BOOLEAN := TRUE;
  CONSTANT c_node_nr         : NATURAL := 0; 
  CONSTANT c_version         : STD_LOGIC_VECTOR(1 DOWNTO 0) := "00";
  CONSTANT c_fw_version      : t_unb2c_board_fw_version := (1, 0);
  
  CONSTANT c_eth_clk_period      : TIME := 8 ns;  -- 125 MHz XO on UniBoard
  CONSTANT c_ext_clk_period      : TIME := 5 ns;
  CONSTANT c_bck_ref_clk_period  : TIME := 5 ns;
  CONSTANT c_sa_clk_period       : TIME := tech_pll_clk_644_period; -- 644MHz
  CONSTANT c_pps_period          : NATURAL := 1000;

  CONSTANT c_tb_clk_period       : TIME := 100 ps; -- use fast tb_clk to speed up M&C
  CONSTANT c_cable_delay         : TIME := 12 ns;

  CONSTANT c_revision_select     : t_lofar2_unb2c_ring_config := func_sel_revision_rec(g_design_name);
  CONSTANT c_nof_lanes           : NATURAL := c_revision_select.N_ring_lanes;

  CONSTANT c_block_period        : NATURAL := 1024;
  CONSTANT c_blocksize           : NATURAL := c_sdp_V_ring_pkt_len_max - c_ring_dp_hdr_field_size;
  CONSTANT c_gapsize             : NATURAL := c_block_period - c_blocksize;
  CONSTANT c_sync_timeout        : NATURAL := c_block_period * g_nof_block_per_sync + 10; -- +10 for extra slack
  CONSTANT c_exp_bsn_at_sync     : NATURAL := g_nof_block_per_sync;
  CONSTANT c_exp_nof_sop         : NATURAL := g_nof_block_per_sync;
  CONSTANT c_exp_nof_valid       : NATURAL := g_nof_block_per_sync * c_blocksize;

 
  -- MM  
  CONSTANT c_mm_file_reg_ring_info               : STRING := mmf_unb_file_prefix(g_unb_nr, c_node_nr) & "REG_RING_INFO";
  CONSTANT c_mm_file_reg_ring_lane_info          : STRING := mmf_unb_file_prefix(g_unb_nr, c_node_nr) & "REG_RING_LANE_INFO";
  CONSTANT c_mm_file_reg_diag_bg                 : STRING := mmf_unb_file_prefix(g_unb_nr, c_node_nr) & "REG_DIAG_BG";
  CONSTANT c_mm_file_ram_diag_bg                 : STRING := mmf_unb_file_prefix(g_unb_nr, c_node_nr) & "RAM_DIAG_BG";
  CONSTANT c_mm_file_reg_dp_xonoff_lane          : STRING := mmf_unb_file_prefix(g_unb_nr, c_node_nr) & "REG_DP_XONOFF_LANE";
  CONSTANT c_mm_file_reg_dp_xonoff_local         : STRING := mmf_unb_file_prefix(g_unb_nr, c_node_nr) & "REG_DP_XONOFF_LOCAL";
  CONSTANT c_mm_file_reg_bsn_monitor_v2_ring_rx  : STRING := mmf_unb_file_prefix(g_unb_nr, c_node_nr) & "REG_BSN_MONITOR_V2_RING_RX";
  CONSTANT c_mm_file_reg_bsn_monitor_v2_ring_tx  : STRING := mmf_unb_file_prefix(g_unb_nr, c_node_nr) & "REG_BSN_MONITOR_V2_RING_TX";

  -- Tb
  SIGNAL sim_done            : STD_LOGIC := '0';
  SIGNAL tb_clk              : STD_LOGIC := '0';  
  SIGNAL i_tb_end            : STD_LOGIC := '0';
  SIGNAL rd_data             : STD_LOGIC_VECTOR(c_32-1 DOWNTO 0);

  SIGNAL i_QSFP_0_TX         : t_unb2c_board_qsfp_bus_2arr(g_nof_rn -1 DOWNTO 0) := (OTHERS => (OTHERS => '0'));
  SIGNAL i_QSFP_0_RX         : t_unb2c_board_qsfp_bus_2arr(g_nof_rn -1 DOWNTO 0) := (OTHERS => (OTHERS => '0'));
  SIGNAL i_RING_0_TX         : t_unb2c_board_qsfp_bus_2arr(g_nof_rn -1 DOWNTO 0) := (OTHERS => (OTHERS => '0'));
  SIGNAL i_RING_0_RX         : t_unb2c_board_qsfp_bus_2arr(g_nof_rn -1 DOWNTO 0) := (OTHERS => (OTHERS => '0'));
  SIGNAL i_RING_1_TX         : t_unb2c_board_qsfp_bus_2arr(g_nof_rn -1 DOWNTO 0) := (OTHERS => (OTHERS => '0'));
  SIGNAL i_RING_1_RX         : t_unb2c_board_qsfp_bus_2arr(g_nof_rn -1 DOWNTO 0) := (OTHERS => (OTHERS => '0'));

  -- DUT
  SIGNAL ext_clk             : STD_LOGIC := '0';
  SIGNAL pps                 : STD_LOGIC := '0';
  SIGNAL ext_pps             : STD_LOGIC := '0'; 
  SIGNAL pps_rst             : STD_LOGIC := '1';

  SIGNAL WDI                 : STD_LOGIC;
  SIGNAL INTA                : STD_LOGIC;
  SIGNAL INTB                : STD_LOGIC;

  SIGNAL eth_clk             : STD_LOGIC_VECTOR(c_unb2c_board_nof_eth-1 downto 0) := (OTHERS => '0');
  SIGNAL eth_txp             : STD_LOGIC_VECTOR(c_unb2c_board_nof_eth-1 downto 0);
  SIGNAL eth_rxp             : STD_LOGIC_VECTOR(c_unb2c_board_nof_eth-1 downto 0);

  SIGNAL sens_scl            : STD_LOGIC;
  SIGNAL sens_sda            : STD_LOGIC;
  SIGNAL pmbus_scl           : STD_LOGIC;
  SIGNAL pmbus_sda           : STD_LOGIC;

  SIGNAL SA_CLK              : STD_LOGIC := '1';
   

BEGIN

  ----------------------------------------------------------------------------
  -- System setup
  ----------------------------------------------------------------------------
  tb_clk <= NOT tb_clk AFTER c_tb_clk_period/2;  -- Testbench MM clock

  ext_clk <= NOT ext_clk OR i_tb_end AFTER c_ext_clk_period/2;  -- External clock (200 MHz)
  eth_clk(0) <= NOT eth_clk(0) OR i_tb_end AFTER c_eth_clk_period/2;  -- Ethernet ref clock (125 MHz)
  SA_CLK <= NOT SA_CLK OR i_tb_end AFTER c_sa_clk_period/2;  -- Serial Gigabit IO sa clock (644 MHz)

  pps_rst <= '0' AFTER c_ext_clk_period*2;

  INTA <= 'H';  -- pull up
  INTB <= 'H';  -- pull up

  sens_scl <= 'H';  -- pull up
  sens_sda <= 'H';  -- pull up
  pmbus_scl <= 'H';  -- pull up
  pmbus_sda <= 'H';  -- pull up

  ------------------------------------------------------------------------------
  -- External PPS
  ------------------------------------------------------------------------------  
  proc_common_gen_pulse(5, c_pps_period, '1', pps_rst, ext_clk, pps);
  ext_pps <= pps;

  ------------------------------------------------------------------------------
  -- DUTs
  ------------------------------------------------------------------------------
    gen_dut_rn : FOR RN IN 0 TO g_nof_rn -1 GENERATE
    u_lofar_unb2c_ring : ENTITY work.lofar2_unb2c_ring
    GENERIC MAP (
      g_design_name            => g_design_name,
      g_design_note            => "",
      g_sim                    => c_sim,
      g_sim_unb_nr             => g_unb_nr + (RN / c_quad),
      g_sim_node_nr            => RN MOD c_quad,
      g_sim_sync_timeout       => c_sync_timeout
    )
    PORT MAP (
      -- GENERAL
      CLK          => ext_clk,
      PPS          => pps,
      WDI          => WDI,
      INTA         => INTA,
      INTB         => INTB,
  
      -- Others
      VERSION      => c_version,
      ID           => ( TO_UVEC(RN / c_quad, c_unb2c_board_nof_uniboard_w) & TO_UVEC(RN MOD c_quad, c_unb2c_board_nof_chip_w) ),
      TESTIO       => open,

      -- 1GbE Control Interface
      ETH_CLK      => eth_clk,
      ETH_SGIN     => eth_rxp,
      ETH_SGOUT    => eth_txp,
  
      -- Transceiver clocks
      SA_CLK       => SA_CLK,
      -- front transceivers
      QSFP_0_RX    => i_QSFP_0_RX(RN), 
      QSFP_0_TX    => i_QSFP_0_TX(RN),
   
      -- ring transceivers
      RING_0_RX    => i_RING_0_RX(RN), 
      RING_0_TX    => i_RING_0_TX(RN), 
      RING_1_RX    => i_RING_1_RX(RN), 
      RING_1_TX    => i_RING_1_TX(RN), 
      -- LEDs
      QSFP_LED     => open
  
    );
  END GENERATE;

  -- Ring connections
  gen_ring : FOR I IN 0 TO g_nof_rn -2 GENERATE
    -- Connect consecutive nodes with RING interfaces (PCB)
    i_RING_0_RX(I+1) <= i_RING_1_TX(I);
    i_RING_1_RX(I)   <= i_RING_0_TX(I+1);
  END GENERATE;
  -- Connect first and last nodes with QSFP interface. 
  i_QSFP_0_RX(0)          <= i_QSFP_0_TX(g_nof_rn-1);
  i_QSFP_0_RX(g_nof_rn-1) <= i_QSFP_0_TX(0);
  


  ------------------------------------------------------------------------------
  -- MM peripeheral accesses via file IO
  ------------------------------------------------------------------------------
  p_mm_stimuli : PROCESS
  BEGIN
    -- Wait for DUT power up after reset
    WAIT FOR 1 us;
    
    proc_common_wait_until_hi_lo(ext_clk, ext_pps);

    -- Write ring configuration to all nodes.
    FOR RN IN 0 TO g_nof_rn-1 LOOP
      mmf_mm_bus_wr(mmf_unb_file_prefix(g_unb_nr + (RN / c_quad), RN MOD c_quad) & "REG_RING_INFO", 2, g_nof_rn, tb_clk); -- N_rn
      mmf_mm_bus_wr(mmf_unb_file_prefix(g_unb_nr + (RN / c_quad), RN MOD c_quad) & "REG_RING_INFO", 3, 0,        tb_clk); -- O_rn
    END LOOP;

    -- Start node specific settings
    mmf_mm_bus_wr(mmf_unb_file_prefix(g_unb_nr, 0) & "REG_RING_INFO", 0, 1, tb_clk); -- use_ring_to_previous_rn = 1
    mmf_mm_bus_wr(mmf_unb_file_prefix(g_unb_nr, 0) & "REG_RING_INFO", 1, 0, tb_clk); -- use_ring_to_next_rn = 0
  
    -- End node specific settings
    mmf_mm_bus_wr(mmf_unb_file_prefix(g_unb_nr + ((g_nof_rn-1) / c_quad), (g_nof_rn-1) MOD c_quad) & "REG_RING_INFO", 0, 0, tb_clk); -- use_ring_to_previous_rn = 0
    mmf_mm_bus_wr(mmf_unb_file_prefix(g_unb_nr + ((g_nof_rn-1) / c_quad), (g_nof_rn-1) MOD c_quad) & "REG_RING_INFO", 1, 1, tb_clk); -- use_ring_to_next_rn = 1
     
    ----------------------------------------------------------------------------
    -- Access scheme 1. A source RN creates the packets and sends them along the ring.
    ----------------------------------------------------------------------------
    IF g_access_scheme = 1 THEN
      FOR I IN 0 TO c_nof_lanes-1 LOOP
        -- Select local input (= 1) on  start node on all lanes.
        mmf_mm_bus_wr(c_mm_file_reg_dp_xonoff_lane,  I*2, 0, tb_clk); -- Disable input from lane
        mmf_mm_bus_wr(c_mm_file_reg_dp_xonoff_local, I*2, 1, tb_clk); -- Enable local input
        
        -- Set transport_nof_hops to N_rn on start node for a full transfer around the ring.
        mmf_mm_bus_wr(mmf_unb_file_prefix(g_unb_nr, 0) & "REG_RING_LANE_INFO", I*2+1, g_nof_rn, tb_clk);
      END LOOP;
     
    ----------------------------------------------------------------------------
    -- Access scheme 2, 3. Each RN creates packets and sends them along the ring.
    ----------------------------------------------------------------------------
    ELSE  
      FOR RN IN 0 TO g_nof_rn-1 LOOP
        FOR I IN 0 TO c_nof_lanes-1 LOOP
          -- Select both local and remote input on all nodes on all lanes.
          mmf_mm_bus_wr(mmf_unb_file_prefix(g_unb_nr + (RN / c_quad), RN MOD c_quad) & "REG_DP_XONOFF_LANE",  I*2, 1, tb_clk); -- Enable input from lane
          mmf_mm_bus_wr(mmf_unb_file_prefix(g_unb_nr + (RN / c_quad), RN MOD c_quad) & "REG_DP_XONOFF_LOCAL", I*2, 1, tb_clk); -- Enable local input
          
          -- Set transport_nof_hops to N_rn-1 on all nodes.
          mmf_mm_bus_wr(mmf_unb_file_prefix(g_unb_nr + (RN / c_quad), RN MOD c_quad) & "REG_RING_LANE_INFO", I*2+1, g_nof_rn-1, tb_clk);
        END LOOP;
      END LOOP;
    END IF;
    ----------------------------------------------------------------------------
    -- Enable BG on all nodes (for bs_sosi)
    ----------------------------------------------------------------------------
    FOR RN IN 0 TO g_nof_rn-1 LOOP
      mmf_mm_bus_wr(mmf_unb_file_prefix(g_unb_nr + (RN / c_quad), RN MOD c_quad) & "REG_DIAG_BG", 1,          c_blocksize, tb_clk); -- samples per packet
      mmf_mm_bus_wr(mmf_unb_file_prefix(g_unb_nr + (RN / c_quad), RN MOD c_quad) & "REG_DIAG_BG", 2, g_nof_block_per_sync, tb_clk); -- blocks per sync
      mmf_mm_bus_wr(mmf_unb_file_prefix(g_unb_nr + (RN / c_quad), RN MOD c_quad) & "REG_DIAG_BG", 3,            c_gapsize, tb_clk); -- gapsize
      mmf_mm_bus_wr(mmf_unb_file_prefix(g_unb_nr + (RN / c_quad), RN MOD c_quad) & "REG_DIAG_BG", 0,                    3, tb_clk); -- enable at sync
    END LOOP;
    ----------------------------------------------------------------------------
    -- Verify Access scheme 1 by reading rx / tx monitors on source RN
    ----------------------------------------------------------------------------
    IF g_access_scheme = 1 THEN
      -- Wait for bsn monitor to have received a sync period.
      mmf_mm_wait_until_value(c_mm_file_reg_bsn_monitor_v2_ring_rx, 4, -- read nof valid
                              "SIGNED", rd_data, ">", 0,               -- this is the wait until condition
                              1 us, tb_clk);                           -- read every 1 us

      FOR I IN 0 TO c_nof_lanes-1 LOOP
        mmf_mm_bus_rd(c_mm_file_reg_bsn_monitor_v2_ring_tx, I*c_sdp_N_pn_max*8+1, rd_data, tb_clk); --bsn at sync
        ASSERT TO_UINT(rd_data) = c_exp_bsn_at_sync REPORT "Wrong bsn_at_sync value from bsn_monitor_v2_ring_tx on source node in access scheme 1." SEVERITY ERROR;
        mmf_mm_bus_rd(c_mm_file_reg_bsn_monitor_v2_ring_tx, I*c_sdp_N_pn_max*8+3, rd_data, tb_clk); --nof_sop
        ASSERT TO_UINT(rd_data) = c_exp_nof_sop     REPORT "Wrong nof_sop value from bsn_monitor_v2_ring_tx on source node in access scheme 1."     SEVERITY ERROR;
        mmf_mm_bus_rd(c_mm_file_reg_bsn_monitor_v2_ring_tx, I*c_sdp_N_pn_max*8+4, rd_data, tb_clk); --nof_valid
        ASSERT TO_UINT(rd_data) = c_exp_nof_valid   REPORT "Wrong nof_valid value from bsn_monitor_v2_ring_tx on source node in access scheme 1."   SEVERITY ERROR;
        mmf_mm_bus_rd(c_mm_file_reg_bsn_monitor_v2_ring_tx, I*c_sdp_N_pn_max*8+5, rd_data, tb_clk); --nof_err
        ASSERT TO_UINT(rd_data) = 0                 REPORT "Wrong nof_err value from bsn_monitor_v2_ring_tx on source node in access scheme 1."     SEVERITY ERROR;

        mmf_mm_bus_rd(c_mm_file_reg_bsn_monitor_v2_ring_rx, I*c_sdp_N_pn_max*8+1, rd_data, tb_clk); --bsn at sync
        ASSERT TO_UINT(rd_data) = c_exp_bsn_at_sync REPORT "Wrong bsn_at_sync value from bsn_monitor_v2_ring_rx on source node in access scheme 1." SEVERITY ERROR;
        mmf_mm_bus_rd(c_mm_file_reg_bsn_monitor_v2_ring_rx, I*c_sdp_N_pn_max*8+3, rd_data, tb_clk); --nof_sop
        ASSERT TO_UINT(rd_data) = c_exp_nof_sop     REPORT "Wrong nof_sop value from bsn_monitor_v2_ring_rx on source node in access scheme 1."     SEVERITY ERROR;
        mmf_mm_bus_rd(c_mm_file_reg_bsn_monitor_v2_ring_rx, I*c_sdp_N_pn_max*8+4, rd_data, tb_clk); --nof_valid
        ASSERT TO_UINT(rd_data) = c_exp_nof_valid   REPORT "Wrong nof_valid value from bsn_monitor_v2_ring_rx on source node in access scheme 1."   SEVERITY ERROR;
        mmf_mm_bus_rd(c_mm_file_reg_bsn_monitor_v2_ring_rx, I*c_sdp_N_pn_max*8+5, rd_data, tb_clk); --nof_err
        ASSERT TO_UINT(rd_data) = 0                 REPORT "Wrong nof_err value from bsn_monitor_v2_ring_rx on source node in access scheme 1."     SEVERITY ERROR;
      END LOOP;

    ----------------------------------------------------------------------------
    -- Verify Access scheme 2,3 by reading rx / tx monitors on all RN
    ----------------------------------------------------------------------------
    ELSE
    -- Wait for bsn monitor to have received a sync period.
    mmf_mm_wait_until_value(mmf_unb_file_prefix(g_unb_nr + ((g_nof_rn-1) / c_quad), (g_nof_rn-1) MOD c_quad) & "REG_BSN_MONITOR_V2_RING_RX", 4, -- read nof valid
                            "SIGNED", rd_data, ">", 0,               -- this is the wait until condition
                            1 us, tb_clk);                           -- read every 1 us

      FOR RN IN 0 TO g_nof_rn-1 LOOP
        FOR I IN 0 TO c_nof_lanes-1 LOOP -- lane index
          FOR J IN 0 TO g_nof_rn-1 LOOP  -- bsn_monitor index
            -- No packets transmitted from next RN (this_rn + 1 for even lanes, this_rn - 1 for odd lanes) as this RN should have removed it from the ring.
            IF (I MOD 2 = 0 AND (RN + 1) MOD g_nof_rn = J) OR (I MOD 2 = 1 AND (RN + g_nof_rn-1) MOD g_nof_rn = J) THEN 
              mmf_mm_bus_rd(mmf_unb_file_prefix(g_unb_nr + (RN / c_quad), RN MOD c_quad) & "REG_BSN_MONITOR_V2_RING_TX", (I*c_sdp_N_pn_max + J) * 8+0, rd_data, tb_clk); --status bits
              ASSERT rd_data(2) = '1'                     REPORT "Wrong sync_timout, expected 1, got 0. From bsn_monitor_v2_ring_tx on RN_" & INTEGER'IMAGE(RN) & " in access scheme 2/3." SEVERITY ERROR;          
            ELSE
              mmf_mm_bus_rd(mmf_unb_file_prefix(g_unb_nr + (RN / c_quad), RN MOD c_quad) & "REG_BSN_MONITOR_V2_RING_TX", (I*c_sdp_N_pn_max + J) * 8+0, rd_data, tb_clk); --status bits
              ASSERT rd_data(2) = '0'                     REPORT "Wrong sync_timout, expected 0, got 1. From bsn_monitor_v2_ring_tx on RN_" & INTEGER'IMAGE(RN) & " in access scheme 2/3." SEVERITY ERROR;  
              mmf_mm_bus_rd(mmf_unb_file_prefix(g_unb_nr + (RN / c_quad), RN MOD c_quad) & "REG_BSN_MONITOR_V2_RING_TX", (I*c_sdp_N_pn_max + J) * 8+1, rd_data, tb_clk); --bsn at sync
              ASSERT TO_UINT(rd_data) = c_exp_bsn_at_sync REPORT "Wrong bsn_at_sync value from bsn_monitor_v2_ring_tx on RN_" & INTEGER'IMAGE(RN) & " in access scheme 2/3." SEVERITY ERROR;
              mmf_mm_bus_rd(mmf_unb_file_prefix(g_unb_nr + (RN / c_quad), RN MOD c_quad) & "REG_BSN_MONITOR_V2_RING_TX", (I*c_sdp_N_pn_max + J) * 8+3, rd_data, tb_clk); --nof_sop
              ASSERT TO_UINT(rd_data) = c_exp_nof_sop     REPORT "Wrong nof_sop value from bsn_monitor_v2_ring_tx on RN_"     & INTEGER'IMAGE(RN) & " in access scheme 2/3." SEVERITY ERROR;
              mmf_mm_bus_rd(mmf_unb_file_prefix(g_unb_nr + (RN / c_quad), RN MOD c_quad) & "REG_BSN_MONITOR_V2_RING_TX", (I*c_sdp_N_pn_max + J) * 8+4, rd_data, tb_clk); --nof_valid
              ASSERT TO_UINT(rd_data) = c_exp_nof_valid   REPORT "Wrong nof_valid value from bsn_monitor_v2_ring_tx on RN_"   & INTEGER'IMAGE(RN) & " in access scheme 2/3." SEVERITY ERROR;
              mmf_mm_bus_rd(mmf_unb_file_prefix(g_unb_nr + (RN / c_quad), RN MOD c_quad) & "REG_BSN_MONITOR_V2_RING_TX", (I*c_sdp_N_pn_max + J) * 8+5, rd_data, tb_clk); --nof_err
              ASSERT TO_UINT(rd_data) = 0                 REPORT "Wrong nof_err value from bsn_monitor_v2_ring_tx on RN_"     & INTEGER'IMAGE(RN) & " in access scheme 2/3." SEVERITY ERROR;
            END IF; 
            IF RN = J THEN -- No packets received from itself as the previous RN should have removed it from the ring.
              mmf_mm_bus_rd(mmf_unb_file_prefix(g_unb_nr + (RN / c_quad), RN MOD c_quad) & "REG_BSN_MONITOR_V2_RING_RX", (I*c_sdp_N_pn_max + J) * 8+0, rd_data, tb_clk); --status bits
              ASSERT rd_data(2) = '1'                     REPORT "Wrong sync_timout, expected 1, got 0. From bsn_monitor_v2_ring_rx on RN_" & INTEGER'IMAGE(RN) & " in access scheme 2/3." SEVERITY ERROR;          
            ELSE
              mmf_mm_bus_rd(mmf_unb_file_prefix(g_unb_nr + (RN / c_quad), RN MOD c_quad) & "REG_BSN_MONITOR_V2_RING_RX", (I*c_sdp_N_pn_max + J) * 8+0, rd_data, tb_clk); --status bits
              ASSERT rd_data(2) = '0'                     REPORT "Wrong sync_timout, expected 0, got 1. From bsn_monitor_v2_ring_rx on RN_" & INTEGER'IMAGE(RN) & " in access scheme 2/3." SEVERITY ERROR;  
              mmf_mm_bus_rd(mmf_unb_file_prefix(g_unb_nr + (RN / c_quad), RN MOD c_quad) & "REG_BSN_MONITOR_V2_RING_RX", (I*c_sdp_N_pn_max + J) * 8+1, rd_data, tb_clk); --bsn at sync
              ASSERT TO_UINT(rd_data) = c_exp_bsn_at_sync REPORT "Wrong bsn_at_sync value from bsn_monitor_v2_ring_rx on RN_" & INTEGER'IMAGE(RN) & " in access scheme 2/3." SEVERITY ERROR;
              mmf_mm_bus_rd(mmf_unb_file_prefix(g_unb_nr + (RN / c_quad), RN MOD c_quad) & "REG_BSN_MONITOR_V2_RING_RX", (I*c_sdp_N_pn_max + J) * 8+3, rd_data, tb_clk); --nof_sop
              ASSERT TO_UINT(rd_data) = c_exp_nof_sop     REPORT "Wrong nof_sop value from bsn_monitor_v2_ring_rx on RN_"     & INTEGER'IMAGE(RN) & " in access scheme 2/3." SEVERITY ERROR;
              mmf_mm_bus_rd(mmf_unb_file_prefix(g_unb_nr + (RN / c_quad), RN MOD c_quad) & "REG_BSN_MONITOR_V2_RING_RX", (I*c_sdp_N_pn_max + J) * 8+4, rd_data, tb_clk); --nof_valid
              ASSERT TO_UINT(rd_data) = c_exp_nof_valid   REPORT "Wrong nof_valid value from bsn_monitor_v2_ring_rx on RN_"   & INTEGER'IMAGE(RN) & " in access scheme 2/3." SEVERITY ERROR;
              mmf_mm_bus_rd(mmf_unb_file_prefix(g_unb_nr + (RN / c_quad), RN MOD c_quad) & "REG_BSN_MONITOR_V2_RING_RX", (I*c_sdp_N_pn_max + J) * 8+5, rd_data, tb_clk); --nof_err
              ASSERT TO_UINT(rd_data) = 0                 REPORT "Wrong nof_err value from bsn_monitor_v2_ring_rx on RN_"     & INTEGER'IMAGE(RN) & " in access scheme 2/3." SEVERITY ERROR;
            END IF;
          END LOOP;
        END LOOP;
      END LOOP;
    END IF;

    ---------------------------------------------------------------------------
    -- End Simulation 
    ---------------------------------------------------------------------------   
    sim_done <= '1';
    proc_common_wait_some_cycles(ext_clk, 100);
    proc_common_stop_simulation(NOT g_multi_tb, ext_clk, sim_done, i_tb_end);
    WAIT;
  END PROCESS;

  tb_end <= i_tb_end;
END tb;
