-------------------------------------------------------------------------------
--
-- Copyright 2021
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-------------------------------------------------------------------------------

-- Authors : J Hargreaves, L Hiemstra, R van der Walle
-- Purpose:  
--   AIT - ADC (Jesd) receiver, input, timing and associated diagnostic blocks
-- Description:
--   Unb2b version for lab testing
--   Contains all the signal processing blocks to receive and time the ADC input data
--   See https://support.astron.nl/confluence/display/STAT/L5+SDPFW+DD%3A+ADC+data+input+and+timestamp

LIBRARY IEEE, common_lib, diag_lib, aduh_lib, dp_lib, tech_jesd204b_lib, st_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE IEEE.NUMERIC_STD.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE diag_lib.diag_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE work.sdp_pkg.ALL;

ENTITY node_sdp_adc_input_and_timing IS
  GENERIC (
    g_no_jesd                 : BOOLEAN := FALSE;
    g_buf_nof_data            : NATURAL := c_sdp_V_si_db;
    g_bsn_nof_clk_per_sync    : NATURAL := c_sdp_N_clk_per_sync;  -- Default 200M, overide for short simulation
    g_sim                     : BOOLEAN := FALSE  
  );
  PORT (
    -- clocks and resets
    mm_clk                         : IN STD_LOGIC;
    mm_rst                         : IN STD_LOGIC;
    dp_clk                         : IN STD_LOGIC;
    dp_rst                         : IN STD_LOGIC;
    dp_pps                         : IN STD_LOGIC := '0';

    -- mm control buses
    -- JESD 
    jesd204b_mosi                  : IN  t_mem_mosi := c_mem_mosi_rst;
    jesd204b_miso                  : OUT t_mem_miso := c_mem_miso_rst; 
 
    -- Shiftram (applies per-antenna delay)
    reg_dp_shiftram_mosi           : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_dp_shiftram_miso           : OUT t_mem_miso := c_mem_miso_rst;

    -- bsn source
    reg_bsn_source_v2_mosi         : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_bsn_source_v2_miso         : OUT t_mem_miso := c_mem_miso_rst;

    -- bsn scheduler
    reg_bsn_scheduler_wg_mosi      : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_bsn_scheduler_wg_miso      : OUT t_mem_miso := c_mem_miso_rst;

    -- WG
    reg_wg_mosi                    : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_wg_miso                    : OUT t_mem_miso := c_mem_miso_rst;
    ram_wg_mosi                    : IN  t_mem_mosi := c_mem_mosi_rst;
    ram_wg_miso                    : OUT t_mem_miso := c_mem_miso_rst;

    -- BSN MONITOR
    reg_bsn_monitor_input_mosi     : IN  t_mem_mosi;
    reg_bsn_monitor_input_miso     : OUT t_mem_miso;

    -- Data buffer for framed samples (variable depth)
    ram_diag_data_buf_bsn_mosi     : IN  t_mem_mosi;
    ram_diag_data_buf_bsn_miso     : OUT t_mem_miso;
    reg_diag_data_buf_bsn_mosi     : IN  t_mem_mosi;
    reg_diag_data_buf_bsn_miso     : OUT t_mem_miso;

    -- ST Histogram
    ram_st_histogram_mosi          : IN  t_mem_mosi;
    ram_st_histogram_miso          : OUT t_mem_miso;

    -- Aduh (statistics) monitor
    reg_aduh_monitor_mosi          : IN  t_mem_mosi;
    reg_aduh_monitor_miso          : OUT t_mem_miso;

    -- JESD control
    jesd_ctrl_mosi                 : IN  t_mem_mosi;
    jesd_ctrl_miso                 : OUT t_mem_miso;

    -- JESD io signals
    jesd204b_serial_data           : IN    STD_LOGIC_VECTOR(c_sdp_S_pn-1 downto 0) := (OTHERS => '0'); 
    jesd204b_refclk                : IN    STD_LOGIC := '0'; 
    jesd204b_sysref                : IN    STD_LOGIC := '0';
    jesd204b_sync_n                : OUT   STD_LOGIC_VECTOR(c_sdp_N_sync_jesd - 1 DOWNTO 0) := (OTHERS => '0');

    -- Streaming data output
    out_sosi_arr                   : OUT t_dp_sosi_arr(c_sdp_S_pn-1 DOWNTO 0);        
    dp_bsn_source_restart          : OUT STD_LOGIC

  );
END node_sdp_adc_input_and_timing;


ARCHITECTURE str OF node_sdp_adc_input_and_timing IS

  -- Waveform Generator
  CONSTANT c_wg_buf_directory       : STRING := "data/";
  CONSTANT c_wg_buf_dat_w           : NATURAL := 18; --default value of WG that fits 14 bits of ADC data
  CONSTANT c_wg_buf_addr_w          : NATURAL := 10; --default value of WG for 1024 samples;
  SIGNAL trigger_wg                 : STD_LOGIC;

  -- Frame parameters
  CONSTANT c_bs_sync_timeout        : NATURAL := g_bsn_nof_clk_per_sync + g_bsn_nof_clk_per_sync / 10;  -- +10% margin
  CONSTANT c_bs_bsn_w               : NATURAL := 64; -- > 51;
  CONSTANT c_bs_block_size          : NATURAL := c_sdp_N_fft; -- =1024;
  CONSTANT c_dp_fifo_dc_size        : NATURAL := 64;

  -- JESD signals
  SIGNAL rx_clk                     : STD_LOGIC; -- formerly jesd204b_frame_clk
  SIGNAL rx_rst                     : STD_LOGIC; 
  SIGNAL rx_sysref                  : STD_LOGIC; 

  SIGNAL rx_bsn_source_restart      : STD_LOGIC;

  -- Sosis and sosi arrays
  SIGNAL rx_sosi_arr                : t_dp_sosi_arr(c_sdp_S_pn-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);         
  SIGNAL dp_shiftram_snk_in_arr     : t_dp_sosi_arr(c_sdp_S_pn-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);         
  SIGNAL ant_sosi_arr               : t_dp_sosi_arr(c_sdp_S_pn-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);
  SIGNAL bs_sosi                    : t_dp_sosi := c_dp_sosi_rst;    
  SIGNAL wg_sosi_arr                : t_dp_sosi_arr(c_sdp_S_pn-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);    
  SIGNAL mux_sosi_arr               : t_dp_sosi_arr(c_sdp_S_pn-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);         
  SIGNAL nxt_mux_sosi_arr           : t_dp_sosi_arr(c_sdp_S_pn-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);
  SIGNAL st_sosi_arr                : t_dp_sosi_arr(c_sdp_S_pn-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);    

  SIGNAL mm_rst_internal            : STD_LOGIC; 
  SIGNAL mm_jesd_ctrl_reg           : STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);
  SIGNAL jesd204b_disable_arr       : STD_LOGIC_VECTOR(c_sdp_S_pn-1 DOWNTO 0);

BEGIN

  gen_no_jesd : IF g_no_jesd = TRUE GENERATE
    rx_clk <= dp_clk;
    rx_rst <= dp_rst;
    rx_sysref <= dp_pps;
  END GENERATE;

  -- The node AIT is reset at power up by mm_rst and under software control by jesd204b_disable_arr.
  -- The mm_rst internal will cause a reset on the rx_rst by the reset sequencer in the u_jesd204b.
  -- The MM jesd204b_disable_arr is intended for node AIT resynchronisation tests of the u_jesd204b.
  -- The MM jesd204b_disable_arr should not be applied in an SDP application, because this will cause
  -- a disturbance in the block timing of the out_sosi_arr(i).sync,bsn,sop,eop. The other logic
  -- in an SDP application assumes that the block timing of the out_sosi_arr(i) only contains
  -- complete blocks, so from sop to eop.

  mm_rst_internal <= mm_rst OR mm_jesd_ctrl_reg(31);
  gen_jesd_disable : FOR I IN 0 TO c_sdp_S_pn-1 GENERATE
    jesd204b_disable_arr(i) <= mm_jesd_ctrl_reg(i);
  END GENERATE;

  gen_jesd : IF g_no_jesd = FALSE GENERATE
    -----------------------------------------------------------------------------
    -- JESD204B IP (ADC Handler)
    -----------------------------------------------------------------------------
    
    u_jesd204b: ENTITY tech_jesd204b_lib.tech_jesd204b 
    GENERIC MAP(
      g_sim                => g_sim,               
      g_nof_streams        => c_sdp_S_pn,
      g_nof_sync_n         => c_sdp_N_sync_jesd,
      g_jesd_freq          => c_sdp_jesd204b_freq
    )
    PORT MAP(
      jesd204b_refclk      => JESD204B_REFCLK,   
      jesd204b_sysref      => JESD204B_SYSREF,   
      jesd204b_sync_n_arr  => jesd204b_sync_n,   
    
      rx_sosi_arr          => rx_sosi_arr,          
      rx_clk               => rx_clk,          
      rx_rst               => rx_rst,          
      rx_sysref            => rx_sysref,          
  
      jesd204b_disable_arr  => jesd204b_disable_arr,
    
      -- MM
      mm_clk               => mm_clk,           
      mm_rst               => mm_rst_internal,           
    
      jesd204b_mosi        => jesd204b_mosi,         
      jesd204b_miso        => jesd204b_miso,         
    
       -- Serial
      serial_tx_arr        => open,
      serial_rx_arr        => JESD204B_SERIAL_DATA(c_sdp_S_pn-1 downto 0)
    );
  
  
    -----------------------------------------------------------------------------
    -- Time delay: dp_shiftram
    -- . copied from unb1_bn_capture_input (apertif)
    --   Array range reversal is not done because everything is DOWNTO
    -- . the input valid is always '1', even when there is no data 
    -----------------------------------------------------------------------------
    
    p_dp_shiftram_snk_in_arr : PROCESS(rx_sosi_arr)
    BEGIN
      dp_shiftram_snk_in_arr <= rx_sosi_arr;
      FOR I IN 0 TO c_sdp_S_pn-1 LOOP
        -- ADC data is stored in the upper 14 bits of the jesd rx_sosi.
        dp_shiftram_snk_in_arr(I).data    <= RESIZE_DP_SDATA(rx_sosi_arr(I).data(c_sdp_W_adc_jesd-1 DOWNTO (c_sdp_W_adc_jesd - c_sdp_W_adc) ));
        -- Force valid.
        dp_shiftram_snk_in_arr(I).valid   <= '1';
      END LOOP;
    END PROCESS;
  
  
    u_dp_shiftram : ENTITY dp_lib.dp_shiftram
    GENERIC MAP (
      g_nof_streams => c_sdp_S_pn, 
      g_nof_words   => c_sdp_V_sample_delay,
      g_data_w      => c_sdp_W_adc, 
      g_use_sync_in => TRUE
    )
    PORT MAP (
      dp_rst   => rx_rst,
      dp_clk   => rx_clk,
  
      mm_rst   => mm_rst_internal,
      mm_clk   => mm_clk,
  
      sync_in  => bs_sosi.sync,
  
      reg_mosi => reg_dp_shiftram_mosi,
      reg_miso => reg_dp_shiftram_miso,
  
      snk_in_arr => dp_shiftram_snk_in_arr,
  
      src_out_arr => ant_sosi_arr
    );
  END GENERATE;

  -----------------------------------------------------------------------------
  -- Timestamp
  -----------------------------------------------------------------------------  
  u_bsn_source_v2 : ENTITY dp_lib.mms_dp_bsn_source_v2
  GENERIC MAP (
    g_cross_clock_domain     => TRUE,
    g_block_size             => c_bs_block_size,
    g_nof_clk_per_sync       => g_bsn_nof_clk_per_sync,
    g_bsn_w                  => c_bs_bsn_w
  )
  PORT MAP (
    -- Clocks and reset
    mm_rst            => mm_rst_internal,
    mm_clk            => mm_clk,
    dp_rst            => rx_rst,
    dp_clk            => rx_clk,
    dp_pps            => rx_sysref,
    
    -- Memory-mapped clock domain
    reg_mosi          => reg_bsn_source_v2_mosi,
    reg_miso          => reg_bsn_source_v2_miso,
    
    -- Streaming clock domain
    bs_sosi           => bs_sosi,

    bs_restart        => rx_bsn_source_restart
  );

  u_bsn_trigger_wg : ENTITY dp_lib.mms_dp_bsn_scheduler
  GENERIC MAP (
    g_cross_clock_domain => TRUE,
    g_bsn_w              => c_bs_bsn_w
  )
  PORT MAP (
    -- Memory-mapped clock domain
    mm_rst      => mm_rst_internal,
    mm_clk      => mm_clk,

    reg_mosi    => reg_bsn_scheduler_wg_mosi,
    reg_miso    => reg_bsn_scheduler_wg_miso,

    -- Streaming clock domain
    dp_rst      => rx_rst,
    dp_clk      => rx_clk,

    snk_in      => bs_sosi,  -- only uses eop (= block sync), bsn[]
    trigger_out => trigger_wg
  );


  -----------------------------------------------------------------------------
  -- WG (Test Signal Generator)
  -----------------------------------------------------------------------------

  u_wg_arr : ENTITY diag_lib.mms_diag_wg_wideband_arr
  GENERIC MAP (
    g_nof_streams        => c_sdp_S_pn, 
    g_cross_clock_domain => TRUE,
    g_buf_dir            => c_wg_buf_directory,

    -- Wideband parameters
    g_wideband_factor    => 1, 

    -- Basic WG parameters, see diag_wg.vhd for their meaning
    g_buf_dat_w          => c_wg_buf_dat_w,
    g_buf_addr_w         => c_wg_buf_addr_w,
    g_calc_support       => TRUE,
    g_calc_gain_w        => 1,
    g_calc_dat_w         => c_sdp_W_adc
  )
  PORT MAP (
    -- Memory-mapped clock domain
    mm_rst              => mm_rst_internal,
    mm_clk              => mm_clk,

    reg_mosi            => reg_wg_mosi,
    reg_miso            => reg_wg_miso,

    buf_mosi            => ram_wg_mosi,
    buf_miso            => ram_wg_miso,

    -- Streaming clock domain
    st_rst              => rx_rst,
    st_clk              => rx_clk,
    st_restart          => trigger_wg,

    out_sosi_arr        => wg_sosi_arr
  );


  -----------------------------------------------------------------------------
  -- ADC/WG Mux (Input Select)
  -----------------------------------------------------------------------------
  
  gen_mux : FOR I IN 0 TO c_sdp_S_pn-1 GENERATE
    p_sosi : PROCESS(ant_sosi_arr(I), wg_sosi_arr(I))
    BEGIN
      -- Default use the ADC data
      nxt_mux_sosi_arr(I).data  <= ant_sosi_arr(I).data;
      IF wg_sosi_arr(I).valid='1' THEN
        -- Valid WG data overrules ADC data
        nxt_mux_sosi_arr(I).data <= wg_sosi_arr(I).data;
      END IF;
    END PROCESS;
  END GENERATE;

  mux_sosi_arr  <= nxt_mux_sosi_arr WHEN rising_edge(rx_clk);

  -----------------------------------------------------------------------------
  -- Concatenate muxed data streams with bsn framing
  -----------------------------------------------------------------------------
  
  gen_concat : FOR I IN 0 TO c_sdp_S_pn-1 GENERATE
    p_sosi : PROCESS(mux_sosi_arr(I), bs_sosi)
    BEGIN
      st_sosi_arr(I)       <= bs_sosi;
      st_sosi_arr(I).data  <= mux_sosi_arr(I).data;
    END PROCESS;
  END GENERATE;


  ---------------------------------------------------------------------------------------
  -- Diagnostics on the bsn-framed data
  --   . BSN Monitor (ToDo: can be removed as not part of the spec)
  --   . Aduh monitor
  --   . Data Buffer (variable depth from 1k-128k)
  ---------------------------------------------------------------------------------------
  

  ---------------------------------------------------------------------------------------
  -- BSN monitor (Block Checker)
  ---------------------------------------------------------------------------------------
  u_bsn_monitor : ENTITY dp_lib.mms_dp_bsn_monitor
  GENERIC MAP (
    g_nof_streams        => 1,  -- They're all the same
    g_sync_timeout       => c_bs_sync_timeout,
    g_bsn_w              => c_bs_bsn_w,
    g_log_first_bsn      => FALSE
  )
  PORT MAP (
    -- Memory-mapped clock domain
    mm_rst      => mm_rst_internal,
    mm_clk      => mm_clk,
    reg_mosi    => reg_bsn_monitor_input_mosi,
    reg_miso    => reg_bsn_monitor_input_miso,

    -- Streaming clock domain
    dp_rst      => rx_rst,
    dp_clk      => rx_clk,
    in_sosi_arr => st_sosi_arr(0 downto 0)
  );


  -----------------------------------------------------------------------------
  -- Monitor ADU/WG output
  -----------------------------------------------------------------------------
  u_aduh_monitor : ENTITY aduh_lib.mms_aduh_monitor_arr
  GENERIC MAP (
    g_cross_clock_domain   => TRUE,
    g_nof_streams          => c_sdp_S_pn,
    g_symbol_w             => c_sdp_W_adc,  
    g_nof_symbols_per_data => 1,          -- Wideband factor is 1          
    g_nof_accumulations    => g_bsn_nof_clk_per_sync
  )
  PORT MAP (
    -- Memory-mapped clock domain
    mm_rst         => mm_rst_internal,
    mm_clk         => mm_clk,

    reg_mosi       => reg_aduh_monitor_mosi, -- read only access to the signal path data mean sum and power sum registers
    reg_miso       => reg_aduh_monitor_miso,
    buf_mosi       => c_mem_mosi_rst,         -- Unused
    buf_miso       => OPEN,

    -- Streaming clock domain
    st_rst         => rx_rst,
    st_clk         => rx_clk,

    in_sosi_arr    => st_sosi_arr
  );


  -----------------------------------------------------------------------------
  -- Diagnostic Data Buffer
  -----------------------------------------------------------------------------

  u_diag_data_buffer_bsn : ENTITY diag_lib.mms_diag_data_buffer
  GENERIC MAP (
    g_nof_streams  => c_sdp_S_pn,
    g_data_w       => c_sdp_W_adc,
    g_buf_nof_data => g_buf_nof_data,
    g_buf_use_sync => TRUE -- when TRUE start filling the buffer at the in_sync, else after the last word was read
  )
  PORT MAP (
    mm_rst            => mm_rst_internal,
    mm_clk            => mm_clk,
    dp_rst            => rx_rst,
    dp_clk            => rx_clk,

    ram_data_buf_mosi => ram_diag_data_buf_bsn_mosi,
    ram_data_buf_miso => ram_diag_data_buf_bsn_miso,
    reg_data_buf_mosi => reg_diag_data_buf_bsn_mosi,
    reg_data_buf_miso => reg_diag_data_buf_bsn_miso,

    in_sosi_arr       => st_sosi_arr,
    in_sync           => st_sosi_arr(0).sync
  );

  -----------------------------------------------------------------------------
  -- ST Histogram
  -----------------------------------------------------------------------------
  u_st_histogram : ENTITY st_lib.mmp_st_histogram
  GENERIC MAP (
    g_nof_instances          => c_sdp_S_pn,
    g_data_w                 => c_sdp_W_adc,
    g_nof_bins               => c_sdp_V_si_histogram,
    g_nof_data_per_sync      => g_bsn_nof_clk_per_sync,
    g_nof_data_per_sync_diff => c_sdp_N_fft/2 
  )
  PORT MAP (
    mm_rst            => mm_rst_internal,
    mm_clk            => mm_clk,
    dp_rst            => rx_rst,
    dp_clk            => rx_clk,

    ram_copi          => ram_st_histogram_mosi,
    ram_cipo          => ram_st_histogram_miso,

    snk_in_arr        => st_sosi_arr
  );


  -----------------------------------------------------------------------------
  -- Output Stage
  --   . Thin dual clock fifo to cross from jesd frame clock (rx_clk) to dp_clk domain
  -----------------------------------------------------------------------------
  u_dp_fifo_dc_arr : ENTITY dp_lib.dp_fifo_dc_arr
    GENERIC MAP (
      g_nof_streams    => c_sdp_S_pn,
      g_data_w         => c_sdp_W_adc,
      g_data_signed    => TRUE,
      g_bsn_w          => c_bs_bsn_w,
      g_use_empty      => FALSE,
      g_use_ctrl       => TRUE,
      g_use_sync       => TRUE,
      g_use_bsn        => TRUE,
      g_use_aux        => TRUE,
      g_fifo_size      => c_dp_fifo_dc_size
    )
    PORT MAP (
      wr_rst           => rx_rst,
      wr_clk           => rx_clk,
      rd_rst           => dp_rst,
      rd_clk           => dp_clk,
      snk_in_arr       => st_sosi_arr,
      src_out_arr      => out_sosi_arr,
      in_aux(0)        => rx_bsn_source_restart,
      out_aux(0)       => dp_bsn_source_restart
    );

  
-----------------------------------------------------------------------------
  -- JESD Control register
  -----------------------------------------------------------------------------
  u_mm_jesd_ctrl_reg : ENTITY common_lib.common_reg_r_w
  GENERIC MAP (
    g_reg       => c_sdp_jesd204b_mm_jesd_ctrl_reg,
    g_init_reg  => (OTHERS => '0')
  )
  PORT MAP (
    rst       => mm_rst,
    clk       => mm_clk,
    -- control side
    wr_en     => jesd_ctrl_mosi.wr,
    wr_adr    => jesd_ctrl_mosi.address(c_sdp_jesd204b_mm_jesd_ctrl_reg.adr_w-1 DOWNTO 0),
    wr_dat    => jesd_ctrl_mosi.wrdata(c_sdp_jesd204b_mm_jesd_ctrl_reg.dat_w-1 DOWNTO 0),
    rd_en     => jesd_ctrl_mosi.rd,
    rd_adr    => jesd_ctrl_mosi.address(c_sdp_jesd204b_mm_jesd_ctrl_reg.adr_w-1 DOWNTO 0),
    rd_dat    => jesd_ctrl_miso.rddata(c_sdp_jesd204b_mm_jesd_ctrl_reg.dat_w-1 DOWNTO 0),
    rd_val    => OPEN,
    -- data side
    out_reg   => mm_jesd_ctrl_reg,
    in_reg    => mm_jesd_ctrl_reg
  );

END str;
