-------------------------------------------------------------------------------
--
-- Copyright 2021
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--
-- Author: R. van der Walle
-- Purpose: 
-- . Implements the functionality of the Subband Correlator in the 
--   LOFAR2 SDPFW design.
-- Description:
-- Remark:
-- .
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib, reorder_lib, st_lib, mm_lib, ring_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_network_layers_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE ring_lib.ring_pkg.ALL;
USE work.sdp_pkg.ALL;

ENTITY node_sdp_correlator IS
  GENERIC (
    g_sim            : BOOLEAN := FALSE;
    g_sim_sdp        : t_sdp_sim := c_sdp_sim;
    g_P_sq           : NATURAL := c_sdp_P_sq
  );
  PORT (
    dp_clk        : IN  STD_LOGIC;
    dp_rst        : IN  STD_LOGIC;

    in_sosi_arr   : IN  t_dp_sosi_arr(c_sdp_P_pfb-1 DOWNTO 0);
    xst_udp_sosi  : OUT t_dp_sosi;
    xst_udp_siso  : IN  t_dp_siso;
    from_ri_sosi  : IN  t_dp_sosi := c_dp_sosi_rst;
    to_ri_sosi    : OUT t_dp_sosi;
    bs_sosi       : OUT t_dp_sosi;
    mm_rst        : IN  STD_LOGIC;
    mm_clk        : IN  STD_LOGIC;

    reg_bsn_sync_scheduler_xsub_copi         : IN  t_mem_copi := c_mem_copi_rst;
    reg_bsn_sync_scheduler_xsub_cipo         : OUT t_mem_cipo;
    reg_crosslets_info_copi                  : IN  t_mem_copi := c_mem_copi_rst;
    reg_crosslets_info_cipo                  : OUT t_mem_cipo;
    reg_nof_crosslets_copi                   : IN  t_mem_copi := c_mem_copi_rst;
    reg_nof_crosslets_cipo                   : OUT t_mem_cipo;
    ram_st_xsq_copi                          : IN  t_mem_copi := c_mem_copi_rst;
    ram_st_xsq_cipo                          : OUT t_mem_cipo;
    reg_stat_enable_copi                     : IN  t_mem_copi := c_mem_copi_rst;
    reg_stat_enable_cipo                     : OUT t_mem_cipo;    
    reg_stat_hdr_dat_copi                    : IN  t_mem_copi := c_mem_copi_rst;
    reg_stat_hdr_dat_cipo                    : OUT t_mem_cipo;
    reg_bsn_align_copi                       : IN  t_mem_copi := c_mem_copi_rst;
    reg_bsn_align_cipo                       : OUT t_mem_cipo;
    reg_bsn_monitor_v2_bsn_align_input_copi  : IN  t_mem_copi := c_mem_copi_rst;
    reg_bsn_monitor_v2_bsn_align_input_cipo  : OUT t_mem_cipo;    
    reg_bsn_monitor_v2_bsn_align_output_copi : IN  t_mem_copi := c_mem_copi_rst;
    reg_bsn_monitor_v2_bsn_align_output_cipo : OUT t_mem_cipo;
    reg_bsn_monitor_v2_xst_offload_copi      : IN  t_mem_copi := c_mem_copi_rst;
    reg_bsn_monitor_v2_xst_offload_cipo      : OUT t_mem_cipo;

    sdp_info          : IN t_sdp_info;
    ring_info         : IN t_ring_info;
    gn_id             : IN STD_LOGIC_VECTOR(c_sdp_W_gn_id-1 DOWNTO 0);
    stat_eth_src_mac  : IN STD_LOGIC_VECTOR(c_network_eth_mac_addr_w-1 DOWNTO 0);
    stat_ip_src_addr  : IN STD_LOGIC_VECTOR(c_network_ip_addr_w-1 DOWNTO 0);
    stat_udp_src_port : IN STD_LOGIC_VECTOR(c_network_udp_port_w-1 DOWNTO 0)
  );
END node_sdp_correlator;

ARCHITECTURE str OF node_sdp_correlator IS
   
  CONSTANT c_nof_controllers      : POSITIVE := 2;
  CONSTANT c_block_size           : NATURAL  := c_sdp_N_crosslets_max * c_sdp_S_pn;
  CONSTANT c_block_size_longwords : NATURAL  := ceil_div(c_block_size, 2); -- 32b -> 64b
  CONSTANT c_data_w               : NATURAL  := c_sdp_W_crosslet * c_nof_complex;

-- The size for 1 block is probably already enough as the number of blocks received
-- on the remote input of the mux probably have enough gap time in between. Just
-- to be sure to not run into issues in the future, the fifo size is increased to 
-- buffer the maximum nof blocks per block period.
  CONSTANT c_mux_fifo_size   : NATURAL  := 2**ceil_log2(g_P_sq * c_block_size_longwords); 
-- c_fifo_fill_size should be at least 2 * c_block_size_longwords as dp_repack_data
-- repacks from 64bit to 32bit. Chosing 3x to have some room.
  CONSTANT c_fifo_fill_size  : NATURAL  := 2**ceil_log2(3 * c_block_size_longwords); 

-- crosslet statistics offload
  SIGNAL ram_st_offload_copi           : t_mem_copi := c_mem_copi_rst;
  SIGNAL ram_st_offload_cipo           : t_mem_cipo := c_mem_cipo_rst;

  SIGNAL controller_mem_mux_copi       : t_mem_copi := c_mem_copi_rst;
  SIGNAL controller_mem_mux_cipo       : t_mem_cipo := c_mem_cipo_rst;
  SIGNAL controller_copi_arr           : t_mem_copi_arr(0 TO c_nof_controllers-1) := (OTHERS=>c_mem_copi_rst);
  SIGNAL controller_cipo_arr           : t_mem_cipo_arr(0 TO c_nof_controllers-1) := (OTHERS=>c_mem_cipo_rst);

  SIGNAL quant_sosi_arr                : t_dp_sosi_arr(c_sdp_P_pfb-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);
  SIGNAL dp_bsn_sync_scheduler_src_out : t_dp_sosi := c_dp_sosi_rst;
  SIGNAL xsel_sosi                     : t_dp_sosi := c_dp_sosi_rst;
  SIGNAL xsel_data_sosi                : t_dp_sosi := c_dp_sosi_rst;
  SIGNAL local_sosi                    : t_dp_sosi := c_dp_sosi_rst;

  SIGNAL ring_mux_sosi                 : t_dp_sosi := c_dp_sosi_rst;
  SIGNAL ring_mux_siso                 : t_dp_siso := c_dp_siso_rdy;
  SIGNAL dp_fifo_fill_sosi             : t_dp_sosi := c_dp_sosi_rst;
  SIGNAL dp_fifo_fill_siso             : t_dp_siso := c_dp_siso_rdy;
  SIGNAL rx_sosi                       : t_dp_sosi := c_dp_sosi_rst;
  SIGNAL dispatch_invert_sosi_arr      : t_dp_sosi_arr(0 TO g_P_sq-1) := (OTHERS => c_dp_sosi_rst);
  SIGNAL dispatch_sosi_arr             : t_dp_sosi_arr(g_P_sq-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);

  SIGNAL crosslets_sosi                : t_dp_sosi  := c_dp_sosi_rst;
  SIGNAL crosslets_copi                : t_mem_copi := c_mem_copi_rst; 
  SIGNAL crosslets_cipo_arr            : t_mem_cipo_arr(g_P_sq-1 DOWNTO 0) := (OTHERS => c_mem_cipo_rst); 
  SIGNAL mon_xst_udp_sosi_arr          : t_dp_sosi_arr(0 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);

  SIGNAL crosslets_info                : STD_LOGIC_VECTOR(c_sdp_crosslets_info_reg_w-1 DOWNTO 0);
  SIGNAL nof_crosslets_reg             : STD_LOGIC_VECTOR(c_sdp_nof_crosslets_reg_w-1 DOWNTO 0);
  SIGNAL nof_crosslets                 : STD_LOGIC_VECTOR(c_sdp_nof_crosslets_reg_w-1 DOWNTO 0);
BEGIN
  ---------------------------------------------------------------
  -- Requantize 18b to 16b 
  ---------------------------------------------------------------
  gen_requantize : FOR I IN 0 TO c_sdp_P_pfb-1 GENERATE
    u_dp_requantize : ENTITY dp_lib.dp_requantize
    GENERIC MAP (
      g_complex            => TRUE,
      g_representation     => "SIGNED",
      g_lsb_w              => 0,
      g_lsb_round          => TRUE,
      g_lsb_round_clip     => FALSE,
      g_msb_clip           => TRUE,
      g_msb_clip_symmetric => FALSE,
      g_in_dat_w           => c_sdp_W_subband,
      g_out_dat_w          => c_sdp_W_crosslet
    )
    PORT MAP(
      rst => dp_rst,
      clk => dp_clk,

      snk_in  => in_sosi_arr(I),
      src_out => quant_sosi_arr(I)
    );
  END GENERATE;

  ---------------------------------------------------------------
  -- Crosslet Subband Select
  ---------------------------------------------------------------
  u_crosslets_subband_select : ENTITY work.sdp_crosslets_subband_select
  GENERIC MAP (
    g_N_crosslets            => c_sdp_N_crosslets_max,
    g_ctrl_interval_size_min => sel_a_b(g_sim, g_sim_sdp.xst_nof_clk_per_sync_min, c_sdp_xst_nof_clk_per_sync_min)
  )
  PORT MAP(
    dp_clk         => dp_clk,
    dp_rst         => dp_rst,

    in_sosi_arr    => quant_sosi_arr,
    out_sosi       => xsel_sosi,

    mm_rst         => mm_rst,
    mm_clk         => mm_clk,

    reg_crosslets_info_mosi => reg_crosslets_info_copi,
    reg_crosslets_info_miso => reg_crosslets_info_cipo,

    reg_bsn_sync_scheduler_xsub_mosi => reg_bsn_sync_scheduler_xsub_copi,
    reg_bsn_sync_scheduler_xsub_miso => reg_bsn_sync_scheduler_xsub_cipo,
                            
    out_crosslets_info => crosslets_info
  );
 
  -- Use xsel_sosi as local bsn and sync reference since the sync
  -- is generated by the bsn_sync_scheduler in sdp_crosslets_subband_select.
  bs_sosi <= xsel_sosi;

  ---------------------------------------------------------------
  -- Repack 32b to 64b 
  ---------------------------------------------------------------
  -- repacking xsel re/im to data field.
  p_wire_xsel_sosi : PROCESS(xsel_sosi)
  BEGIN
    xsel_data_sosi <= xsel_sosi;
    xsel_data_sosi.data(                c_sdp_W_crosslet -1 DOWNTO 0)                <= xsel_sosi.re(c_sdp_W_crosslet-1 DOWNTO 0);
    xsel_data_sosi.data(c_nof_complex * c_sdp_W_crosslet -1 DOWNTO c_sdp_W_crosslet) <= xsel_sosi.im(c_sdp_W_crosslet-1 DOWNTO 0);
  END PROCESS;

  u_dp_repack_data_local : ENTITY dp_lib.dp_repack_data
  GENERIC MAP (
    g_in_dat_w       => c_data_w,
    g_in_nof_words   => c_longword_w/c_data_w,
    g_out_dat_w      => c_longword_w,
    g_out_nof_words  => 1,
    g_pipeline_ready => TRUE -- Needed for src_in.ready to snk_out.ready. 
  )
  PORT MAP (
    rst => dp_rst,
    clk => dp_clk,

    snk_in  => xsel_data_sosi,
    src_out => local_sosi
  );

  ---------------------------------------------------------------
  -- ring_mux
  ---------------------------------------------------------------
  u_ring_mux : ENTITY ring_lib.ring_mux
  GENERIC MAP (
    g_bsn_w        => c_dp_stream_bsn_w,
    g_data_w       => c_longword_w,
    g_channel_w    => c_word_w,
    g_use_error    => FALSE,
    g_fifo_size    => array_init(c_mux_fifo_size, 2)
  )
  PORT MAP (
    dp_clk => dp_clk,
    dp_rst => dp_rst,

    remote_sosi => from_ri_sosi,
    local_sosi  => local_sosi,
    mux_sosi    => ring_mux_sosi,
    mux_siso    => ring_mux_siso
  );

  to_ri_sosi <= ring_mux_sosi;

  -- fill fifo to remove gaps
  u_dp_fifo_fill_eop : ENTITY dp_lib.dp_fifo_fill_eop
  GENERIC MAP (
    g_data_w         => c_longword_w,
    g_bsn_w          => c_dp_stream_bsn_w, 
    g_empty_w        => c_dp_stream_empty_w, 
    g_channel_w      => c_dp_stream_channel_w,
    g_error_w        => c_dp_stream_error_w, 
    g_use_bsn        => TRUE, 
    g_use_empty      => TRUE, 
    g_use_channel    => TRUE, 
    g_use_error      => TRUE, 
    g_use_sync       => TRUE, 
    g_fifo_fill      => c_block_size_longwords,
    g_fifo_size      => c_fifo_fill_size
  )
  PORT MAP (
    wr_rst      => dp_rst,
    wr_clk      => dp_clk,
    rd_rst      => dp_rst,
    rd_clk      => dp_clk,

    snk_out     => ring_mux_siso,
    snk_in      => ring_mux_sosi,

    src_in      => dp_fifo_fill_siso,
    src_out     => dp_fifo_fill_sosi
  );

  ---------------------------------------------------------------
  -- Repack 64b to 32b
  ---------------------------------------------------------------
  u_dp_repack_data_rx : ENTITY dp_lib.dp_repack_data
  GENERIC MAP (
    g_in_dat_w       => c_longword_w,
    g_in_nof_words   => 1,
    g_out_dat_w      => c_data_w,
    g_out_nof_words  => c_longword_w/c_data_w,
    g_pipeline_ready => TRUE -- Needed for src_in.ready to snk_out.ready.
  )
  PORT MAP (
    rst => dp_rst,
    clk => dp_clk,

    snk_in  => dp_fifo_fill_sosi,
    snk_out => dp_fifo_fill_siso,
    src_out => rx_sosi
  );

  ---------------------------------------------------------------
  -- dp_demux
  ---------------------------------------------------------------
  u_dp_demux : ENTITY dp_lib.dp_demux
  GENERIC MAP (
    g_mode              => 0,
    g_nof_output        => g_P_sq,
    g_remove_channel_lo => FALSE,
    g_sel_ctrl_invert   => TRUE --TRUE when indexed (g_nof_input-1 DOWNTO 0)
  )
  PORT MAP (
    rst => dp_rst,
    clk => dp_clk,

    snk_in      => rx_sosi,
    src_out_arr => dispatch_invert_sosi_arr
  );

  dispatch_sosi_arr <= func_dp_stream_arr_reverse_range(dispatch_invert_sosi_arr);
    

  ---------------------------------------------------------------
  -- dp_bsn_aligner_v2 
  ---------------------------------------------------------------
  u_mmp_dp_bsn_align_v2 : ENTITY dp_lib.mmp_dp_bsn_align_v2
  GENERIC MAP(
    -- for dp_bsn_align_v2
    g_nof_streams             => g_P_sq,    
    g_bsn_latency_max         => 2,    
    g_nof_aligners_max        => 1, -- 1 for Access scheme 3.
    g_block_size              => c_block_size,    
    g_data_w                  => c_data_w,    
    g_use_mm_output           => TRUE,   
    g_rd_latency              => 1, -- Required for st_xst 
    -- for mms_dp_bsn_monitor_v2
    g_nof_clk_per_sync        => c_sdp_N_clk_sync_timeout, -- Using c_sdp_N_clk_sync_timeout as g_nof_clk_per_sync is used for BSN monitor timeout.
    g_nof_input_bsn_monitors  => g_P_sq,  
    g_use_bsn_output_monitor  => TRUE  
    )
  PORT MAP (
    -- Memory-mapped clock domain
    mm_rst                  => mm_rst, 
    mm_clk                  => mm_clk, 

    reg_bsn_align_copi      => reg_bsn_align_copi, 
    reg_bsn_align_cipo      => reg_bsn_align_cipo, 

    reg_input_monitor_copi  => reg_bsn_monitor_v2_bsn_align_input_copi, 
    reg_input_monitor_cipo  => reg_bsn_monitor_v2_bsn_align_input_cipo, 

    reg_output_monitor_copi => reg_bsn_monitor_v2_bsn_align_output_copi, 
    reg_output_monitor_cipo => reg_bsn_monitor_v2_bsn_align_output_cipo, 

    -- Streaming clock domain
    dp_rst     => dp_rst,             
    dp_clk     => dp_clk,             

    -- Streaming input
    in_sosi_arr => dispatch_sosi_arr,             

    -- Output via local MM interface in dp_clk domain, when g_use_mm_output = TRUE.
    mm_sosi     => crosslets_sosi,            
    mm_copi     => crosslets_copi,            
    mm_cipo_arr => crosslets_cipo_arr            
  );

  ---------------------------------------------------------------
  -- Crosslets Statistics (XST) 
  ---------------------------------------------------------------
  u_crosslets_stats : ENTITY st_lib.st_xst
  GENERIC MAP(
    g_nof_streams       => g_P_sq,
    g_nof_crosslets     => c_sdp_N_crosslets_max,
    g_nof_signal_inputs => c_sdp_S_pn,
    g_in_data_w         => c_sdp_W_crosslet,
    g_stat_data_w       => c_longword_w,
    g_stat_data_sz      => c_longword_sz/c_word_sz
  )
  PORT MAP (
    mm_rst          => mm_rst,
    mm_clk          => mm_clk,
    dp_rst          => dp_rst,
    dp_clk          => dp_clk,
    in_sosi         => crosslets_sosi,
    mm_mosi         => crosslets_copi,
    mm_miso_arr     => crosslets_cipo_arr,

    ram_st_xsq_mosi => controller_mem_mux_copi,
    ram_st_xsq_miso => controller_mem_mux_cipo
  );

  ---------------------------------------------------------------
  -- MM controller multiplexer 
  ---------------------------------------------------------------
  -- Connect 2 mm_controllers to the common_mem_mux output
  controller_copi_arr(0)  <= ram_st_xsq_copi;    -- MM access via QSYS MM bus
  ram_st_xsq_cipo     <= controller_cipo_arr(0);
  controller_copi_arr(1)  <= ram_st_offload_copi;   -- MM access by UDP offload
  ram_st_offload_cipo <= controller_cipo_arr(1);

  u_mem_controller_mux : ENTITY mm_lib.mm_master_mux
  GENERIC MAP (
    g_nof_masters    => c_nof_controllers,
    g_rd_latency_min => 1  -- read latency of statistics RAM is 1
  )
  PORT MAP (
    mm_clk => mm_clk,

    master_mosi_arr => controller_copi_arr,
    master_miso_arr => controller_cipo_arr,
    mux_mosi        => controller_mem_mux_copi,
    mux_miso        => controller_mem_mux_cipo
  );

  ---------------------------------------------------------------
  -- REG_NOF_CROSSLETS
  ---------------------------------------------------------------
  u_nof_crosslets : ENTITY common_lib.mms_common_reg
  GENERIC MAP(
    g_mm_reg => c_sdp_mm_reg_nof_crosslets
  )
  PORT MAP(
    -- Clocks and reset
    mm_rst => mm_rst, 
    mm_clk => mm_clk,
    st_rst => dp_rst,
    st_clk => dp_clk,
    
    -- MM bus access in memory-mapped clock domain
    reg_mosi => reg_nof_crosslets_copi,
    reg_miso => reg_nof_crosslets_cipo, 
    
    in_reg   => nof_crosslets,
    out_reg  => nof_crosslets_reg
  );
  -- Force nof crosslets to max nof crosslets if a higher value is written or to 1 if a lower value is written via MM.
  nof_crosslets <= TO_UVEC(1, c_sdp_nof_crosslets_reg_w) WHEN TO_UINT(nof_crosslets_reg) < 1 ELSE 
                   nof_crosslets_reg WHEN TO_UINT(nof_crosslets_reg) <= c_sdp_N_crosslets_max ELSE 
                   TO_UVEC(c_sdp_N_crosslets_max, c_sdp_nof_crosslets_reg_w);

  ---------------------------------------------------------------
  -- XST UDP offload 
  ---------------------------------------------------------------
  xst_udp_sosi <= mon_xst_udp_sosi_arr(0);
  u_sdp_xst_udp_offload: ENTITY work.sdp_statistics_offload
  GENERIC MAP (
    g_statistics_type     => "XST",
    g_offload_time        => sel_a_b(g_sim, g_sim_sdp.offload_time, c_sdp_offload_time),
    g_P_sq                => g_P_sq,
    g_crosslets_direction => 1  -- = lane direction
  )
  PORT MAP (
    mm_clk    => mm_clk,
    mm_rst    => mm_rst,

    dp_clk    => dp_clk,
    dp_rst    => dp_rst,

    master_mosi => ram_st_offload_copi,
    master_miso => ram_st_offload_cipo,

    reg_enable_mosi  => reg_stat_enable_copi,
    reg_enable_miso  => reg_stat_enable_cipo,

    reg_hdr_dat_mosi => reg_stat_hdr_dat_copi,
    reg_hdr_dat_miso => reg_stat_hdr_dat_cipo,

    reg_bsn_monitor_v2_offload_copi => reg_bsn_monitor_v2_xst_offload_copi,
    reg_bsn_monitor_v2_offload_cipo => reg_bsn_monitor_v2_xst_offload_cipo,

    in_sosi   => crosslets_sosi,
    out_sosi  => mon_xst_udp_sosi_arr(0),
    out_siso  => xst_udp_siso,

    eth_src_mac    => stat_eth_src_mac,
    udp_src_port   => stat_udp_src_port,
    ip_src_addr    => stat_ip_src_addr,

    gn_index       => TO_UINT(gn_id),
    ring_info      => ring_info,
    sdp_info       => sdp_info,
    weighted_subbands_flag  => '1',  -- because XSub uses in_sosi_arr = fsub_sosi_arr, so weighted subbands
    nof_crosslets  => nof_crosslets,
    crosslets_info => crosslets_info
  );

END str;
