-------------------------------------------------------------------------------
--
-- Copyright 2020
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--
-- Author: R. van der Walle
-- Purpose: 
-- . Implements the functionality of the subband filterbank (Fsub) in the 
--   LOFAR2 SDPFW design.
-- Description:
-- . The subband filterbank seperates the incoming timestamped ADC samples into
--   512 frequency bands called subbands. 
-- . It implements a critically sampled poly-phase filterbank (PFB). The PFB consists of a 
--   poly-phase finite impulse response (PFIR) filter per real input and a 
--   complex fast fourier transform (FFT) per 2 real inputs. 
-- . The number of points of the FFT is 1024.
-- Remark:
-- .
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib, rTwoSDF_lib, wpfb_lib, filter_lib, si_lib, st_lib, mm_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_network_layers_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE rTwoSDF_lib.rTwoSDFPkg.ALL;
USE filter_lib.fil_pkg.ALL;
USE wpfb_lib.wpfb_pkg.ALL;
USE work.sdp_pkg.ALL;

ENTITY node_sdp_filterbank IS
  GENERIC (
    g_sim                    : BOOLEAN := FALSE;
    g_sim_sdp                : t_sdp_sim := c_sdp_sim;
    g_wpfb                   : t_wpfb := c_sdp_wpfb_subbands;
    g_scope_selected_subband : NATURAL := 0
  );
  PORT (
    dp_clk        : IN  STD_LOGIC;
    dp_rst        : IN  STD_LOGIC;

    in_sosi_arr   : IN  t_dp_sosi_arr(c_sdp_S_pn-1 DOWNTO 0);
    pfb_sosi_arr  : OUT t_dp_sosi_arr(c_sdp_P_pfb-1 DOWNTO 0);
    fsub_sosi_arr : OUT t_dp_sosi_arr(c_sdp_P_pfb-1 DOWNTO 0);
    sst_udp_sosi  : OUT t_dp_sosi;
    sst_udp_siso  : IN  t_dp_siso := c_dp_siso_rst;

    dp_bsn_source_restart : IN STD_LOGIC;

    mm_rst             : IN  STD_LOGIC;
    mm_clk             : IN  STD_LOGIC;

    reg_si_mosi        : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_si_miso        : OUT t_mem_miso;    
    ram_st_sst_mosi    : IN  t_mem_mosi := c_mem_mosi_rst; 
    ram_st_sst_miso    : OUT t_mem_miso;
    ram_fil_coefs_mosi : IN  t_mem_mosi := c_mem_mosi_rst; 
    ram_fil_coefs_miso : OUT t_mem_miso;
    ram_gains_mosi     : IN  t_mem_mosi := c_mem_mosi_rst; 
    ram_gains_miso     : OUT t_mem_miso;
    reg_selector_mosi  : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_selector_miso  : OUT t_mem_miso;    
    reg_enable_mosi    : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_enable_miso    : OUT t_mem_miso;    
    reg_hdr_dat_mosi   : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_hdr_dat_miso   : OUT t_mem_miso;
    reg_bsn_monitor_v2_sst_offload_copi : IN  t_mem_copi := c_mem_copi_rst;
    reg_bsn_monitor_v2_sst_offload_cipo : OUT t_mem_cipo;

    sdp_info : IN t_sdp_info;
    gn_id    : IN STD_LOGIC_VECTOR(c_sdp_W_gn_id-1 DOWNTO 0);

    eth_src_mac  : IN STD_LOGIC_VECTOR(c_network_eth_mac_addr_w-1 DOWNTO 0);
    ip_src_addr  : IN STD_LOGIC_VECTOR(c_network_ip_addr_w-1 DOWNTO 0);
    udp_src_port : IN STD_LOGIC_VECTOR(c_network_udp_port_w-1 DOWNTO 0) 
  );
END node_sdp_filterbank;

ARCHITECTURE str OF node_sdp_filterbank IS
  
  CONSTANT c_coefs_file_prefix : STRING := "data/Coeffs16384Kaiser-quant_1wb";
  CONSTANT c_gains_file_name : STRING := "data/gains_1024_complex_16b13f_unit"; -- Can be generated by src/python/sdp_hex.py

  CONSTANT c_subband_equalizer_latency : NATURAL := 5;

  CONSTANT c_nof_masters : POSITIVE := 2;

  CONSTANT c_si_pipeline : NATURAL := 1;

  SIGNAL ram_st_sst_mosi_arr : t_mem_mosi_arr(c_sdp_P_pfb-1 DOWNTO 0) := (OTHERS => c_mem_mosi_rst);
  SIGNAL ram_st_sst_miso_arr : t_mem_miso_arr(c_sdp_P_pfb-1 DOWNTO 0) := (OTHERS => c_mem_miso_rst);

  -- Subband statistics
  SIGNAL ram_st_offload_mosi : t_mem_mosi := c_mem_mosi_rst;
  SIGNAL ram_st_offload_miso : t_mem_miso := c_mem_miso_rst;
  
  SIGNAL master_mem_mux_mosi : t_mem_mosi := c_mem_mosi_rst;
  SIGNAL master_mem_mux_miso : t_mem_miso := c_mem_miso_rst;
  SIGNAL master_mosi_arr     : t_mem_mosi_arr(0 TO c_nof_masters-1) := (OTHERS=>c_mem_mosi_rst);
  SIGNAL master_miso_arr     : t_mem_miso_arr(0 TO c_nof_masters-1) := (OTHERS=>c_mem_miso_rst);

  SIGNAL si_sosi_arr                    : t_dp_sosi_arr(c_sdp_S_pn-1 DOWNTO 0)  := (OTHERS => c_dp_sosi_rst);
  SIGNAL wpfb_unit_out_sosi_arr         : t_dp_sosi_arr(c_sdp_P_pfb-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);
  SIGNAL wpfb_unit_fil_sosi_arr         : t_dp_sosi_arr(c_sdp_P_pfb-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);
  SIGNAL wpfb_unit_in_sosi_arr          : t_dp_sosi_arr(c_sdp_P_pfb-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);
  SIGNAL subband_equalizer_out_sosi_arr : t_dp_sosi_arr(c_sdp_P_pfb-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);
  SIGNAL dp_selector_out_sosi_arr       : t_dp_sosi_arr(c_sdp_P_pfb-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);
  SIGNAL scope_sosi_arr                 : t_dp_sosi_integer_arr(c_sdp_S_pn-1 DOWNTO 0);
  
  SIGNAL selector_en                : STD_LOGIC;
  SIGNAL weighted_subbands_flag     : STD_LOGIC;
  SIGNAL dp_bsn_source_restart_pipe : STD_LOGIC;
BEGIN
  ---------------------------------------------------------------
  -- SPECTRAL INVERSION 
  ---------------------------------------------------------------
  u_si_arr : ENTITY si_lib.si_arr
  GENERIC MAP (
    g_nof_streams => c_sdp_S_pn, 
    g_pipeline    => c_si_pipeline,  
    g_dat_w       => c_sdp_W_adc 
  )
  PORT MAP(
    in_sosi_arr  => in_sosi_arr, 
    out_sosi_arr => si_sosi_arr, 
    
    reg_si_mosi  => reg_si_mosi, 
    reg_si_miso  => reg_si_miso, 

    mm_rst       => mm_rst, 
    mm_clk       => mm_clk, 
    dp_clk       => dp_clk, 
    dp_rst       => dp_rst 
  );

  ---------------------------------------------------------------
  -- POLY-PHASE FILTERBANK
  ---------------------------------------------------------------
  -- Connect the 12 ADC streams to the re and im fields of the PFB input.
  p_pfb_streams : PROCESS(si_sosi_arr)
  BEGIN
    FOR I IN 0 TO c_sdp_P_pfb-1 LOOP
      wpfb_unit_in_sosi_arr(I) <= si_sosi_arr(2*I);
      wpfb_unit_in_sosi_arr(I).re <= RESIZE_DP_DSP_DATA(si_sosi_arr(2*I).data);
      wpfb_unit_in_sosi_arr(I).im <= RESIZE_DP_DSP_DATA(si_sosi_arr(2*I+1).data);
    END LOOP;
  END PROCESS;

  -- pipeline bsn restart signal to keep dp_bsn_source_restart aligned with si_sosi_arr
  u_common_pipeline_sl : ENTITY common_lib.common_pipeline_sl
  GENERIC MAP (
    g_pipeline  => c_si_pipeline
  )
  PORT MAP (
    rst        => dp_rst,
    clk        => dp_clk,
    in_dat     => dp_bsn_source_restart,
    out_dat    => dp_bsn_source_restart_pipe
  ); 

  -- PFB 
  u_wpfb_unit_dev : ENTITY wpfb_lib.wpfb_unit_dev
  GENERIC MAP (
    g_wpfb                   => g_wpfb,
    g_use_prefilter          => TRUE,
    g_stats_ena              => FALSE,
    g_use_bg                 => FALSE,
    g_coefs_file_prefix      => c_coefs_file_prefix,
    g_restart_on_valid       => FALSE 
  )
  PORT MAP (
    dp_rst             => dp_rst, 
    dp_clk             => dp_clk, 
    mm_rst             => mm_rst, 
    mm_clk             => mm_clk, 

    ram_fil_coefs_mosi => ram_fil_coefs_mosi, 
    ram_fil_coefs_miso => ram_fil_coefs_miso, 

    in_sosi_arr        => wpfb_unit_in_sosi_arr, 
    fil_sosi_arr       => wpfb_unit_fil_sosi_arr, 
    out_sosi_arr       => wpfb_unit_out_sosi_arr,
 
    dp_bsn_source_restart => dp_bsn_source_restart_pipe
  );
  
  -- Output PFB streams
  pfb_sosi_arr <= wpfb_unit_out_sosi_arr;

  ---------------------------------------------------------------
  -- SUBBAND EQUALIZER
  ---------------------------------------------------------------
  u_sdp_subband_equalizer : ENTITY work.sdp_subband_equalizer
    GENERIC MAP (
      g_gains_file_name => c_gains_file_name
    )
    PORT MAP(
      dp_clk         => dp_clk,
      dp_rst         => dp_rst,

      in_sosi_arr    => wpfb_unit_out_sosi_arr,
      out_sosi_arr   => subband_equalizer_out_sosi_arr,
  
      mm_rst         => mm_rst,
      mm_clk         => mm_clk,
  
      ram_gains_mosi => ram_gains_mosi,  
      ram_gains_miso => ram_gains_miso 
    );

  -- Output fsub streams
  fsub_sosi_arr <= subband_equalizer_out_sosi_arr; 

  ---------------------------------------------------------------
  -- DP SELECTOR 
  ---------------------------------------------------------------
  u_dp_selector_arr : ENTITY dp_lib.dp_selector_arr
    GENERIC MAP (
      g_nof_arr => c_sdp_P_pfb,
      g_pipeline => c_subband_equalizer_latency
    )
    PORT MAP (
      mm_rst => mm_rst,
      mm_clk => mm_clk,
      dp_rst => dp_rst,
      dp_clk => dp_clk,

      reg_selector_mosi => reg_selector_mosi,
      reg_selector_miso => reg_selector_miso,

      pipe_sosi_arr  => wpfb_unit_out_sosi_arr, 
      ref_sosi_arr   => subband_equalizer_out_sosi_arr,    
      out_sosi_arr   => dp_selector_out_sosi_arr,

      selector_en    => selector_en
    );
 
  ---------------------------------------------------------------
  -- SUBBAND STATISTICS
  ---------------------------------------------------------------
  gen_stats_streams: FOR I IN 0 TO c_sdp_P_pfb-1 GENERATE
      u_subband_stats : ENTITY st_lib.st_sst
      GENERIC MAP(
        g_nof_stat      => c_sdp_N_sub*c_sdp_Q_fft,
        g_in_data_w     => c_sdp_W_subband,
        g_stat_data_w   => g_wpfb.stat_data_w,
        g_stat_data_sz  => g_wpfb.stat_data_sz
      )
      PORT MAP (
        mm_rst          => mm_rst,
        mm_clk          => mm_clk,
        dp_rst          => dp_rst,
        dp_clk          => dp_clk,
        in_complex      => dp_selector_out_sosi_arr(I),
        ram_st_sst_mosi => ram_st_sst_mosi_arr(I),
        ram_st_sst_miso => ram_st_sst_miso_arr(I)
      );
  END GENERATE;

  ---------------------------------------------------------------
  -- COMBINE MEMORY MAPPED INTERFACES OF SST
  ---------------------------------------------------------------
  -- Combine the internal array of mm interfaces for the subband
  -- statistics to one array. 
  u_mem_mux_sst : ENTITY common_lib.common_mem_mux
  GENERIC MAP (
    g_nof_mosi    => c_sdp_P_pfb,
    g_mult_addr_w => ceil_log2(c_sdp_N_sub*c_sdp_Q_fft*g_wpfb.stat_data_sz)
  )
  PORT MAP (
    mosi     => master_mem_mux_mosi,
    miso     => master_mem_mux_miso,
    mosi_arr => ram_st_sst_mosi_arr,
    miso_arr => ram_st_sst_miso_arr
  );

  -- Connect 2 mm_masters to the common_mem_mux output
  master_mosi_arr(0)  <= ram_st_sst_mosi;    -- MM access via QSYS MM bus
  ram_st_sst_miso     <= master_miso_arr(0);
  master_mosi_arr(1)  <= ram_st_offload_mosi;   -- MM access by SST offload
  ram_st_offload_miso <= master_miso_arr(1);

  u_mem_master_mux : ENTITY mm_lib.mm_master_mux
  GENERIC MAP (
    g_nof_masters    => c_nof_masters,
    g_rd_latency_min => 1  -- read latency of statistics RAM is 1
  )
  PORT MAP (
    mm_clk => mm_clk,

    master_mosi_arr => master_mosi_arr,
    master_miso_arr => master_miso_arr,
    mux_mosi        => master_mem_mux_mosi,
    mux_miso        => master_mem_mux_miso
  );
  
  ---------------------------------------------------------------
  -- SIGNAL SCOPE
  ---------------------------------------------------------------
  u_sdp_scope : ENTITY work.sdp_scope
    GENERIC MAP (
      g_sim            => g_sim,
      g_selection      => g_scope_selected_subband,
      g_nof_input      => c_sdp_P_pfb,
      g_n_deinterleave => c_sdp_Q_fft,
      g_dat_w          => c_sdp_W_subband
    )
    PORT MAP (
      clk            => dp_clk,
      rst            => dp_rst,
      sp_sosi_arr    => subband_equalizer_out_sosi_arr,
      scope_sosi_arr => scope_sosi_arr
    );

  ---------------------------------------------------------------
  -- STATISTICS OFFLOAD
  ---------------------------------------------------------------
  weighted_subbands_flag <= NOT selector_en;

  u_sdp_sst_udp_offload: ENTITY work.sdp_statistics_offload
  GENERIC MAP (
    g_statistics_type => "SST",
    g_offload_time    => sel_a_b(g_sim, g_sim_sdp.offload_time, c_sdp_offload_time)
  )
  PORT MAP (
    mm_clk    => mm_clk,
    mm_rst    => mm_rst,

    dp_clk    => dp_clk,
    dp_rst    => dp_rst,

    master_mosi => ram_st_offload_mosi,
    master_miso => ram_st_offload_miso,

    reg_enable_mosi  => reg_enable_mosi,
    reg_enable_miso  => reg_enable_miso,

    reg_hdr_dat_mosi  => reg_hdr_dat_mosi,
    reg_hdr_dat_miso  => reg_hdr_dat_miso,

    reg_bsn_monitor_v2_offload_copi => reg_bsn_monitor_v2_sst_offload_copi,
    reg_bsn_monitor_v2_offload_cipo => reg_bsn_monitor_v2_sst_offload_cipo,

    in_sosi   => dp_selector_out_sosi_arr(0),
    out_sosi  => sst_udp_sosi,
    out_siso  => sst_udp_siso,

    eth_src_mac  => eth_src_mac,
    udp_src_port => udp_src_port,
    ip_src_addr  => ip_src_addr,

    gn_index                => TO_UINT(gn_id),
    sdp_info                => sdp_info,
    weighted_subbands_flag  => weighted_subbands_flag
  );

END str;
