-------------------------------------------------------------------------------
--
-- Copyright 2020
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--
-- Author: P. Donker

-- Purpose:
-- . SDP info register
-- Description:
--
-- https://plm.astron.nl/polarion/#/project/LOFAR2System/wiki/L2%20Interface%20Control%20Documents/SC%20to%20SDP%20ICD
--
-- Remark:
-- .
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, mm_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_field_pkg.ALL;
USE work.sdp_pkg.ALL;

ENTITY sdp_info IS
  PORT (
    -- Clocks and reset
    mm_rst             : IN  STD_LOGIC;   -- reset synchronous with mm_clk
    mm_clk             : IN  STD_LOGIC;   -- memory-mapped bus clock

    dp_clk             : IN  STD_LOGIC;
    dp_rst             : IN  STD_LOGIC;

    reg_mosi           : IN  t_mem_mosi;
    reg_miso           : OUT t_mem_miso;

    -- inputs from other blocks
    gn_index           : IN NATURAL;
    f_adc              : IN STD_LOGIC;
    fsub_type          : IN STD_LOGIC;

    -- sdp info
    sdp_info           : OUT t_sdp_info
  );
END sdp_info;


ARCHITECTURE str OF sdp_info IS

  SIGNAL sdp_info_ro: t_sdp_info := c_sdp_info_rst;  -- ro = read only
  SIGNAL block_period: STD_LOGIC_VECTOR(15 DOWNTO 0);

BEGIN

  u_mm_fields: ENTITY work.sdp_info_reg
  PORT MAP (

    mm_clk    => mm_clk,
    mm_rst    => mm_rst,

    dp_clk    => dp_clk,
    dp_rst    => dp_rst,

    reg_mosi  => reg_mosi,
    reg_miso  => reg_miso,

    -- sdp info
    sdp_info_ro => sdp_info_ro,
    sdp_info    => sdp_info
  );

  
  -- f_adc    : '0' => 160M, '1' => 200M
  -- fsub_type: '0' => critical sampled PFB, '1' => oversampled PFB  
  p_block_period : PROCESS(f_adc, fsub_type)
  BEGIN
    IF f_adc = '0' THEN
      IF fsub_type = '0' THEN
        block_period <= TO_UVEC(6400, block_period'LENGTH);  -- 160M, critical sampled
      ELSE
        block_period <= TO_UVEC(5400, block_period'LENGTH);  -- 160M, oversampled
      END IF;
    ELSE
      IF fsub_type = '0' THEN
        block_period <= TO_UVEC(5120, block_period'LENGTH);  -- 200M, critical sampled
      ELSE
        block_period <= TO_UVEC(4320, block_period'LENGTH);  -- 200M, oversampled
      END IF;  
    END IF;
  END PROCESS;

  sdp_info_ro.f_adc              <= f_adc;
  sdp_info_ro.fsub_type          <= fsub_type;
  sdp_info_ro.block_period       <= block_period;

END str;
