-------------------------------------------------------------------------------
--
-- Copyright 2020
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--
-- Author: P. Donker, R van der Walle, E. Kooistra

-- Purpose:
-- . SDP statistics offload
-- Description:
-- [1] https://support.astron.nl/confluence/display/L2M/L5+SDPFW+Design+Document%3A+Subband+filterbank
--     . See figure 4.3
-- [2] https://support.astron.nl/confluence/display/L2M/L5+SDPFW+Design+Document%3A+Subband+Correlator
--     . See Figure 3.7
-- [3] https://plm.astron.nl/polarion/#/project/LOFAR2System/wiki/L2%20Interface%20Control%20Documents/SC%20to%20SDP%20ICD
--     . See 2.9.4 Station Control (L3-SC) - SDP Firmware (L4-SDPFW)
--
-- . endianess
--   Within a 32bit MM word the values are stored with LSByte at lowest byte
--   address (so relative byte address 0) and MSByte at highest byte address
--   (so relative byte address 3). Internally in the RAM the MM words are thus
--   stored with LSByte first, which is called little endian. However,
--   externally the MM words are send with MSByte first, so in the statistics
--   offload header and offload payload the 32bit words are send with MSByte
--   first, which is called big endian.
--   The dp_offload_tx_v3 sends all multi byte header fields in big endian
--   order, hence also the 16bit, 24bit and 64bit values are send MSByte
--   first. The internet packet headers also use big endian, which is
--   therefore also called network order. Hence the entire statistics offload
--   packet is in big endian order.
--   If values are shown with first byte received left and last byte received
--   right, then big endian hex values can directly be interpreted when read
--   by a human, because they are shown in the normal order.
--
-- . g_reverse_word_order
--   The statistics consist of c_sdp_W_statistic_sz = 2 MM words that are
--   with LSWord at low (even) word address and MSWord at high (odd) word
--   address. Default the statistics offload reads the low address first, but
--   the statistics have to be send with MSWord first. Therefore the read
--   order needs to be reversed per statistic value, so g_reverse_word_order
--   = TRUE. The combination of sending the MSWord first and the MSByte first
--   results that the entire 64b statistics words are send in big endian
--   order.
--
--   Stored order:
--   word address:   0,   1,   2,   3
--   SST            Ul,  Uh
--   BST            Xl,  Xh,  Yl,  Yh
--   XST            Rl,  Rh,  Il,  Ih
--
--   The g_reverse_word_order = TRUE is needed to achieve have MSWord first
--   and 64bit big endian.
--   The g_user_size defines the number of words that get reversed:
--   . For the SST there is only one uint64 part, so g_user_size = 2.
--   . For the BST the X and Y polarization parts are treated as an array of
--     [N_pol_bf], so index 0 = X is send first and therefore g_user_size = 2
--     to preserve the polarization X, Y order.
--   . For the XST the Re and Im complex parts of cint64 are treated as an
--     array of [N_complex], so index 0 = Re is send first and therefore
--     g_user_size = 2 to preserve the complex Re, Im order.
--
--   The () show the parts that are contained in g_user_size and that got
--   reversed by g_reverse_word_order = TRUE compared to the stored order:
--
--   Transport order:                   (g_user_size)
--   SST           (Uh,  Ul),                2
--   BST           (Xh,  Xl), (Yh,  Yl),     2   keep X, Y parts order
--   XST           (Rh,  Rl), (Ih,  Il),     2   keep Re, Im parts order
--
-- . g_P_sq and nof_used_P_sq
--   The g_P_sq defines the number of correlator cells that is available in
--   the SDPFW. Use generic to support P_sq = 1 for one node and P_sq =
--   c_sdp_P_sq for multiple nodes (with ring).
--   The nof_used_P_sq is the number of correlator cells that is actually
--   used and that will output XST packets. Unused correlator cells yield
--   zero data that should not be output. The nof_used_P_sq is the smallest
--   of g_P_sq and ring_info.N_rn/2 + 1. In this way the XST offload can work
--   with g_P_sq = 1 when N_rn > 1 and also in a ring with N_rn < N_pn when
--   g_P_sq = 9.
--
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, mm_lib, dp_lib, ring_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.common_field_pkg.ALL;
USE common_lib.common_network_layers_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE ring_lib.ring_pkg.ALL;
USE work.sdp_pkg.ALL;

ENTITY sdp_statistics_offload IS
  GENERIC (
    g_statistics_type     : STRING  := "SST";
    g_offload_time        : NATURAL := c_sdp_offload_time;
    g_beamset_id          : NATURAL := 0;
    g_P_sq                : NATURAL := c_sdp_P_sq;  -- number of available correlator cells,
    g_crosslets_direction : NATURAL := 1; -- > 0 for crosslet transport in positive direction (incrementing RN), else 0 for negative direction
    g_reverse_word_order  : BOOLEAN := TRUE  -- default word order is MSB after LSB, we need to stream LSB after MSB.
  );
  PORT (
    -- Clocks and reset
    mm_rst : IN  STD_LOGIC;   -- reset synchronous with mm_clk
    mm_clk : IN  STD_LOGIC;   -- memory-mapped bus clock

    dp_clk : IN  STD_LOGIC;
    dp_rst : IN  STD_LOGIC;

    -- Memory access to statistics values
    master_mosi      : OUT  t_mem_mosi;  -- := c_mem_mosi_rst;
    master_miso      : IN t_mem_miso;

    -- Memory access to read/write settings
    reg_enable_mosi  : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_enable_miso  : OUT t_mem_miso;

    reg_hdr_dat_mosi : IN  t_mem_mosi := c_mem_mosi_rst;
    reg_hdr_dat_miso : OUT t_mem_miso;

    -- Memory access bsn monitor udp offload
    reg_bsn_monitor_v2_offload_copi : IN  t_mem_copi := c_mem_copi_rst;
    reg_bsn_monitor_v2_offload_cipo : OUT t_mem_cipo;

    -- Input timing regarding the integration interval of the statistics
    in_sosi          : IN t_dp_sosi;
    
    -- Streaming output of offloaded statistics packets
    out_sosi         : OUT t_dp_sosi;
    out_siso         : IN t_dp_siso;

    -- inputs from other blocks
    eth_src_mac             : IN STD_LOGIC_VECTOR(c_network_eth_mac_addr_w-1 DOWNTO 0);
    udp_src_port            : IN STD_LOGIC_VECTOR(c_network_udp_port_w-1 DOWNTO 0);
    ip_src_addr             : IN STD_LOGIC_VECTOR(c_network_ip_addr_w-1 DOWNTO 0);

    gn_index                : IN NATURAL;
    ring_info               : IN t_ring_info := c_ring_info_rst;  -- only needed for XST
    sdp_info                : IN t_sdp_info;
    weighted_subbands_flag  : IN STD_LOGIC := '0';
    nof_crosslets           : IN STD_LOGIC_VECTOR(c_sdp_nof_crosslets_reg_w-1 DOWNTO 0) := (OTHERS => '0');
    crosslets_info          : IN STD_LOGIC_VECTOR(c_sdp_crosslets_info_reg_w-1 DOWNTO 0) := (OTHERS => '0')
  );
END sdp_statistics_offload;


ARCHITECTURE str OF sdp_statistics_offload IS

  CONSTANT c_nof_streams               : NATURAL := 1;
  
  -- header fields
  CONSTANT c_marker                    : NATURAL := func_sdp_get_stat_marker(g_statistics_type);
  CONSTANT c_nof_signal_inputs         : NATURAL := func_sdp_get_stat_nof_signal_inputs(g_statistics_type);
  CONSTANT c_nof_statistics_per_packet : NATURAL := func_sdp_get_stat_nof_statistics_per_packet(g_statistics_type);
  CONSTANT c_udp_total_length          : NATURAL := func_sdp_get_stat_udp_total_length(g_statistics_type);
  CONSTANT c_ip_total_length           : NATURAL := func_sdp_get_stat_ip_total_length(g_statistics_type);
  CONSTANT c_nof_packets_max           : NATURAL := func_sdp_get_stat_nof_packets(g_statistics_type, c_sdp_S_pn, g_P_sq, c_sdp_N_crosslets_max);

  CONSTANT c_beamlet_id                : NATURAL := g_beamset_id * c_sdp_S_sub_bf;

  -- MM access settings per packet for u_dp_block_from_mm_dc
  CONSTANT c_mm_user_size              : NATURAL := func_sdp_get_stat_from_mm_user_size(g_statistics_type);
  CONSTANT c_mm_data_size              : NATURAL := func_sdp_get_stat_from_mm_data_size(g_statistics_type);
  CONSTANT c_mm_step_size              : NATURAL := func_sdp_get_stat_from_mm_step_size(g_statistics_type);
  CONSTANT c_mm_nof_data               : NATURAL := func_sdp_get_stat_from_mm_nof_data(g_statistics_type);
  CONSTANT c_mm_ram_size               : NATURAL := c_mm_nof_data * c_mm_data_size * c_nof_packets_max;

  -- offload control
  TYPE t_reg IS RECORD
    packet_count         : NATURAL RANGE 0 TO c_nof_packets_max;
    start_address        : NATURAL RANGE 0 TO c_mm_ram_size;
    start_pulse          : STD_LOGIC;
    sync                 : STD_LOGIC;
    dp_header_info       : STD_LOGIC_VECTOR(1023 DOWNTO 0);
    payload_err          : STD_LOGIC;
    in_sop_cnt           : NATURAL;
    integration_interval : NATURAL;
    interleave_count     : NATURAL RANGE 0 TO c_sdp_Q_fft;
    interleave_address   : NATURAL RANGE 0 TO c_mm_ram_size;
    crosslet_count       : NATURAL RANGE 0 TO c_sdp_N_crosslets_max;
    instance_count       : NATURAL RANGE 0 TO c_sdp_P_sq;
    instance_address     : NATURAL RANGE 0 TO c_mm_ram_size;
    nof_crosslets        : NATURAL RANGE 0 TO c_sdp_N_crosslets_max;
    crosslets_info_rec   : t_sdp_crosslets_info;
  END RECORD;

  CONSTANT c_crosslets_info_rst : t_sdp_crosslets_info := (offset_arr => (OTHERS => 0), step => 0);
  CONSTANT c_reg_rst            : t_reg := (0, 0, '0', '0', (OTHERS => '0'), '0', 0, 0, 0, 0, 0, 0, 0, 0, c_crosslets_info_rst);

  SIGNAL r     : t_reg;
  SIGNAL nxt_r : t_reg;

  SIGNAL gn_index_reg             : NATURAL;  -- index of this global node
  SIGNAL pn_index                 : NATURAL;  -- index of this node in antenna band
  SIGNAL rn_index                 : NATURAL;  -- index of this ring node
  SIGNAL local_si_offset          : NATURAL;  -- index of first signal input on this node
  SIGNAL remote_rn                : NATURAL;  -- index of remote ring node
  SIGNAL remote_gn                : NATURAL;  -- index of remote global node
  SIGNAL remote_pn                : NATURAL;  -- index of remote node in antenna band
  SIGNAL remote_si_offset         : NATURAL;  -- index of first signal input on remote node
  SIGNAL nof_cycles_dly           : NATURAL;  -- trigger_offload delay for this node
  SIGNAL offset_rn                : NATURAL;  -- = ring_info.O_rn, GN index of first ring node
  SIGNAL nof_rn                   : NATURAL;  -- = ring_info.N_rn, number of GN in the ring
  SIGNAL nof_used_P_sq            : NATURAL;  -- number of used correlator cells <= g_P_sq (is number of available correlator cells)
  SIGNAL nof_packets              : NATURAL;  -- nof XST offload packets per integration interval

  SIGNAL data_id_rec              : t_sdp_stat_data_id;
  SIGNAL data_id_slv              : STD_LOGIC_VECTOR(31 DOWNTO 0) := (OTHERS => '0');

  SIGNAL trigger_en               : STD_LOGIC := '0';
  SIGNAL trigger_offload          : STD_LOGIC := '0';
  SIGNAL mm_done                  : STD_LOGIC := '0';
  SIGNAL dp_sop                   : STD_LOGIC := '0';
  SIGNAL dp_block_from_mm_src_out : t_dp_sosi;
  SIGNAL dp_block_from_mm_src_in  : t_dp_siso;
  
  SIGNAL dp_offload_snk_in        : t_dp_sosi;
  SIGNAL dp_offload_snk_out       : t_dp_siso;

  SIGNAL udp_sosi                 : t_dp_sosi;

  SIGNAL bsn_at_sync              : STD_LOGIC_VECTOR(c_dp_stream_bsn_w-1 DOWNTO 0) := (OTHERS => '0');
  SIGNAL dp_header_info           : STD_LOGIC_VECTOR(1023 DOWNTO 0):= (OTHERS => '0');

  -- Debug signals for view in Wave window
  SIGNAL dbg_c_marker                    : NATURAL := c_marker;
  SIGNAL dbg_c_nof_signal_inputs         : NATURAL := c_nof_signal_inputs;
  SIGNAL dbg_c_nof_statistics_per_packet : NATURAL := c_nof_statistics_per_packet;
  SIGNAL dbg_c_udp_total_length          : NATURAL := c_udp_total_length;
  SIGNAL dbg_c_ip_total_length           : NATURAL := c_ip_total_length;
  SIGNAL dbg_c_nof_packets_max           : NATURAL := c_nof_packets_max;
  SIGNAL dbg_c_beamlet_id                : NATURAL := c_beamlet_id;
  SIGNAL dbg_c_mm_data_size              : NATURAL := c_mm_data_size;
  SIGNAL dbg_c_mm_step_size              : NATURAL := c_mm_step_size;
  SIGNAL dbg_c_mm_nof_data               : NATURAL := c_mm_nof_data;

BEGIN

  bsn_at_sync <= RESIZE_UVEC(in_sosi.bsn, c_dp_stream_bsn_w) WHEN rising_edge(dp_clk) AND in_sosi.sync = '1';

  -------------------------------------------------------------------------------
  -- Assemble offload header info, for data path fields that are selected by:
  --   c_sdp_stat_hdr_field_sel = "1"&"101"&"111011111001"&"0100"&"0100"&"000000010"&"1000000"&"0"
  --                                   eth   ip             udp    app
  --   where 0 = data path, 1 = MM controlled. The '0' fields are assigned here in dp_header_info
  --   in order:
  --   . eth: eth_src_mac
  --   . ip:  ip_total_length,
  --          ip_header_checksum (not here, will be filled in by 1GbE eth component),
  --          ip_src_addr
  --   . udp: udp_src_port,
  --          udp_total_length
  --          udp_checksum, (not here, also not calculated in 1GbE eth component because would
  --            require store and forward, send default 0)
  --   . app: - sdp_marker, sdp_observation_id, sdp_station_id,
  --          - sdp_source_info_antenna_band_id,
  --            sdp_source_info_nyquist_zone_id,
  --            sdp_source_info_f_adc,
  --            sdp_source_info_fsub_type,
  --            sdp_source_info_payload_error,
  --            sdp_source_info_beam_repositioning_flag,
  --            sdp_source_info_weighted_subbands_flag,
  --            sdp_source_info_gn_id,
  --          - sdp_integration_interval, sdp_data_id, sdp_nof_signal_inputs,
  --            sdp_nof_bytes_per_statistic,
  --            sdp_nof_statistics_per_packet, sdp_block_period
  --          - dp_bsn
  -------------------------------------------------------------------------------
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "eth_src_mac"                             ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "eth_src_mac"                             )) <= eth_src_mac;
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "ip_total_length"                         ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "ip_total_length"                         )) <= TO_UVEC(c_ip_total_length, 16);
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "ip_src_addr"                             ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "ip_src_addr"                             )) <= ip_src_addr;
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "udp_src_port"                            ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "udp_src_port"                            )) <= udp_src_port;
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "udp_total_length"                        ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "udp_total_length"                        )) <= TO_UVEC(c_udp_total_length, 16);
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "sdp_marker"                              ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "sdp_marker"                              )) <= TO_UVEC(c_marker, 8);
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "sdp_observation_id"                      ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "sdp_observation_id"                      )) <= sdp_info.observation_id;
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "sdp_station_id"                          ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "sdp_station_id"                          )) <= sdp_info.station_id;
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "sdp_source_info_antenna_band_id"         ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "sdp_source_info_antenna_band_id"         )) <= SLV(sdp_info.antenna_band_index);
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "sdp_source_info_nyquist_zone_id"         ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "sdp_source_info_nyquist_zone_id"         )) <= sdp_info.nyquist_zone_index;
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "sdp_source_info_f_adc"                   ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "sdp_source_info_f_adc"                   )) <= SLV(sdp_info.f_adc);
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "sdp_source_info_fsub_type"               ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "sdp_source_info_fsub_type"               )) <= SLV(sdp_info.fsub_type);
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "sdp_source_info_payload_error"           ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "sdp_source_info_payload_error"           )) <= SLV(r.payload_err);
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "sdp_source_info_beam_repositioning_flag" ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "sdp_source_info_beam_repositioning_flag" )) <= SLV(sdp_info.beam_repositioning_flag);
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "sdp_source_info_weighted_subbands_flag"  ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "sdp_source_info_weighted_subbands_flag"  )) <= SLV(weighted_subbands_flag);
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "sdp_source_info_gn_id"                   ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "sdp_source_info_gn_id"                   )) <= TO_UVEC(gn_index, 5);
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "sdp_integration_interval"                ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "sdp_integration_interval"                )) <= TO_UVEC(r.integration_interval, 24);
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "sdp_data_id"                             ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "sdp_data_id"                             )) <= data_id_slv;
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "sdp_nof_signal_inputs"                   ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "sdp_nof_signal_inputs"                   )) <= TO_UVEC(c_nof_signal_inputs, 8);
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "sdp_nof_bytes_per_statistic"             ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "sdp_nof_bytes_per_statistic"             )) <= TO_UVEC(c_sdp_nof_bytes_per_statistic, 8);
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "sdp_nof_statistics_per_packet"           ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "sdp_nof_statistics_per_packet"           )) <= TO_UVEC(c_nof_statistics_per_packet, 16);
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "sdp_block_period"                        ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "sdp_block_period"                        )) <= sdp_info.block_period;
  dp_header_info(field_hi(c_sdp_stat_hdr_field_arr, "dp_bsn"                                  ) DOWNTO field_lo(c_sdp_stat_hdr_field_arr,  "dp_bsn"                                  )) <= bsn_at_sync;

  p_reg : PROCESS(dp_rst, dp_clk)
  BEGIN
    IF dp_rst='1' THEN
      r <= c_reg_rst;
    ELSIF rising_edge(dp_clk) THEN
      r <= nxt_r;
    END IF;
  END PROCESS;

  -- Derive and pipeline dynamic parameters
  p_parameters : PROCESS(dp_clk)
  BEGIN
    IF rising_edge(dp_clk) THEN
      gn_index_reg     <= gn_index;
      pn_index         <= func_sdp_gn_index_to_pn_index(gn_index_reg);
      offset_rn        <= TO_UINT(ring_info.O_rn);
      rn_index         <= gn_index_reg - offset_rn;
      local_si_offset  <= pn_index * c_sdp_S_pn;
      nof_cycles_dly   <= gn_index_reg * g_offload_time;
      nof_rn           <= TO_UINT(ring_info.N_rn);
      nof_used_P_sq    <= smallest(nof_rn / 2 + 1, g_P_sq);
      nof_packets      <= func_sdp_get_stat_nof_packets(g_statistics_type, c_sdp_S_pn, nof_used_P_sq, r.nof_crosslets);
      remote_rn        <= func_ring_nof_hops_to_source_rn(r.instance_count, rn_index, nof_rn, g_crosslets_direction);
      remote_gn        <= offset_rn + remote_rn;
      remote_pn        <= func_sdp_gn_index_to_pn_index(remote_gn);
      remote_si_offset <= remote_pn * c_sdp_S_pn;
    END IF;
  END PROCESS;

  -- Assign application header data_id for different statistic types, use
  -- GENERATE to keep unused fields at 0.
  gen_data_id_sst : IF g_statistics_type = "SST" GENERATE
    data_id_rec.sst_signal_input_index <= r.packet_count + local_si_offset;
  END GENERATE;
  gen_data_id_bst : IF g_statistics_type = "BST" GENERATE
    data_id_rec.bst_beamlet_index <= c_beamlet_id;
  END GENERATE;
  gen_data_id_xst : IF g_statistics_type = "XST" GENERATE
    data_id_rec.xst_subband_index <= func_sdp_modulo_N_sub(r.crosslets_info_rec.offset_arr(r.crosslet_count));
    data_id_rec.xst_signal_input_A_index <= local_si_offset;
    data_id_rec.xst_signal_input_B_index <= remote_si_offset;
  END GENERATE;

  data_id_slv <= func_sdp_map_stat_data_id(g_statistics_type, data_id_rec);

  p_control_packet_offload : PROCESS(r, in_sosi, local_si_offset, trigger_offload, nof_crosslets, crosslets_info, nof_packets, dp_sop, dp_header_info)
    VARIABLE v       : t_reg;
    VARIABLE v_index : NATURAL;
  BEGIN
    v := r;
    v.start_pulse := '0';
    v.sync := '0';
    
    -- Count number of sop in a sync interval and get payload errors and keep them till next sync.
    IF in_sosi.sync = '1' THEN
      v.integration_interval := r.in_sop_cnt + 1;  -- count = index + 1
      v.in_sop_cnt := 0;
      v.payload_err := '0';
    ELSE
      IF in_sosi.eop = '1' THEN
        v.payload_err := r.payload_err OR in_sosi.err(0);
      END IF;

      IF in_sosi.sop = '1' THEN
        v.in_sop_cnt := r.in_sop_cnt + 1;
      END IF;
    END IF;


    -- For XST offload capture nof_crosslets and crosslets_info at in_sosi.sync,
    -- to make sure they do not change during packets offload.
    -- . The sdp_crosslets_subband_select.vhd takes in [2] takes care that
    --   nof_crosslets and crosslets_info are valid at the xsel_sosi.sync. The
    --   mmp_dp_bsn_align_v2 in [2] then aligns the local xsel_sosi with the
    --   remote data and passes on the sync. After some latency the sync
    --   arrives at the sdp_statistics_offload. This latency is very short
    --   compared to the sync period, so the nof_crosslets and crosslet_info
    --   are still valid at the in_sosi.sync.
    IF in_sosi.sync = '1' THEN
      v.nof_crosslets      := TO_UINT(nof_crosslets);
      v.crosslets_info_rec := func_sdp_map_crosslets_info(crosslets_info);
    END IF;

    -- The trigger_offload occurs nof_cycles_dly after the in_sosi.sync and the
    -- offload will have finished before the next in_sosi.sync, because
    -- c_sdp_offload_time is such that all offload will finish within 100 ms
    -- and the integration interval (= sync interval) is 1 s for SST and BST
    -- and minimal 0.1s (= c_sdp_xst_nof_clk_per_sync_min) for XST.
    -- The trigger_offload initializes the control for the first packet offload
    -- in every sync interval.
    -- . Issue a start_pulse per packet offload. The start_pulse is used by
    --   u_dp_block_from_mm_dc to read the packet from statistics memory.
    IF trigger_offload = '1' THEN
      -- Use trigger_offload to start first packet offload, all
      -- g_statistics_type start from start address 0
      v.start_pulse        := '1';
      v.sync               := '1';
      v.start_address      := 0;
      v.packet_count       := 0;
      v.interleave_count   := 0;  -- only used for SST
      v.interleave_address := 0;  -- only used for SST
      v.crosslet_count     := 0;  -- only used for XST
      v.instance_count     := 0;  -- only used for XST
      v.instance_address   := 0;  -- only used for XST

    -- The dp_sop = '1' when the packet has been read from statistics memory
    -- and is about to get out of the dp_fifo_fill_eop in
    -- u_dp_block_from_mm_dc. The difference between dp_sop and the mm_done
    -- output of u_dp_block_from_mm_dc, is that dp_sop also includes any
    -- dp_fifo_fill_eop latency. This ensures that the dp_sop identifies the
    -- sop of the offload packet. At the dp_sop:
    -- . the dp_header_info per packet offload can be released
    -- . the next packet offload can be prepared
    --
    ELSIF dp_sop = '1' THEN
      -- Release dp_header_info for current packet offload
      v.dp_header_info := dp_header_info;

      -- Start next packets offload.
      IF r.packet_count < nof_packets - 1 THEN
        IF g_statistics_type = "SST" THEN
          --                 step        step        step        step        step        step
          -- start_address :    0,    2, 2048, 2050, 4096, 4098, 6144, 6146, 8192, 8194, 10240, 10242
          v.start_address := r.start_address + c_mm_data_size;  -- default step to next packet in this step
          v.interleave_count := r.interleave_count + 1;
          IF r.interleave_count = c_sdp_Q_fft - 1 THEN
            -- jump to first packet for next step,
            v.start_address := r.interleave_address + c_sdp_N_sub * c_sdp_Q_fft * c_sdp_W_statistic_sz;  -- = + 512 * 2 * 2 = + 2048
            v.interleave_count := 0;
            v.interleave_address := v.start_address;
          END IF;
          v.start_pulse := '1';
          v.packet_count := r.packet_count + 1;

        ELSIF g_statistics_type = "BST" THEN
          NULL; -- there is only one BST packet, so no more packets to offload here.

        ELSIF g_statistics_type = "XST" THEN
          -- start_address:
          --   nof_crosslets:     0,     1,     2,     3,     4,     5,     6
          --   X_sq instance:
          --           0          0,   576,  1152,  1728,  2304,  2880,  3456
          --           1       4096,  4672,  5248,  5824,  6400,  6976,  7552
          --           2       8192,  8768,  9344,  9920, 10496, 11072, 11648
          --           3      12288, 12864, 13440, 14016, 14592, 15168, 15744
          --           4      16384, 16960, 17536, 18112, 18688, 19264, 19840
          --           5      20480, 21056, 21632, 22208, 22784, 23360, 23936
          --           6      24576, 25152, 25728, 26304, 26880, 27456, 28032
          --           7      28672, 29248, 29824, 30400, 30976, 31552, 32128
          --           8      32768, 33344, 33920, 34496, 35072, 35648, 36224
          v.start_address := r.start_address + c_sdp_X_sq * c_nof_complex * c_sdp_W_statistic_sz;  -- continue with next packet in this instance
          v.crosslet_count := r.crosslet_count + 1;
          IF r.crosslet_count = TO_UINT(nof_crosslets) - 1 THEN
            v.start_address := r.instance_address + 2**c_sdp_ram_st_xsq_addr_w;  -- jump to first packet in next instance
            v.crosslet_count := 0;
            v.instance_count := r.instance_count + 1;
            v.instance_address := v.start_address;  -- use v.start_address to avoid multipier needed in (r.instance_count + 1) * 2**c_sdp_ram_st_xsq_addr_w
          END IF;
          v.start_pulse := '1';
          v.packet_count := r.packet_count + 1;

        ELSE
          NULL;  -- do nothing in case of unknown g_statistics_type
        END IF;
      END IF;
    END IF;

    nxt_r <= v;
  END PROCESS;

  u_mms_common_variable_delay : ENTITY common_lib.mms_common_variable_delay
  PORT MAP (
    mm_rst => mm_rst,
    mm_clk => mm_clk,
    dp_rst => dp_rst,
    dp_clk => dp_clk,

    -- MM interface
    reg_enable_mosi => reg_enable_mosi,
    reg_enable_miso => reg_enable_miso,

    delay           => nof_cycles_dly,
    trigger         => in_sosi.sync,
    trigger_en      => trigger_en,
    trigger_dly     => trigger_offload
  );
  
  u_dp_block_from_mm_dc : ENTITY dp_lib.dp_block_from_mm_dc
  GENERIC MAP (
    g_user_size          => c_mm_user_size,
    g_data_size          => c_mm_data_size,
    g_step_size          => c_mm_step_size,
    g_nof_data           => c_mm_nof_data,
    g_word_w             => c_word_w,
    g_reverse_word_order => g_reverse_word_order,
    g_bsn_w              => c_dp_stream_bsn_w
  ) 
  PORT MAP(
    dp_rst        => dp_rst,
    dp_clk        => dp_clk,
    mm_rst        => mm_rst,
    mm_clk        => mm_clk,
    start_pulse   => r.start_pulse,
    sync_in       => r.sync,
    bsn_at_sync   => bsn_at_sync, 
    start_address => r.start_address,
    done          => mm_done,  -- not used, use dp_sop instead
    mm_mosi       => master_mosi,
    mm_miso       => master_miso,
    out_sosi      => dp_block_from_mm_src_out,
    out_siso      => dp_block_from_mm_src_in
  );

  -- Use dp_block_from_mm_src_out.sop as dp_sop, to include the
  -- dp_fifo_fill_eop that is in dp_block_from_mm_dc. The dp_sop thus is the
  -- sop of the packet that is about to be offloaded by u_dp_offload_tx_v3.
  -- The r.dp_header_info must be available at the dp_offload_snk_in.sop.
  -- This is guaranteed because:
  -- . r.dp_header_info is available one clock cycle after dp_sop in
  --   p_control_packet_offload.
  -- . The dp_offload_snk_in is delayed also by at least one clock cycle by
  --   u_dp_pipeline_ready.
  dp_sop <= dp_block_from_mm_src_out.sop;

  u_dp_pipeline_ready : ENTITY dp_lib.dp_pipeline_ready
  PORT MAP(
    rst          => dp_rst,
    clk          => dp_clk,
    -- ST sink
    snk_out      => dp_block_from_mm_src_in,
    snk_in       => dp_block_from_mm_src_out,
    -- ST source
    src_in       => dp_offload_snk_out,
    src_out      => dp_offload_snk_in
  );

  u_dp_offload_tx_v3: ENTITY dp_lib.dp_offload_tx_v3
  GENERIC MAP (
    g_nof_streams    => c_nof_streams,
    g_data_w         => c_word_w,
    g_symbol_w       => c_word_w,
    g_hdr_field_arr  => c_sdp_stat_hdr_field_arr,
    g_hdr_field_sel  => c_sdp_stat_hdr_field_sel,
    g_pipeline_ready => TRUE
  )
  PORT MAP(
    mm_rst               => mm_rst,
    mm_clk               => mm_clk,
    dp_rst               => dp_rst,
    dp_clk               => dp_clk,
    reg_hdr_dat_mosi     => reg_hdr_dat_mosi,
    reg_hdr_dat_miso     => reg_hdr_dat_miso,
    snk_in_arr(0)        => dp_offload_snk_in,
    snk_out_arr(0)       => dp_offload_snk_out,
    src_out_arr(0)       => udp_sosi,
    src_in_arr(0)        => out_siso,
    hdr_fields_in_arr(0) => r.dp_header_info
  );

  out_sosi <= udp_sosi;

  u_bsn_mon_udp : ENTITY dp_lib.mms_dp_bsn_monitor_v2
  GENERIC MAP (
    g_nof_streams        => 1,  
    g_cross_clock_domain => TRUE,
    g_sync_timeout       => c_sdp_N_clk_sync_timeout,
    g_bsn_w              => c_dp_stream_bsn_w,
    g_error_bi           => 0,
    g_cnt_sop_w          => c_word_w,
    g_cnt_valid_w        => c_word_w,
    g_cnt_latency_w      => c_word_w
  )
  PORT MAP (
    -- Memory-mapped clock domain
    mm_rst         => mm_rst,
    mm_clk         => mm_clk,
    reg_mosi       => reg_bsn_monitor_v2_offload_copi,
    reg_miso       => reg_bsn_monitor_v2_offload_cipo,

    -- Streaming clock domain
    dp_rst         => dp_rst,
    dp_clk         => dp_clk,
    ref_sync       => in_sosi.sync,

    in_sosi_arr(0) => udp_sosi
  );




END str;
