-------------------------------------------------------------------------------
--
-- Copyright 2021
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
-- Author : R vd Walle
-- Purpose: Verify sdp_crosslets_subband_select.
-- Usage:
-- > as 10
-- > run -all
-- * The tb is self stopping and self checking,tb_end will stop the simulation by
--   stopping the clk and thus all toggling.
--
-- Description: The tb starts the dut by writing a scheduled bsn to the bsn_scheduler
-- via MM. The offsets and step are configured using MM. The dut makes the subband 
-- selection based on the MM configuration and N_crosslets. The TB then verifies out_sosi 
-- and out_crosslets_info of the dut by comparing it to the expected output.

LIBRARY IEEE, common_lib, dp_lib;
USE IEEE.std_logic_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.tb_common_pkg.ALL;
USE common_lib.tb_common_mem_pkg.ALL;
USE common_lib.common_lfsr_sequences_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE dp_lib.tb_dp_pkg.ALL;
USE work.sdp_pkg.ALL;

ENTITY tb_sdp_crosslets_subband_select IS

END tb_sdp_crosslets_subband_select;


ARCHITECTURE tb OF tb_sdp_crosslets_subband_select IS

  CONSTANT c_clk_period           : TIME := 10 ns;
  CONSTANT c_mm_clk_period        : TIME := 20 ns;
  CONSTANT c_rl                   : NATURAL := 1;
  CONSTANT c_nof_sync             : NATURAL := 5;
  CONSTANT c_nof_block_per_sync   : NATURAL := 4;
  CONSTANT c_dsp_data_w           : NATURAL := c_sdp_W_subband;

  CONSTANT c_N_crosslets          : NATURAL := 2;
  CONSTANT c_crosslet_offsets     : t_natural_arr(0 TO c_N_crosslets-1) := (0, 15);
  CONSTANT c_crosslet_step        : NATURAL := 3; -- offset step size to increase per sync interval

  CONSTANT c_nof_ch_in            : NATURAL := 1024; -- nof input words per block, identical for all input streams.
  CONSTANT c_nof_ch_sel_row       : NATURAL := c_sdp_P_pfb;
  CONSTANT c_nof_ch_sel_col       : NATURAL := c_sdp_Q_fft; -- nof of sequential columns to select per row.
  CONSTANT c_nof_ch_sel           : NATURAL := c_N_crosslets*c_nof_ch_sel_col*c_nof_ch_sel_row;

  CONSTANT c_ctrl_interval_size   : NATURAL := c_nof_block_per_sync * c_nof_ch_in;
  CONSTANT c_scheduled_bsn        : NATURAL := 11;
  CONSTANT c_nof_block_dly        : NATURAL := c_nof_block_per_sync;
 
  SIGNAL tb_end             : STD_LOGIC;
  SIGNAL rst                : STD_LOGIC;
  SIGNAL clk                : STD_LOGIC := '1'; 
  SIGNAL mm_clk             : STD_LOGIC := '1'; 

  SIGNAL mm_mosi             : t_mem_mosi := c_mem_mosi_rst;
  SIGNAL mm_miso             : t_mem_miso;
  SIGNAL rd_crosslet_offsets : t_natural_arr(0 TO c_N_crosslets-1) := (0, 15);
  SIGNAL rd_crosslet_step    : NATURAL;

  SIGNAL mm_trigger_mosi    : t_mem_mosi := c_mem_mosi_rst;
  SIGNAL mm_trigger_miso    : t_mem_miso;

  SIGNAL st_en              : STD_LOGIC := '1';
  SIGNAL st_siso_arr        : t_dp_siso_arr(c_sdp_P_pfb-1 DOWNTO 0) := (OTHERS => c_dp_siso_rdy);
  SIGNAL st_sosi_arr        : t_dp_sosi_arr(c_sdp_P_pfb-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);
  SIGNAL exp_sosi           : t_dp_sosi := c_dp_sosi_rst;
  
  SIGNAL bsn                : NATURAL := c_scheduled_bsn - c_nof_block_dly;
  
  SIGNAL in_sosi_arr        : t_dp_sosi_arr(c_sdp_P_pfb-1 DOWNTO 0) := (OTHERS => c_dp_sosi_rst);
  
  SIGNAL out_sosi           : t_dp_sosi;

  SIGNAL exp_crosslets_info_slv : STD_LOGIC_VECTOR(c_sdp_crosslets_info_reg_w-1 DOWNTO 0) := (OTHERS => '0');
  SIGNAL out_crosslets_info_slv : STD_LOGIC_VECTOR(c_sdp_crosslets_info_reg_w-1 DOWNTO 0) := (OTHERS => '0');
  SIGNAL exp_crosslets_info_rec : t_sdp_crosslets_info;
  SIGNAL out_crosslets_info_rec : t_sdp_crosslets_info;

BEGIN

  clk <= (NOT clk) OR tb_end AFTER c_clk_period/2;
  mm_clk <= (NOT mm_clk) OR tb_end AFTER c_mm_clk_period/2;
  rst <= '1', '0' AFTER c_clk_period*7;           
  
  p_mm_stimuli : PROCESS
  VARIABLE k : NATURAL;
  BEGIN
    proc_common_wait_until_low(mm_clk, rst);
    proc_common_wait_some_cycles(mm_clk, 50); -- Give dut some time to start

    -- Set BSN sync scheduler
    proc_mem_mm_bus_wr(1, c_ctrl_interval_size, mm_clk, mm_trigger_miso, mm_trigger_mosi);  
    proc_mem_mm_bus_wr(2, c_scheduled_bsn, mm_clk, mm_trigger_miso, mm_trigger_mosi);  
    proc_mem_mm_bus_wr(3, 0, mm_clk, mm_trigger_miso, mm_trigger_mosi); 
    proc_mem_mm_bus_wr(0, 1, mm_clk, mm_trigger_miso, mm_trigger_mosi); --enable
    
    -- Set crosslet info
    FOR I IN 0 TO c_N_crosslets-1 LOOP
      proc_mem_mm_bus_wr(I, c_crosslet_offsets(I), mm_clk, mm_miso, mm_mosi); --offsets
    END LOOP;
    proc_mem_mm_bus_wr(15, c_crosslet_step, mm_clk, mm_miso, mm_mosi); --step
    proc_common_wait_cross_clock_domain_latency(c_clk_period, c_mm_clk_period);

    -- Verify that MM reads the active crosslets_info
    -- a) Readback crosslet info after rst release
    FOR I IN 0 TO c_N_crosslets-1 LOOP
      proc_mem_mm_bus_rd(I, mm_clk, mm_miso, mm_mosi); --offsets
      proc_mem_mm_bus_rd_latency(1, mm_clk);
      rd_crosslet_offsets(I) <= TO_UINT(mm_miso.rddata(c_word_w-1 DOWNTO 0));
    END LOOP;
    proc_mem_mm_bus_rd(15, mm_clk, mm_miso, mm_mosi);
    proc_mem_mm_bus_rd_latency(1, mm_clk);
    rd_crosslet_step <= TO_UINT(mm_miso.rddata(c_word_w-1 DOWNTO 0));
    proc_common_wait_some_cycles(mm_clk, 1);
    -- Verify that readback crosslet info is active crosslets_info
    FOR I IN 0 TO c_N_crosslets-1 LOOP
      ASSERT rd_crosslet_offsets(I) = 0 REPORT "Wrong crosslet offset after rst." SEVERITY ERROR;
    END LOOP;
    ASSERT rd_crosslet_step = 0 REPORT "Wrong crosslet step after rst." SEVERITY ERROR;

    -- b) Read crosslet_info in every sync interval
    WHILE TRUE LOOP
      proc_common_wait_until_hi_lo(clk, out_sosi.sync);
      proc_common_wait_cross_clock_domain_latency(c_clk_period, c_mm_clk_period);
      -- Readback crosslet info
      FOR I IN 0 TO c_N_crosslets-1 LOOP
        proc_mem_mm_bus_rd(I, mm_clk, mm_miso, mm_mosi); --offsets
        proc_mem_mm_bus_rd_latency(1, mm_clk);
        rd_crosslet_offsets(I) <= TO_UINT(mm_miso.rddata(c_word_w-1 DOWNTO 0));
      END LOOP;
      proc_mem_mm_bus_rd(15, mm_clk, mm_miso, mm_mosi);
      proc_mem_mm_bus_rd_latency(1, mm_clk);
      rd_crosslet_step <= TO_UINT(mm_miso.rddata(c_word_w-1 DOWNTO 0));
      proc_common_wait_some_cycles(mm_clk, 1);
      -- Verify that readback crosslet info is active crosslets_info
      FOR I IN 0 TO c_N_crosslets-1 LOOP
        ASSERT rd_crosslet_offsets(I) = exp_crosslets_info_rec.offset_arr(I) REPORT "Wrong active crosslet offset in output sync interval." SEVERITY ERROR;
      END LOOP;
      ASSERT rd_crosslet_step = exp_crosslets_info_rec.step REPORT "Wrong active crosslet step in output sync interval." SEVERITY ERROR;
    END LOOP;

    WAIT;
  END PROCESS;

  ------------------------------------------------------------------------------
  -- Data blocks
  ------------------------------------------------------------------------------
  gen_stimuli : FOR K IN 0 TO c_sdp_P_pfb-1 GENERATE
    p_st_stimuli : PROCESS
      VARIABLE v_re  : NATURAL := 0+k*2**5;
      VARIABLE v_im  : NATURAL := 1+k*2**5;
    BEGIN  
      tb_end <= '0';
      st_sosi_arr(K) <= c_dp_sosi_rst;
      proc_common_wait_until_low(clk, rst);
    
      -- Run some sync intervals with DSP counter data for the real and imag fields
      WAIT UNTIL rising_edge(clk);
      FOR I IN 0 TO c_nof_sync-1 LOOP
        proc_dp_gen_block_data(c_rl, FALSE, c_dsp_data_w, c_dsp_data_w, 0, v_re, v_im, c_nof_ch_in, 0, 0, '1', "0", clk, st_en, st_siso_arr(K), st_sosi_arr(K));     -- next sync
        v_re := v_re + c_nof_ch_in;
        v_im := v_im + c_nof_ch_in;
        FOR J IN 0 TO c_nof_block_per_sync-2 LOOP  -- provide sop and eop for block reference
          proc_dp_gen_block_data(c_rl, FALSE, c_dsp_data_w, c_dsp_data_w, 0, v_re, v_im, c_nof_ch_in, 0, 0, '0', "0", clk, st_en, st_siso_arr(K), st_sosi_arr(K));   -- no sync
          v_re := v_re + c_nof_ch_in;
          v_im := v_im + c_nof_ch_in;
        END LOOP;
      END LOOP;
      st_sosi_arr(K) <= c_dp_sosi_rst;
      proc_common_wait_some_cycles(clk, c_nof_ch_in);
      proc_common_wait_some_cycles(clk, 10);
      tb_end <= '1';
      WAIT;
    END PROCESS;
  END GENERATE;
  
  -- Time stimuli  
  bsn <= bsn + 1 WHEN rising_edge(clk) AND (st_sosi_arr(0).eop='1'); 

  -- Add BSN to the ST data
  p_in_sosi : PROCESS(st_sosi_arr, bsn)
  BEGIN    
    FOR I IN 0 TO c_sdp_P_pfb-1 LOOP
      in_sosi_arr(I)     <= st_sosi_arr(I);
      in_sosi_arr(I).bsn <= TO_DP_BSN(bsn);
    END LOOP;
  END PROCESS;

  ------------------------------------------------------------------------------
  -- Verification
  ------------------------------------------------------------------------------
  p_generate_exp_data : PROCESS
    VARIABLE v_col : NATURAL := 0;
    VARIABLE v_row : NATURAL := 0;
    VARIABLE v_offset : NATURAL := 0; 
    VARIABLE v_sync_ix : NATURAL := 0;
    VARIABLE v_k : NATURAL := 0;
  BEGIN
    FOR I IN 0 TO c_nof_sync*c_nof_block_per_sync-1 LOOP
      v_sync_ix := I / c_nof_block_per_sync;
      exp_sosi <= c_dp_sosi_rst;
      WAIT UNTIL rising_edge(out_sosi.sop);

      exp_crosslets_info_slv(c_sdp_crosslets_info_reg_w-1 DOWNTO c_sdp_crosslets_info_reg_w - c_sdp_crosslets_index_w) <= TO_UVEC(c_crosslet_step, c_sdp_crosslets_index_w);
      FOR C IN 0 TO c_nof_ch_sel_col-1 LOOP
        exp_crosslets_info_slv((C+1)*c_sdp_crosslets_index_w-1 DOWNTO C*c_sdp_crosslets_index_w) <= TO_UVEC(c_crosslet_offsets(C) + v_sync_ix * c_crosslet_step, c_sdp_crosslets_index_w);
      END LOOP;

      FOR J IN 0 TO c_nof_ch_sel-1 LOOP
        v_offset := J / (c_nof_ch_sel_col*c_nof_ch_sel_row);
        v_col := J MOD c_nof_ch_sel_col;
        v_row := (J/c_nof_ch_sel_col) MOD c_nof_ch_sel_row;
        v_k := c_nof_ch_sel_col * v_sync_ix * c_crosslet_step;

        exp_sosi <= c_dp_sosi_rst;
        exp_sosi.valid <= '1';
        IF J = 0 THEN
          exp_sosi.sop <= '1';
          IF I MOD c_nof_block_per_sync = 0 THEN
            exp_sosi.sync <= '1';

          END IF;
        ELSIF j = c_nof_ch_sel-1 THEN 
          exp_sosi.eop <= '1';
        END IF;

        exp_sosi.re <= RESIZE_DP_DSP_DATA(TO_DP_DSP_DATA(   (I + c_nof_block_dly) * c_nof_ch_in + v_k + c_nof_ch_sel_col*c_crosslet_offsets(v_offset) + v_col + v_row*2**5)(c_sdp_W_crosslet-1 DOWNTO 0));
        exp_sosi.im <= RESIZE_DP_DSP_DATA(TO_DP_DSP_DATA(1+ (I + c_nof_block_dly) * c_nof_ch_in + v_k + c_nof_ch_sel_col*c_crosslet_offsets(v_offset) + v_col + v_row*2**5)(c_sdp_W_crosslet-1 DOWNTO 0));
        proc_common_wait_some_cycles(clk, 1);

      END LOOP;
      exp_sosi <= c_dp_sosi_rst;
    END LOOP;
    WAIT;
  END PROCESS;

  p_verify_out_data : PROCESS(clk)
  BEGIN
    IF rising_edge(clk) THEN
      ASSERT out_sosi.valid = exp_sosi.valid      REPORT "Wrong out_sosi.valid"      SEVERITY ERROR;
      ASSERT out_sosi.sop   = exp_sosi.sop        REPORT "Wrong out_sosi.sop"        SEVERITY ERROR;
      ASSERT out_sosi.eop   = exp_sosi.eop        REPORT "Wrong out_sosi.eop"        SEVERITY ERROR;
      ASSERT out_sosi.sync  = exp_sosi.sync       REPORT "Wrong out_sosi.sync"       SEVERITY ERROR;
      ASSERT out_crosslets_info_slv = exp_crosslets_info_slv  REPORT "Wrong out_crosslets_info_slv"  SEVERITY ERROR;
      IF exp_sosi.valid = '1' THEN
        ASSERT out_sosi.re  = exp_sosi.re    REPORT "Wrong out_sosi.re"    SEVERITY ERROR;
        ASSERT out_sosi.im  = exp_sosi.im    REPORT "Wrong out_sosi.im"    SEVERITY ERROR;
      END IF;
    END IF;
  END PROCESS;

  u_dut : ENTITY work.sdp_crosslets_subband_select
  GENERIC MAP (
    g_N_crosslets  => c_N_crosslets,
    g_ctrl_interval_size_min => 1
  )
  PORT MAP (
    dp_rst         => rst,
    dp_clk         => clk,

    mm_rst         => rst,
    mm_clk         => mm_clk,
   
    reg_crosslets_info_mosi => mm_mosi,
    reg_crosslets_info_miso => mm_miso,
   
    reg_bsn_sync_scheduler_xsub_mosi  => mm_trigger_mosi,
    reg_bsn_sync_scheduler_xsub_miso  => mm_trigger_miso, 
   
    -- Streaming
    in_sosi_arr => in_sosi_arr,
    out_sosi    => out_sosi,

    out_crosslets_info => out_crosslets_info_slv
  );
 
  -- Map crosslets_info slv to record for easier view in Wave window
  exp_crosslets_info_rec <= func_sdp_map_crosslets_info(exp_crosslets_info_slv);
  out_crosslets_info_rec <= func_sdp_map_crosslets_info(out_crosslets_info_slv);

END tb;
