-------------------------------------------------------------------------------
--
-- Copyright 2020
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--
-- Author: P. Donker

-- Purpose:
-- . test bench for sdp_info.vhd (and sdp_info_reg.vhd)
-- Description:
--
-- https://plm.astron.nl/polarion/#/project/LOFAR2System/workitem?id=LOFAR2-9258
-- https://plm.astron.nl/polarion/#/project/LOFAR2System/workitem?id=LOFAR2-8855
--
-- Remark:
-- .
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.tb_common_pkg.ALL; 
USE common_lib.tb_common_mem_pkg.ALL; 
USE work.sdp_pkg.ALL;


ENTITY tb_sdp_info IS
END tb_sdp_info;

ARCHITECTURE tb OF tb_sdp_info IS
  CONSTANT c_dp_clk_period        : TIME := 5 ns;     -- 200 MHz
  CONSTANT c_mm_clk_period        : TIME := 20 ns;    -- 50 MHz

  -- used mm_adresses on mm bus
  CONSTANT c_mm_addr_block_period            : NATURAL := 0;
  CONSTANT c_mm_addr_beam_repositioning_flag : NATURAL := 1;
  CONSTANT c_mm_addr_fsub_type               : NATURAL := 2;
  CONSTANT c_mm_addr_f_adc                   : NATURAL := 3;
  CONSTANT c_mm_addr_nyquist_zone_index      : NATURAL := 4;
  CONSTANT c_mm_addr_observation_id          : NATURAL := 5;
  CONSTANT c_mm_addr_antenna_band_index      : NATURAL := 6;
  CONSTANT c_mm_addr_station_id              : NATURAL := 7;
  
  SIGNAL tb_end              : STD_LOGIC := '0';
  SIGNAL tb_mm_reg_end       : STD_LOGIC := '0';

  SIGNAL dp_clk              : STD_LOGIC := '1';   -- digital data path clock = 200 MHz (deser factor 4);
  SIGNAL dp_rst              : STD_LOGIC;

  SIGNAL mm_clk              : STD_LOGIC := '1';   -- MM control clock = 50 MHz
  SIGNAL mm_rst              : STD_LOGIC;
  
  SIGNAL reg_mosi            : t_mem_mosi := c_mem_mosi_rst;
  SIGNAL reg_miso            : t_mem_miso;

  -- signals used to change settings of sdp_info 
  SIGNAL gn_index  : NATURAL   := 15; 
  SIGNAL f_adc     : STD_LOGIC := '0'; 
  SIGNAL fsub_type : STD_LOGIC := '0';

  SIGNAL sdp_info  : t_sdp_info; 

  -- signals used for response of mm bus
  SIGNAL mm_natural_response : NATURAL;
BEGIN
  dp_rst <= '1', '0' AFTER c_dp_clk_period*7;
  dp_clk <= (NOT dp_clk) OR tb_end AFTER c_dp_clk_period/2;
  
  mm_rst <= '1', '0' AFTER c_mm_clk_period*7;
  mm_clk <= (NOT mm_clk) OR tb_end AFTER c_mm_clk_period/2;

  p_mm_reg_stimuli : PROCESS
  BEGIN
    reg_mosi <= c_mem_mosi_rst;

    -- initialyze
    proc_common_wait_until_low(mm_clk, mm_rst);
    proc_common_wait_some_cycles(mm_clk, 100);

    -- default all register hold value 0, try to write 1 in all registers
    proc_mem_mm_bus_wr(c_mm_addr_block_period           ,11 ,mm_clk, reg_miso, reg_mosi); -- RO
    proc_mem_mm_bus_wr(c_mm_addr_beam_repositioning_flag,1  ,mm_clk, reg_miso, reg_mosi);
    proc_mem_mm_bus_wr(c_mm_addr_fsub_type              ,1  ,mm_clk, reg_miso, reg_mosi); -- RO
    proc_mem_mm_bus_wr(c_mm_addr_f_adc                  ,1  ,mm_clk, reg_miso, reg_mosi); -- RO
    proc_mem_mm_bus_wr(c_mm_addr_nyquist_zone_index     ,3  ,mm_clk, reg_miso, reg_mosi);
    proc_mem_mm_bus_wr(c_mm_addr_observation_id         ,16 ,mm_clk, reg_miso, reg_mosi);
    proc_mem_mm_bus_wr(c_mm_addr_antenna_band_index     ,1  ,mm_clk, reg_miso, reg_mosi);
    proc_mem_mm_bus_wr(c_mm_addr_station_id             ,17 ,mm_clk, reg_miso, reg_mosi);
    proc_common_wait_cross_clock_domain_latency(dp_clk, mm_clk);

    proc_mem_mm_bus_rd(c_mm_addr_block_period, mm_clk, reg_mosi);
    proc_mem_mm_bus_rd_latency(c_mem_reg_rd_latency, mm_clk);
    mm_natural_response <= TO_UINT(reg_miso.rddata);
    proc_common_wait_some_cycles(mm_clk, 1);
    ASSERT mm_natural_response /= 11 REPORT "Wrong block_period (not read only)" SEVERITY ERROR;

    proc_mem_mm_bus_rd(c_mm_addr_beam_repositioning_flag, mm_clk, reg_mosi);
    proc_mem_mm_bus_rd_latency(c_mem_reg_rd_latency, mm_clk);
    mm_natural_response <= TO_UINT(reg_miso.rddata);
    proc_common_wait_some_cycles(mm_clk, 1);
    ASSERT mm_natural_response = 1 REPORT "Wrong beam_repositioning_flag" SEVERITY ERROR;

    proc_mem_mm_bus_rd(c_mm_addr_fsub_type, mm_clk, reg_mosi);
    proc_mem_mm_bus_rd_latency(c_mem_reg_rd_latency, mm_clk);
    mm_natural_response <= TO_UINT(reg_miso.rddata);  proc_common_wait_some_cycles(mm_clk, 1);
    ASSERT mm_natural_response /= 1 REPORT "Wrong fsub_type (not read only)" SEVERITY ERROR;

    proc_mem_mm_bus_rd(c_mm_addr_f_adc, mm_clk, reg_mosi);
    proc_mem_mm_bus_rd_latency(c_mem_reg_rd_latency, mm_clk);
    mm_natural_response <= TO_UINT(reg_miso.rddata);
    proc_common_wait_some_cycles(mm_clk, 1);
    ASSERT mm_natural_response /= 1 REPORT "Wrong f_adc (not read only)" SEVERITY ERROR;

    proc_mem_mm_bus_rd(c_mm_addr_nyquist_zone_index, mm_clk, reg_mosi);
    proc_mem_mm_bus_rd_latency(c_mem_reg_rd_latency, mm_clk);
    mm_natural_response <= TO_UINT(reg_miso.rddata);
    proc_common_wait_some_cycles(mm_clk, 1);
    ASSERT mm_natural_response = 3 REPORT "Wrong nyquist_zone_index" SEVERITY ERROR;

    proc_mem_mm_bus_rd(c_mm_addr_observation_id, mm_clk, reg_mosi);
    proc_mem_mm_bus_rd_latency(c_mem_reg_rd_latency, mm_clk);
    mm_natural_response <= TO_UINT(reg_miso.rddata);
    proc_common_wait_some_cycles(mm_clk, 1);
    ASSERT mm_natural_response = 16 REPORT "Wrong observation_id" SEVERITY ERROR;

    proc_mem_mm_bus_rd(c_mm_addr_antenna_band_index, mm_clk, reg_mosi);
    proc_mem_mm_bus_rd_latency(c_mem_reg_rd_latency, mm_clk);
    mm_natural_response <= TO_UINT(reg_miso.rddata);
    proc_common_wait_some_cycles(mm_clk, 1);
    ASSERT mm_natural_response = 1 REPORT "Wrong antenna_band_index" SEVERITY ERROR;

    proc_mem_mm_bus_rd(c_mm_addr_station_id, mm_clk, reg_mosi);
    proc_mem_mm_bus_rd_latency(c_mem_reg_rd_latency, mm_clk);
    mm_natural_response <= TO_UINT(reg_miso.rddata);
    proc_common_wait_some_cycles(mm_clk, 1);
    ASSERT mm_natural_response = 17 REPORT "Wrong station_id" SEVERITY ERROR;

    -- check block_period if f_adc and fsub_type are changed
    -- f_adc = '0' and fsub_type = '0' => block_period = 6400
    -- f_adc = '1' and fsub_type = '0' => block_period = 5120
    -- f_adc = '0' and fsub_type = '1' => block_period = 5400
    -- f_adc = '1' and fsub_type = '1' => block_period = 4320
    f_adc <= '0'; fsub_type <= '0';
    proc_common_wait_cross_clock_domain_latency(dp_clk, mm_clk);
    proc_mem_mm_bus_rd(c_mm_addr_block_period, mm_clk, reg_mosi);
    proc_mem_mm_bus_rd_latency(c_mem_reg_rd_latency, mm_clk);
    mm_natural_response <= TO_UINT(reg_miso.rddata(15 DOWNTO 0));
    proc_common_wait_some_cycles(mm_clk, 1);
    ASSERT mm_natural_response = 6400 REPORT "wrong block_period, expected 6400" SEVERITY ERROR;
    f_adc <= '1'; fsub_type <= '0';
    proc_common_wait_cross_clock_domain_latency(dp_clk, mm_clk);
    proc_mem_mm_bus_rd(c_mm_addr_block_period, mm_clk, reg_mosi);
    proc_mem_mm_bus_rd_latency(c_mem_reg_rd_latency, mm_clk);
    mm_natural_response <= TO_UINT(reg_miso.rddata(15 DOWNTO 0));
    proc_common_wait_some_cycles(mm_clk, 1);
    ASSERT mm_natural_response = 5120 REPORT "wrong block_period, expected 5120" SEVERITY ERROR;
    f_adc <= '0'; fsub_type <= '1';
    proc_common_wait_cross_clock_domain_latency(dp_clk, mm_clk);
    proc_mem_mm_bus_rd(c_mm_addr_block_period, mm_clk, reg_mosi);
    proc_mem_mm_bus_rd_latency(c_mem_reg_rd_latency, mm_clk);
    mm_natural_response <= TO_UINT(reg_miso.rddata(15 DOWNTO 0));
    proc_common_wait_some_cycles(mm_clk, 1);
    ASSERT mm_natural_response = 5400 REPORT "wrong block_period, expected 5400" SEVERITY ERROR;
    f_adc <= '1'; fsub_type <= '1';
    proc_common_wait_cross_clock_domain_latency(dp_clk, mm_clk);
    proc_mem_mm_bus_rd(c_mm_addr_block_period, mm_clk, reg_mosi);
    proc_mem_mm_bus_rd_latency(c_mem_reg_rd_latency, mm_clk);
    mm_natural_response <= TO_UINT(reg_miso.rddata(15 DOWNTO 0));
    proc_common_wait_some_cycles(mm_clk, 1);
    ASSERT mm_natural_response = 4320 REPORT "wrong block_period, expected 4320" SEVERITY ERROR;

    proc_common_wait_some_cycles(mm_clk, 100);
    tb_mm_reg_end <= '1';
    WAIT;

  END PROCESS;

  -- check if values in sdp_info match with expected values
  p_sdp_info_stimuli : PROCESS
  BEGIN
    proc_common_wait_until_high(mm_clk, tb_mm_reg_end); -- wait for p_mm_reg_stimuli done

    ASSERT TO_UINT(sdp_info.block_period)       = 4320 REPORT "wrong sdp_info.block_period value"            SEVERITY ERROR;
    ASSERT sdp_info.beam_repositioning_flag     = '1'  REPORT "wrong sdp_info.beam_repositioning_flag value" SEVERITY ERROR;
    ASSERT sdp_info.fsub_type                   = '1'  REPORT "wrong sdp_info.fsub_type value"               SEVERITY ERROR;
    ASSERT sdp_info.f_adc                       = '1'  REPORT "wrong sdp_info.f_adc value"                   SEVERITY ERROR;
    ASSERT TO_UINT(sdp_info.nyquist_zone_index) = 3    REPORT "wrong sdp_info.nyquist_zone_index value"      SEVERITY ERROR;
    ASSERT TO_UINT(sdp_info.observation_id)     = 16   REPORT "wrong sdp_info.observation_id value"          SEVERITY ERROR;
    ASSERT sdp_info.antenna_band_index          = '1'  REPORT "wrong sdp_info.antenna_band_index value"      SEVERITY ERROR;
    ASSERT TO_UINT(sdp_info.station_id)         = 17   REPORT "wrong sdp_info.station_id value"              SEVERITY ERROR;

    proc_common_wait_some_cycles(mm_clk, 100);
    tb_end <= '1';
    WAIT;
  END PROCESS;


  -- SDP info
  u_dut: ENTITY work.sdp_info
    PORT MAP (
      mm_clk    => mm_clk,
      mm_rst    => mm_rst,

      dp_clk    => dp_clk,
      dp_rst    => dp_rst,

      reg_mosi  => reg_mosi,
      reg_miso  => reg_miso,

      gn_index  => gn_index,
      f_adc     => f_adc,
      fsub_type => fsub_type,

      sdp_info  => sdp_info
    );

END tb;
