-------------------------------------------------------------------------------
--
-- Copyright 2020
-- ASTRON (Netherlands Institute for Radio Astronomy) <http://www.astron.nl/>
-- P.O.Box 2, 7990 AA Dwingeloo, The Netherlands
--
-- Licensed under the Apache License, Version 2.0 (the "License");
-- you may not use this file except in compliance with the License.
-- You may obtain a copy of the License at
--
--     http://www.apache.org/licenses/LICENSE-2.0
--
-- Unless required by applicable law or agreed to in writing, software
-- distributed under the License is distributed on an "AS IS" BASIS,
-- WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-- See the License for the specific language governing permissions and
-- limitations under the License.
--
-------------------------------------------------------------------------------

-------------------------------------------------------------------------------
--
-- Author: P. Donker, E. Kooistra

-- Purpose:
-- . test bench for sdp_statistics_offload.vhd
-- Description:
-- 
-- https://support.astron.nl/confluence/display/L2M/L5+SDPFW+Design+Document%3A+Subband+filterbank
-- . See figure 4.8
--
-- Usage:
-- > as 8
-- > run -a
-- . for header: view rx_offload_sosi and the rx_sdp_stat_header.app fields
-- . for payload: view rx_val, rx_data and exp_data
-------------------------------------------------------------------------------

LIBRARY IEEE, common_lib, dp_lib, ring_lib;
USE IEEE.STD_LOGIC_1164.ALL;
USE common_lib.common_pkg.ALL;
USE common_lib.common_mem_pkg.ALL;
USE common_lib.tb_common_pkg.ALL; 
USE common_lib.tb_common_mem_pkg.ALL;
USE common_lib.common_network_layers_pkg.ALL;
USE common_lib.common_field_pkg.ALL;
USE common_lib.common_str_pkg.ALL;
USE dp_lib.dp_stream_pkg.ALL;
USE ring_lib.ring_pkg.ALL;
USE work.sdp_pkg.ALL;
USE work.tb_sdp_pkg.ALL;


ENTITY tb_sdp_statistics_offload IS
  GENERIC (
    -- All
    g_fast_mm_clk              : BOOLEAN := TRUE;  -- When TRUE use 1 GHz mm_clk  to speed up simulation, else use 100 MHz mm_clk
                                                   -- for real speed of u_dp_block_from_mm_dc in sdp_statistics_offload
    g_statistics_type          : STRING := "XST";
    g_offload_time             : NATURAL := 50;
    g_reverse_word_order       : BOOLEAN := TRUE;  -- when TRUE then stream LSB word after MSB word.
    g_gn_index                 : NATURAL := 4;  -- global node (GN) index, must be in range(O_rn, O_rn + N_rn), use > 0 to see effect of g_offload_time
    g_nof_sync                 : NATURAL := 3;  -- simulate some sync periods, choose >= 3
    -- BST
    g_beamset_id               : NATURAL := 0;  -- < c_sdp_N_beamsets
    -- XST
    g_O_rn                     : NATURAL := 0;  -- GN index of first ring node (RN)
    g_N_rn                     : NATURAL := 8;  -- <= c_sdp_N_rn_max = 16, number of nodes in ring
    g_P_sq                     : NATURAL := 9;  -- <= c_sdp_P_sq, nof available correlator cells
    g_nof_crosslets            : NATURAL := 4;  -- <= c_sdp_N_crosslets_max
    g_crosslets_direction      : NATURAL := 1   -- > 0 for crosslet transport in positive direction (incrementing RN), else 0 for negative direction
  );
END tb_sdp_statistics_offload;

ARCHITECTURE tb OF tb_sdp_statistics_offload IS

  CONSTANT c_dp_clk_period : TIME := 5 ns;     -- 200 MHz
  CONSTANT c_mm_clk_period : TIME := sel_a_b(g_fast_mm_clk, 1 , 10) * 1 ns;
  CONSTANT c_mm_dp_clk_ratio : NATURAL := sel_a_b(c_mm_clk_period > c_dp_clk_period, c_mm_clk_period / c_dp_clk_period, 1);

  CONSTANT c_cross_clock_domain_latency : NATURAL := 20;

  CONSTANT c_offload_time              : NATURAL := g_offload_time * g_gn_index;

  -- In this tb simply use fixed network src addresses
  CONSTANT c_eth_src_mac  : STD_LOGIC_VECTOR(c_network_eth_mac_addr_w-1 DOWNTO 0) := x"123456789ABC";
  CONSTANT c_ip_src_addr  : STD_LOGIC_VECTOR(c_network_ip_addr_w-1 DOWNTO 0) := x"0A090807";
  CONSTANT c_udp_src_port : STD_LOGIC_VECTOR(c_network_udp_port_w-1 DOWNTO 0) := x"D001";

  -- Used mm_adresses on mm bus "enable_mosi/miso".
  CONSTANT c_reg_enable_mm_addr_enable : NATURAL := 0;
  
  -- header fields
  CONSTANT c_nof_statistics_per_packet : NATURAL := func_sdp_get_stat_nof_statistics_per_packet(g_statistics_type);
  CONSTANT c_udp_total_length          : NATURAL := func_sdp_get_stat_udp_total_length(g_statistics_type);
  CONSTANT c_ip_total_length           : NATURAL := func_sdp_get_stat_ip_total_length(g_statistics_type);
  CONSTANT c_marker                    : NATURAL := func_sdp_get_stat_marker(g_statistics_type);
  CONSTANT c_nof_signal_inputs         : NATURAL := func_sdp_get_stat_nof_signal_inputs(g_statistics_type);
  CONSTANT c_nof_packets_max           : NATURAL := func_sdp_get_stat_nof_packets(g_statistics_type, c_sdp_S_pn, g_P_sq, c_sdp_N_crosslets_max);

  CONSTANT c_exp_sdp_info  :  t_sdp_info := (TO_UVEC(601, 16),   -- station_id
                                              '0',                -- antenna_band_index
                                              x"FFFFFFFF",        -- observation_id
                                              b"01",              -- nyquist_zone_index, 0 = first, 1 = second, 2 = third
                                              '1',                -- f_adc, 0 = 160 MHz, 1 = 200 MHz
                                              '0',                -- fsub_type, 0 = critically sampled, 1 = oversampled
                                              '0',                -- beam_repositioning_flag
                                              x"1400"             -- block_period = 5120
                                            );

  CONSTANT c_exp_ring_info  :  t_ring_info := (TO_UVEC(g_O_rn, 8),  -- GN index of first GN in ring
                                               TO_UVEC(g_N_rn, 8),  -- number of GN in ring
                                              '0',                  -- use_cable_to_next_rn
                                              '0'                   -- use_cable_to_previous_rn
                                              );

  CONSTANT c_beamlet_index             : NATURAL := g_beamset_id * c_sdp_S_sub_bf;

  --CONSTANT c_crosslets_info_rec        : t_sdp_crosslets_info := (offset_arr => (1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15), step => 16);
  CONSTANT c_crosslets_info_rec        : t_sdp_crosslets_info := (offset_arr => (0, 1, 2, 3, 4, 5, 6, 10, 11, 12, 13, 14, 15, 16, 17), step => 7);
  CONSTANT c_crosslets_info_slv        : STD_LOGIC_VECTOR(c_sdp_crosslets_info_reg_w-1 DOWNTO 0) := func_sdp_map_crosslets_info(c_crosslets_info_rec);

  -- Crosslets settings
  CONSTANT c_mm_nof_crosslets          : STD_LOGIC_VECTOR(c_sdp_nof_crosslets_reg_w-1 DOWNTO 0) := TO_UVEC(g_nof_crosslets, c_sdp_nof_crosslets_reg_w);
  CONSTANT c_nof_used_P_sq             : NATURAL := smallest(g_N_rn / 2 + 1, g_P_sq);  -- number of used correlator cells <= g_P_sq
  CONSTANT c_rx_nof_packets            : NATURAL := func_sdp_get_stat_nof_packets(g_statistics_type, c_sdp_S_pn, c_nof_used_P_sq, g_nof_crosslets);

  -- payload data
  CONSTANT c_packet_size : NATURAL := c_nof_statistics_per_packet * c_sdp_W_statistic_sz;

  -- Define statistics RAM size for c_nof_packets_max.
  CONSTANT c_ram_size  : NATURAL := c_packet_size * c_nof_packets_max;
  CONSTANT c_ram_w     : NATURAL := ceil_log2(c_ram_size);
  --CONSTANT c_ram_buf   : t_c_mem := (c_mem_ram_rd_latency, c_ram_w,  32, 2**c_ram_w, 'X');
  CONSTANT c_ram_buf   : t_c_mem := (1, c_ram_w,  32, 2**c_ram_w, 'X');

  -- RAM dimensions
  -- . nof_statistics_per_packet = c_mm_nof_data * c_mm_data_size / c_sdp_W_statistic_sz
  CONSTANT c_mm_user_size        : NATURAL := func_sdp_get_stat_from_mm_user_size(g_statistics_type);
  CONSTANT c_mm_data_size        : NATURAL := func_sdp_get_stat_from_mm_data_size(g_statistics_type);
  CONSTANT c_mm_step_size        : NATURAL := func_sdp_get_stat_from_mm_step_size(g_statistics_type);
  CONSTANT c_mm_nof_data         : NATURAL := func_sdp_get_stat_from_mm_nof_data(g_statistics_type);
  CONSTANT c_mm_ram_size         : NATURAL := c_mm_nof_data * c_mm_data_size * c_nof_packets_max;  -- = c_ram_size

  CONSTANT c_mm_nof_step         : NATURAL := c_mm_step_size / c_mm_data_size;
  CONSTANT c_mm_Xsq_span         : NATURAL := 2**ceil_log2(c_sdp_N_crosslets_max * c_packet_size);  -- XST: 2**ceil_log2(7 * 576) = 4096

  -- Define block timing.
  CONSTANT c_bsn_init            : NATURAL := 0;
  -- Sufficient c_nof_block_per_sync to fit more than c_nof_packets_max offload packets per sync interval.
  CONSTANT c_nof_block_per_sync  : NATURAL := 3 + c_mm_dp_clk_ratio * (ceil_div(c_offload_time, c_packet_size) + c_nof_packets_max);
  CONSTANT c_nof_clk_per_block   : NATURAL := c_packet_size;
  CONSTANT c_nof_clk_per_sync    : NATURAL := c_nof_block_per_sync * c_nof_clk_per_block;

  SIGNAL tb_end : STD_LOGIC := '0';

  SIGNAL dp_clk : STD_LOGIC := '1';   -- Digital data path clock = 200 MHz (deser factor 4);
  SIGNAL dp_rst : STD_LOGIC;

  SIGNAL mm_clk : STD_LOGIC := '1';   -- MM control clock = 50 MHz
  SIGNAL mm_rst : STD_LOGIC;
  
  SIGNAL master_mosi : t_mem_mosi := c_mem_mosi_rst;
  SIGNAL master_miso : t_mem_miso;

  SIGNAL enable_mosi : t_mem_mosi := c_mem_mosi_rst;
  SIGNAL enable_miso : t_mem_miso;

  SIGNAL hdr_dat_mosi : t_mem_mosi := c_mem_mosi_rst;
  SIGNAL hdr_dat_miso : t_mem_miso;

  SIGNAL offload_rx_hdr_dat_mosi : t_mem_mosi := c_mem_mosi_rst;
  SIGNAL offload_rx_hdr_dat_miso : t_mem_miso;

  SIGNAL in_sosi                 : t_dp_sosi := c_dp_sosi_rst;
  SIGNAL in_crosslets_info_rec   : t_sdp_crosslets_info;
  SIGNAL in_crosslets_info_slv   : STD_LOGIC_VECTOR(c_sdp_crosslets_info_reg_w-1 DOWNTO 0);

  SIGNAL sdp_offload_data        : STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);  -- 32 bit
  SIGNAL sdp_offload_sosi        : t_dp_sosi;
  SIGNAL sdp_offload_siso        : t_dp_siso := c_dp_siso_rst;

  SIGNAL rx_offload_en           : STD_LOGIC := '0';
  SIGNAL rx_offload_data         : STD_LOGIC_VECTOR(c_word_w-1 DOWNTO 0);  -- 32 bit
  SIGNAL rx_offload_sosi         : t_dp_sosi := c_dp_sosi_rst;
  SIGNAL rx_offload_sop_cnt      : NATURAL := 0;
  SIGNAL rx_offload_eop_cnt      : NATURAL := 0;

  SIGNAL rx_hdr_fields_out       : STD_LOGIC_VECTOR(1023 DOWNTO 0);
  SIGNAL rx_hdr_fields_raw       : STD_LOGIC_VECTOR(1023 DOWNTO 0) := (OTHERS => '0');
  SIGNAL rx_sdp_stat_header      : t_sdp_stat_header;
  SIGNAL exp_sdp_stat_header     : t_sdp_stat_header;

  SIGNAL exp_dp_bsn              : STD_LOGIC_VECTOR(63 DOWNTO 0);
  SIGNAL exp_sst_signal_input    : NATURAL;
  SIGNAL exp_bst_beamlet_index   : NATURAL;
  SIGNAL cur_X_sq_cell           : NATURAL;
  SIGNAL cur_crosslet            : NATURAL;
  SIGNAL exp_subband_index       : NATURAL;
  SIGNAL exp_xst_signal_input_A  : NATURAL;
  SIGNAL exp_xst_signal_input_B  : NATURAL;

  SIGNAL rx_val                  : STD_LOGIC := '0';
  SIGNAL rx_data                 : NATURAL;
  SIGNAL exp_data                : NATURAL;

  SIGNAL gn_index                : NATURAL := g_gn_index;           -- this node GN
  SIGNAL rn_index                : NATURAL := g_gn_index - g_O_rn;  -- this node RN
  SIGNAL source_rn               : NATURAL;  -- source node RN
  SIGNAL source_gn               : NATURAL;  -- source node GN

  SIGNAL weighted_subbands_flag  : STD_LOGIC := '0';

  -- Signals used for starting processes.
  SIGNAL ram_wr_data      : STD_LOGIC_VECTOR(c_ram_buf.dat_w-1 DOWNTO 0);
  SIGNAL ram_wr_addr      : STD_LOGIC_VECTOR(c_ram_buf.adr_w-1 DOWNTO 0);
  SIGNAL ram_wr_en        : STD_LOGIC;
  SIGNAL init_ram_done    : STD_LOGIC := '0';

  SIGNAL rx_sync_cnt      : INTEGER := 0;
  SIGNAL rx_packet_cnt    : NATURAL := 0;
  SIGNAL rx_valid_cnt     : NATURAL := 0;

  -- Debug signals, to view in Wave window
  SIGNAL dbg_c_mm_dp_clk_ratio           : NATURAL := c_mm_dp_clk_ratio;
  SIGNAL dbg_c_nof_statistics_per_packet : NATURAL := c_nof_statistics_per_packet;
  SIGNAL dbg_c_udp_total_length          : NATURAL := c_udp_total_length;
  SIGNAL dbg_c_ip_total_length           : NATURAL := c_ip_total_length;
  SIGNAL dbg_c_marker                    : NATURAL := c_marker;
  SIGNAL dbg_c_nof_signal_inputs         : NATURAL := c_nof_signal_inputs;
  SIGNAL dbg_c_nof_packets_max           : NATURAL := c_nof_packets_max;
  SIGNAL dbg_c_rx_nof_packets            : NATURAL := c_rx_nof_packets;
  SIGNAL dbg_c_beamlet_index             : NATURAL := c_beamlet_index;
  SIGNAL dbg_c_packet_size               : NATURAL := c_packet_size;
  SIGNAL dbg_c_mm_user_size              : NATURAL := c_mm_user_size;
  SIGNAL dbg_c_mm_data_size              : NATURAL := c_mm_data_size;
  SIGNAL dbg_c_mm_step_size              : NATURAL := c_mm_step_size;
  SIGNAL dbg_c_mm_nof_data               : NATURAL := c_mm_nof_data;
  SIGNAL dbg_c_mm_ram_size               : NATURAL := c_mm_ram_size;
  SIGNAL dbg_c_mm_nof_step               : NATURAL := c_mm_nof_step;
  SIGNAL dbg_c_mm_Xsq_span               : NATURAL := c_mm_Xsq_span;
  SIGNAL dbg_c_ram_size                  : NATURAL := c_ram_size;
  SIGNAL dbg_c_crosslets_info_rec        : t_sdp_crosslets_info := c_crosslets_info_rec;
  SIGNAL dbg_c_crosslets_info_slv        : STD_LOGIC_VECTOR(c_sdp_crosslets_info_reg_w-1 DOWNTO 0) := c_crosslets_info_slv;

BEGIN

  -- Check consistency of constant value dependencies
  ASSERT c_ram_size = c_mm_ram_size REPORT "Wrong derivation of RAM size." SEVERITY FAILURE;

  dp_rst <= '1', '0' AFTER c_dp_clk_period*7;
  dp_clk <= (NOT dp_clk) OR tb_end AFTER c_dp_clk_period/2;
  
  mm_rst <= '1', '0' AFTER c_mm_clk_period*7;
  mm_clk <= (NOT mm_clk) OR tb_end AFTER c_mm_clk_period/2;

  -- Fill statistics RAM with data, data is same as address number.
  p_mm_statistics_ram : PROCESS
  BEGIN
    ram_wr_en <= '0';
    -- Initialyze
    proc_common_wait_until_low(mm_clk, mm_rst);
    proc_common_wait_some_cycles(mm_clk, 10);

    FOR i IN 0 TO c_ram_buf.nof_dat-1 LOOP
      ram_wr_addr <= TO_UVEC(i, c_ram_buf.adr_w);
      ram_wr_data <= TO_UVEC(i, c_ram_buf.dat_w);
      ram_wr_en   <= '1';
      proc_common_wait_some_cycles(mm_clk, 1);
    END LOOP;
    ram_wr_en <= '0';

    init_ram_done <= '1';
    WAIT;
  END PROCESS;

  -- Start the input when statistics RAM is initialized
  p_in_sosi : PROCESS
  BEGIN
    proc_common_wait_until_high(mm_clk, init_ram_done);
    proc_common_wait_some_cycles(dp_clk, 10);
    in_sosi.bsn <= TO_DP_BSN(c_bsn_init);
    in_sosi.valid <= '1';
    in_crosslets_info_rec <= c_crosslets_info_rec;
    WHILE TRUE LOOP
      FOR i IN 0 TO c_nof_block_per_sync-1 LOOP
        FOR j IN 0 TO c_nof_clk_per_block-1 LOOP
          in_sosi.sync  <= '0';
          in_sosi.sop   <= '0';
          in_sosi.eop   <= '0';
          IF i = 0 AND j = 0 THEN
            in_sosi.sync <= '1';
          END IF;
          IF i = 0 AND j = 1 THEN
            -- Increment crosslets_info offsets for next sync interval
            in_crosslets_info_rec <= func_sdp_step_crosslets_info(in_crosslets_info_rec);
          END IF;
          IF j = 0 THEN
            in_sosi.sop  <= '1';
            in_sosi.bsn  <= INCR_UVEC(in_sosi.bsn, 1);
          END IF;
          IF j = c_nof_clk_per_block-1 THEN
            in_sosi.eop  <= '1';
          END IF;
          proc_common_wait_some_cycles(dp_clk, 1);
        END LOOP;
      END LOOP;
    END LOOP;
    WAIT;
  END PROCESS;

  in_crosslets_info_slv <= func_sdp_map_crosslets_info(in_crosslets_info_rec);

  -- Enable the statistics offload when input is running
  p_enable_trigger : PROCESS
  BEGIN
    -- Wait at least one sync interval, so that DUT can have measured the integration_interval
    proc_common_wait_until_hi_lo(dp_clk, in_sosi.sync);
    proc_common_wait_some_cycles(mm_clk, 10);
    -- Enable common variable delay.
    proc_mem_mm_bus_wr(c_reg_enable_mm_addr_enable, 1, mm_clk, enable_miso, enable_mosi);
    proc_common_wait_some_cycles(mm_clk, c_cross_clock_domain_latency);
    proc_common_wait_some_cycles(dp_clk, 1);
    rx_offload_en <= '1';
    WAIT;
  END PROCESS;

  -- Counters to time expected exp_sdp_stat_header fields per offload packet
  p_test_counters : PROCESS(dp_clk)
  BEGIN
    IF rising_edge(dp_clk) THEN
      -- Count rx_offload_sosi packets
      IF rx_offload_sosi.sop = '1' THEN
        rx_offload_sop_cnt <= rx_offload_sop_cnt + 1;  -- early count
      END IF;
      IF rx_offload_sosi.eop = '1' THEN
        rx_offload_eop_cnt <= rx_offload_eop_cnt + 1;  -- after count
      END IF;
    END IF;
  END PROCESS;

  rx_sync_cnt <= rx_sync_cnt + 1 WHEN rising_edge(dp_clk) AND rx_offload_sosi.sync = '1';

  -- derive current X_sq correlator cell index
  cur_X_sq_cell <= (rx_offload_eop_cnt / g_nof_crosslets) MOD c_nof_used_P_sq;
  -- derive current N_crosslets index index
  cur_crosslet <= rx_offload_eop_cnt MOD g_nof_crosslets;

  -- derive source RN index
  source_rn <= func_ring_nof_hops_to_source_rn(cur_X_sq_cell, rn_index, g_N_rn, g_crosslets_direction);
  source_gn <= g_O_rn + source_rn;

  -- Prepare exp_sdp_stat_header before rx_offload_sosi.eop, so that p_exp_sdp_stat_header can
  -- verify it at rx_offload_sosi.eop.

  -- For all statistics
  exp_dp_bsn <= TO_SVEC(c_bsn_init + 1 + rx_sync_cnt * c_nof_block_per_sync, 64);
  -- SST
  exp_sst_signal_input <= rx_packet_cnt + c_sdp_S_pn * gn_index;
  -- BST
  exp_bst_beamlet_index <= c_beamlet_index;
  -- XST
  -- . prepare expected XST subband_index
  exp_subband_index <= (c_crosslets_info_rec.offset_arr(cur_crosslet) + rx_sync_cnt * c_crosslets_info_rec.step) MOD c_sdp_N_sub;

  -- . prepare expected XST signal_input_A index
  exp_xst_signal_input_A <= (gn_index MOD c_sdp_N_pn_max) * c_sdp_S_pn;

  -- . prepare expected XST signal_input_B index, assume crosslet transport in positive direction
  exp_xst_signal_input_B <= (source_gn MOD c_sdp_N_pn_max) * c_sdp_S_pn;

  p_exp_sdp_stat_header : PROCESS(weighted_subbands_flag, gn_index, exp_dp_bsn, exp_sst_signal_input, exp_subband_index, exp_xst_signal_input_A, exp_xst_signal_input_B)
  BEGIN
    -- eth header
    exp_sdp_stat_header.eth.dst_mac        <= c_sdp_stat_eth_dst_mac;
    exp_sdp_stat_header.eth.src_mac        <= c_eth_src_mac;
    exp_sdp_stat_header.eth.eth_type       <= x"0800";

    -- ip header
    exp_sdp_stat_header.ip.version         <= TO_UVEC(                4, c_network_ip_version_w);
    exp_sdp_stat_header.ip.header_length   <= TO_UVEC(                5, c_network_ip_header_length_w);
    exp_sdp_stat_header.ip.services        <= TO_UVEC(                0, c_network_ip_services_w);
    exp_sdp_stat_header.ip.total_length    <= TO_UVEC(c_ip_total_length, c_network_ip_total_length_w);
    exp_sdp_stat_header.ip.identification  <= TO_UVEC(                0, c_network_ip_identification_w);
    exp_sdp_stat_header.ip.flags           <= TO_UVEC(                2, c_network_ip_flags_w);
    exp_sdp_stat_header.ip.fragment_offset <= TO_UVEC(                0, c_network_ip_fragment_offset_w);
    exp_sdp_stat_header.ip.time_to_live    <= TO_UVEC(              127, c_network_ip_time_to_live_w);
    exp_sdp_stat_header.ip.protocol        <= TO_UVEC(               17, c_network_ip_protocol_w);
    exp_sdp_stat_header.ip.header_checksum <= TO_UVEC(                0, c_network_ip_header_checksum_w);
    exp_sdp_stat_header.ip.src_ip_addr     <=             c_ip_src_addr;  -- c_network_ip_addr_w
    exp_sdp_stat_header.ip.dst_ip_addr     <=    c_sdp_stat_ip_dst_addr;  -- c_network_ip_addr_w

    -- udp header
    exp_sdp_stat_header.udp.src_port       <=             c_udp_src_port;
    exp_sdp_stat_header.udp.dst_port       <= TO_UVEC(              5001, c_network_udp_port_w);
    exp_sdp_stat_header.udp.total_length   <= TO_UVEC(c_udp_total_length, c_network_udp_port_w);
    exp_sdp_stat_header.udp.checksum       <= TO_UVEC(                 0, c_network_udp_checksum_w);

    -- app header
    exp_sdp_stat_header.app.sdp_marker                              <= TO_UVEC(c_marker, 8);
    exp_sdp_stat_header.app.sdp_version_id                          <= TO_UVEC(c_sdp_stat_version_id, 8);
    exp_sdp_stat_header.app.sdp_observation_id                      <= c_exp_sdp_info.observation_id;
    exp_sdp_stat_header.app.sdp_station_id                          <= c_exp_sdp_info.station_id;

    exp_sdp_stat_header.app.sdp_source_info_antenna_band_id         <= slv(c_exp_sdp_info.antenna_band_index);
    exp_sdp_stat_header.app.sdp_source_info_nyquist_zone_id         <=     c_exp_sdp_info.nyquist_zone_index;
    exp_sdp_stat_header.app.sdp_source_info_f_adc                   <= slv(c_exp_sdp_info.f_adc);
    exp_sdp_stat_header.app.sdp_source_info_fsub_type               <= slv(c_exp_sdp_info.fsub_type);
    exp_sdp_stat_header.app.sdp_source_info_payload_error           <= TO_UVEC(0, 1);
    exp_sdp_stat_header.app.sdp_source_info_beam_repositioning_flag <= slv(c_exp_sdp_info.beam_repositioning_flag);
    exp_sdp_stat_header.app.sdp_source_info_weighted_subbands_flag  <= slv(weighted_subbands_flag);
    exp_sdp_stat_header.app.sdp_source_info_reserved                <= TO_UVEC(0, 3);
    exp_sdp_stat_header.app.sdp_source_info_gn_id                   <= TO_UVEC(gn_index, 5);

    exp_sdp_stat_header.app.sdp_reserved                            <= TO_UVEC(                   0,  8);
    exp_sdp_stat_header.app.sdp_integration_interval                <= TO_UVEC(c_nof_block_per_sync, 24);
    IF g_statistics_type = "SST" THEN
      exp_sdp_stat_header.app.sdp_data_id                           <= TO_UVEC(exp_sst_signal_input, 32);
      exp_sdp_stat_header.app.sdp_data_id_sst_signal_input_index    <= TO_UVEC(exp_sst_signal_input,  8);
    ELSIF g_statistics_type = "BST" THEN
      exp_sdp_stat_header.app.sdp_data_id                           <= TO_UVEC(c_beamlet_index, 32);
      exp_sdp_stat_header.app.sdp_data_id_bst_beamlet_index         <= TO_UVEC(c_beamlet_index, 16);
    ELSIF g_statistics_type = "XST" THEN
      exp_sdp_stat_header.app.sdp_data_id                           <= TO_UVEC(0, 7) & TO_UVEC(exp_subband_index, 9) & TO_UVEC(exp_xst_signal_input_A, 8) & TO_UVEC(exp_xst_signal_input_B, 8);
      exp_sdp_stat_header.app.sdp_data_id_xst_subband_index         <= TO_UVEC(exp_subband_index, 9);
      exp_sdp_stat_header.app.sdp_data_id_xst_signal_input_A_index  <= TO_UVEC(exp_xst_signal_input_A, 8);
      exp_sdp_stat_header.app.sdp_data_id_xst_signal_input_B_index  <= TO_UVEC(exp_xst_signal_input_B, 8);
    END IF;
    exp_sdp_stat_header.app.sdp_nof_signal_inputs                   <= TO_UVEC(          c_nof_signal_inputs,  8);
    exp_sdp_stat_header.app.sdp_nof_bytes_per_statistic             <= TO_UVEC(c_sdp_nof_bytes_per_statistic,  8);
    exp_sdp_stat_header.app.sdp_nof_statistics_per_packet           <= TO_UVEC(  c_nof_statistics_per_packet, 16);
    exp_sdp_stat_header.app.sdp_block_period                        <= c_exp_sdp_info.block_period;

    exp_sdp_stat_header.app.dp_bsn <= exp_dp_bsn;
  END PROCESS;

  rx_sdp_stat_header <= func_sdp_map_stat_header(rx_hdr_fields_raw);

  p_verify_header : PROCESS(rx_offload_sosi)
    VARIABLE v_bool : BOOLEAN;
  BEGIN
    -- Prepare exp_sdp_stat_header before rx_offload_sosi.eop, so that it can be verified at rx_offload_sosi.eop
    IF rx_offload_sosi.eop = '1' THEN
      v_bool := func_sdp_verify_stat_header(g_statistics_type, rx_sdp_stat_header, exp_sdp_stat_header);
    END IF;
  END PROCESS;

  -- Count number of packets in a sync interval.
  p_rx_packet_cnt : PROCESS(dp_clk)
  BEGIN
    IF rising_edge(dp_clk) THEN
      IF rx_offload_sosi.sync = '1' THEN
        rx_packet_cnt <= 0;
      ELSIF rx_offload_sosi.eop = '1' THEN
        rx_packet_cnt <= rx_packet_cnt + 1;
      END IF;
    END IF;
  END PROCESS;

  -- Verify number of packets per sync interval
  p_verify_nof_packets : PROCESS(dp_clk)
  BEGIN
    IF rising_edge(dp_clk) THEN
      IF rx_offload_sosi.sync = '1' AND rx_sync_cnt > 1 THEN
        ASSERT rx_packet_cnt = c_rx_nof_packets REPORT "Wrong number of packets per sync interval" SEVERITY ERROR;
      END IF;
    END IF;
  END PROCESS;

  p_verify_nof_valid_per_packet : PROCESS(dp_clk)
  BEGIN
    IF rising_edge(dp_clk) THEN
      IF rx_offload_sosi.eop = '1' THEN
        rx_valid_cnt <= 0;
        ASSERT rx_valid_cnt = c_packet_size - 1 REPORT "Wrong number of valid per packet" SEVERITY ERROR;
      ELSIF rx_offload_sosi.valid = '1' THEN
        rx_valid_cnt <= rx_valid_cnt + 1;
      END IF;
    END IF;
  END PROCESS;

  p_verify_payload : PROCESS(dp_clk)
    VARIABLE W            : NATURAL;  -- 32bit Words
    VARIABLE D            : NATURAL;  -- Data values of c_mm_data_size
    VARIABLE S            : NATURAL;  -- Statistic values of c_sdp_W_statistic_sz
    VARIABLE P            : NATURAL;  -- Packet count
    VARIABLE I, J, K, L   : NATURAL;  -- auxiliary index
    VARIABLE U            : NATURAL;  -- real sUbband SST values
    VARIABLE B            : NATURAL;  -- dual polarization Beamlet BST values
    VARIABLE X            : NATURAL;  -- complex crosslet XST values
    VARIABLE v_rx_data    : NATURAL;  -- received (rx) 32bit word
    VARIABLE v_exp_data   : NATURAL;  -- expected (exp) 32bit word
  BEGIN
    IF rising_edge(dp_clk) THEN
      rx_val <= '0';
      v_rx_data := TO_UINT(rx_offload_sosi.data);
      IF rx_offload_sosi.valid = '1' THEN
        IF g_statistics_type = "SST" THEN
          --        Indices:
          --         W:    0     1      2     3      4     5 ...  1022  1023
          -- U = D = S:    0            1            2             511
          --         I:    0     1      0     1      0     1 ...     0     1
          --    P: J: Words values:
          --    0  0       0     1      4     5      8     9 ...  2044  2045
          --    1  1       2     3      6     7     10    11 ...  2046  2047
          --    2  0    2048  2049   2052  2053   2056  2057 ...  4092  4093
          --    3  1    2050  2051   2054  2055   2058  2059 ...  4094  4095
          --   ...                                           ...
          --   11  1   10242 10243  10246 10247  10250 10251 ... 12286 12287
          --
          -- g_reverse_word_order = TRUE: swaps odd and even W columns, because
          --   c_mm_user_size = c_sdp_W_statistic_sz = 2

          W := rx_valid_cnt;                -- range c_packet_size = 1024 32bit Words
          S := W / c_sdp_W_statistic_sz;    -- range c_nof_statistics_per_packet = 512 Statistic values
          D := S;                           -- range c_mm_nof_data = 512 Data values, because
                                            -- c_mm_data_size / c_sdp_W_statistic_sz = 1
          U := S;                           -- range c_sdp_N_sub = 512 SST values
          I := W MOD c_mm_user_size;        -- range c_mm_user_size = c_sdp_W_statistic_sz = 2 words
          P := rx_packet_cnt MOD c_rx_nof_packets;  -- range c_nof_packets_max = 12 = c_sdp_S_pn packets
          J := P MOD c_mm_nof_step;         -- range c_mm_nof_step = 2 = c_sdp_Q_fft

          v_exp_data := S * 4;  -- due to c_mm_step_size = 4 = c_sdp_W_statistic_sz * c_sdp_Q_fft;
          IF I = sel_a_b(g_reverse_word_order, 0, 1) THEN
            v_exp_data := v_exp_data + 1;  -- due to c_mm_user_size = 2
          END IF;
          IF J = 1 THEN
            v_exp_data := v_exp_data + 2;  -- due to c_sdp_W_statistic_sz = 2 and c_mm_nof_step = 2 > 1
          END IF;
          v_exp_data := v_exp_data + (P/2) * 2048;  -- due to c_packet_size = 1024 and c_mm_nof_step = 2 > 1
          ASSERT v_exp_data = v_rx_data REPORT "Wrong SST payload data Rx" SEVERITY ERROR;

        ELSIF g_statistics_type = "BST" THEN
          --    Indices:
          --     W:    0     1      2     3      4     5 ...  1948  1949   1950  1951
          --     S:    0            1            2       ...   974          975
          -- B = D:    0                         1       ...   487
          --     I:    0     1      0     1      0       ...     0     1      0     1
          --  P:  Words values:
          --  0        0     1      2     3      4     5 ...  1948  1949   1950  1951
          --
          -- g_reverse_word_order = TRUE: swaps odd and even W columns, because
          --   c_mm_user_size = c_sdp_W_statistic_sz = 2

          W := rx_valid_cnt;                -- range c_packet_size = 1952
          S := W / c_sdp_W_statistic_sz;    -- range c_nof_statistics_per_packet = 976 Statistic values
          D := S / c_sdp_N_pol_bf;          -- range c_mm_nof_data = 488 Data values, because
                                            -- c_mm_data_size / c_sdp_W_statistic_sz = 2 = c_sdp_N_pol_bf
          B := D;                           -- range c_sdp_S_sub_bf = 488 dual polarization BST values
          I := W MOD c_mm_user_size;        -- range c_mm_user_size = c_sdp_W_statistic_sz = 2 words
          P := rx_packet_cnt MOD c_rx_nof_packets;  -- range c_nof_packets_max = 1 packet

          v_exp_data := S * c_mm_user_size;  -- c_mm_user_size = 2
          IF g_reverse_word_order = FALSE THEN
            v_exp_data := v_exp_data + I;
          ELSE
            v_exp_data := v_exp_data - I + c_mm_user_size-1;
          END IF;
          ASSERT v_exp_data = v_rx_data REPORT "Wrong BST payload data Rx" SEVERITY ERROR;

        ELSIF g_statistics_type = "XST" THEN
          -- . c_nof_used_P_sq = 4
          -- . g_nof_crosslets = 3
          -- . c_sdp_N_crosslets_max = 7 --> c_mm_Xsq_span = 2**ceil_log2(7 * 576) = 4096
          --
          --      W:        0     1      2     3      4     5 ...   572   573   574   575
          --      S:        0            1            2       ...   286         287
          --  X = D:        0                         1       ...   143
          --      I:        0     1      0     1      0       ...     0     1     0     1
          --   P: J: K: Word values:
          --   0  0  0      0     1      2     3      4     5 ...   572   573   574   575
          --   1  1       576                                 ...
          --   2  2      1052                                 ...
          --
          --   3  0  1   4096                                 ...
          --   4  1      4672                                 ...
          --   5  2      5244                                 ...
          --   ...
          --   9  0  3  12288                                 ...
          --  10  1     12864                                 ...
          --  11  2     13436                                 ...
          --
          -- g_reverse_word_order = TRUE: swaps odd and even W columns, because
          --   c_mm_user_size = c_sdp_W_statistic_sz = 2

          W := rx_valid_cnt;                -- range c_packet_size = 576
          S := W / c_sdp_W_statistic_sz;    -- range c_nof_statistics_per_packet = 288 Statistic values
          D := S / c_nof_complex;           -- range c_mm_nof_data = 144 Data values, because
                                            -- c_mm_data_size / c_sdp_W_statistic_sz = 2 = c_nof_complex
          X := D;                           -- range c_sdp_X_sq = 144 complex XST values
          I := W MOD c_mm_user_size;        -- range c_mm_user_size = c_sdp_W_statistic_sz = 2 words
          P := rx_packet_cnt MOD c_rx_nof_packets;  -- range c_nof_packets_max = c_nof_used_P_sq * g_nof_crosslets packets
          J := P MOD g_nof_crosslets;       -- range g_nof_crosslets
          K := P / g_nof_crosslets;         -- range c_nof_used_P_sq

          v_exp_data := S * c_mm_user_size;  -- c_mm_user_size = 2
          IF g_reverse_word_order = FALSE THEN
            v_exp_data := v_exp_data + I;
          ELSE
            v_exp_data := v_exp_data - I + c_mm_user_size-1;
          END IF;
          v_exp_data := v_exp_data + J * c_packet_size;  -- c_packet_size = 576
          v_exp_data := v_exp_data + K * c_mm_Xsq_span;  -- c_mm_Xsq_span = 4096
          ASSERT v_exp_data = v_rx_data REPORT "Wrong XST payload data Rx" SEVERITY ERROR;

        END IF;
        -- for debug in wave window
        rx_val <= '1';
        rx_data <= v_rx_data;
        exp_data <= v_exp_data;
      END IF;
    END IF;
  END PROCESS;

  p_dp_end : PROCESS
  BEGIN
    proc_common_wait_until_high(mm_clk, init_ram_done);
    proc_common_wait_some_cycles(dp_clk, g_nof_sync * c_nof_clk_per_sync);  -- will show some sync periods
    tb_end <= '1';
    WAIT;
  END PROCESS;

  u_ram: ENTITY common_lib.common_ram_crw_crw
  GENERIC MAP (
    g_ram => c_ram_buf
  )
  PORT MAP (
    -- MM write port clock domain.
    rst_a    => mm_rst,
    clk_a    => mm_clk,
    wr_en_a  => ram_wr_en,
    wr_dat_a => ram_wr_data,
    adr_a    => ram_wr_addr,

    -- DP read only port clock domain.
    rst_b    => mm_rst,
    clk_b    => mm_clk,
    adr_b    => master_mosi.address(c_ram_buf.adr_w-1 DOWNTO 0),
    rd_en_b  => master_mosi.rd,
    rd_dat_b => master_miso.rddata(c_ram_buf.dat_w-1 DOWNTO 0),
    rd_val_b => master_miso.rdval
  );

  u_rx : ENTITY dp_lib.dp_offload_rx
  GENERIC MAP (
    g_nof_streams         => 1,
    g_data_w              => c_word_w,
    g_hdr_field_arr       => c_sdp_stat_hdr_field_arr,
    g_remove_crc          => FALSE,
    g_crc_nof_words       => 0
  )
  PORT MAP (
    mm_rst                => mm_rst,
    mm_clk                => mm_clk,

    dp_rst                => dp_rst,
    dp_clk                => dp_clk,

    reg_hdr_dat_mosi      => offload_rx_hdr_dat_mosi,
    reg_hdr_dat_miso      => offload_rx_hdr_dat_miso,

    snk_in_arr(0)         => sdp_offload_sosi,
    snk_out_arr(0)        => sdp_offload_siso,

    src_out_arr(0)        => rx_offload_sosi,

    hdr_fields_out_arr(0) => rx_hdr_fields_out,
    hdr_fields_raw_arr(0) => rx_hdr_fields_raw
  );

  -- SDP info
  u_dut: ENTITY work.sdp_statistics_offload
  GENERIC MAP (
    g_statistics_type     => g_statistics_type,
    g_offload_time        => g_offload_time,
    g_reverse_word_order  => g_reverse_word_order,
    g_beamset_id          => g_beamset_id,
    g_P_sq                => g_P_sq,
    g_crosslets_direction => g_crosslets_direction
  )
  PORT MAP (
    mm_clk => mm_clk,
    mm_rst => mm_rst,

    dp_clk => dp_clk,
    dp_rst => dp_rst,

    -- MM
    master_mosi      => master_mosi,
    master_miso      => master_miso,

    reg_enable_mosi  => enable_mosi,
    reg_enable_miso  => enable_miso,

    reg_hdr_dat_mosi => hdr_dat_mosi,
    reg_hdr_dat_miso => hdr_dat_miso,

    -- ST
    in_sosi          => in_sosi,
    out_sosi         => sdp_offload_sosi,
    out_siso         => sdp_offload_siso,

    -- Inputs from other blocks
    eth_src_mac             => c_eth_src_mac,
    udp_src_port            => c_udp_src_port,
    ip_src_addr             => c_ip_src_addr,

    gn_index                => gn_index,
    ring_info               => c_exp_ring_info,
    sdp_info                => c_exp_sdp_info,
    weighted_subbands_flag  => weighted_subbands_flag,
    nof_crosslets           => c_mm_nof_crosslets,
    crosslets_info          => in_crosslets_info_slv
  );

  -- Check crosslet_info functions
  ASSERT c_crosslets_info_rec = func_sdp_map_crosslets_info(c_crosslets_info_slv) REPORT "Error in func_sdp_map_crosslets_info()" SEVERITY FAILURE;

  -- To view the 32 bit 1GbE offload data more easily in the Wave window
  sdp_offload_data <= sdp_offload_sosi.data(c_word_w-1 DOWNTO 0);
  rx_offload_data <= rx_offload_sosi.data(c_word_w-1 DOWNTO 0);

END tb;
